|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
||									     ||
|| Basic states: wait for P0 and P1, keep DRAM refreshed (using a counter    ||
||	in SRAM location 0xFF).  Responds as follows:		             ||
||									     ||
||   P0:	Dispatch to state whose (OPCODE) value is given in the	     ||
||		  data switches when P0 is pressed.			     ||
||   P1:	Examine contents of SRAM location whose address is	     ||
||		  give in the switches.					     ||
||	P1,P0:	If P0 is pressed momentarily while P1 is being held down,    ||
||		  contents of switches are stored into current SRAM	     ||
||		  location.						     ||
||									     ||
||   Entry to main loop is via state 0, which stores the condition           ||
||	register into SRAM location 0.					     ||
||									     ||
||									     ||
|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||


||LABELS
0x00 = state0
| State zero:
| Copies condition codes into SRAM <0>, then goes to state 5.
|			           ________________________________________
|	 CC Format:   ctl ROM <-   |  N | -C |  E | D0 | P0 | P1 | IOF |  |
|			           ----------------------------------------
|				      <-- Direction of shift <--
0 * | MAR <- 0; !REFR
1 0 | A <- A+A; CC=!C
1 1 | A <- A+A+1; CC=!C
2 0 | A <- A+A; CC=E
2 1 | A <- A+A+1; CC=E
3 0 | A <- A+A; CC=D0
3 1 | A <- A+A+1; CC=D0
4 0 | A <- A+A; CC=P0
4 1 | A <- A+A+1; CC=P0
5 0 | A <- A+A; CC=P1
5 1 | A <- A+A+1; CC=P1
6 0 | A <- A+A; CC=I/O
6 1 | A <- A+A+1; CC=I/O
7 0 | A <- A+A; CC=SP!C
7 1 | A <- A+A+1; CC=SP!C
8 0 | A <- A+A;
8 1 | A <- A+A+1;
9 * | SRAM <- A
A * | A   <- 0
B * | A   <- A+A+1	# 1
C * | A   <- A+A   	# 2
D * | ADR <- 0
E * | ADR <- A+A+1	# 5
F * | NOOP; LDOP;

0x05 = state5
| State five:
| Refreshes DRAM, using SRAM adr 0xFF as a refresh counter.
| Loops until P0 pressed; then enters state 3.
| Displays <B> in lights, untouched.
| Refresh DRAM row, increment refresh counter:
0 * | MAR <- 0xFF; !REFR
1 * | A   <- SRAM
2 * | SRAM<- A+1; !REFR
3 * | DRAM<- A+1; CC; !REFR

4 * | A <- 0; !REFR
5 * | A <- A+A+1
6 * | A <- A+A+1; CC=P0	# (=3)

| If P0 is 1 (depressed): switches into B, go to state 3.
7 1 | B   <- SWT
8 1 | ADR <- 0
9 1 | ADR <- A
A 1 | noop; LDOP

| If P0 is 0 (not depressed): Build a constant, go to state 7 (check P1)
7 0 | ADR <- 0
8 0 | ADR <- A+A+1	# (=7)
9 0 | noop; LDOP

0x07 = state7
| State 7: Entered from state 5 when P0 is not depressed.  If P1 
| is pressed, switches into B and go to state 6.  Otherwise, go back
| to 5.
0 * | A <- A+A; CC=P1; !REFR

| P1 is not pressed.
1 0 | ADR <- 0
2 0 | ADR <- A-1
3 0 | noop; LDOP

| P1 is pressed.
1 1 | B <- SWT
2 1 | ADR <- 0
3 1 | ADR <- A
4 1 | noop; LDOP

0x03 = state3
| State 3: Entered when P0 pressed.
| Wait on ~P0  (refreshing DRAM).
| Refresh DRAM row, increment refresh counter:
0 * | MAR <- 0xFF; !REFR
1 * | A   <- SRAM 
2 * | SRAM<- A+1; !REFR
3 * | DRAM<- A+1; CC; CC=P0; !REFR

| If P0 is 1 (depressed): Stay in state 3.
4 1 | A   <- 0; !REFR
5 1 | A   <- A+1 	# (= 1)
6 1 | ADR <- 0
7 1 | ADR <- A+A+1	# (=3)
8 1 | noop; LDOP

| If P0 is 0 (not depressed): Jump to state given in B.
4 0 | ADR <- 0; !REFR
5 0 | ADR <- B
6 0 | noop; LDOP

0x06 = state6
| State 6: Entered when P1 pressed.
| Examines SRAM location whose address is in the switches.
| Displays address until P1 released, then returns to state 5,
|   displaying contents of that address.
| If P0 pressed while P1 is held down, stores contents of switches
|   in indicated address.
| Refresh DRAM row, increment refresh counter:
0 * | MAR <- 0xFF; !REFR
1 * | A   <- SRAM 
2 * | SRAM<- A+1; !REFR
3 * | DRAM<- A+1; !REFR

4 * | noop; CC; CC=P0; !REFR
5 * | A <- 0 
6 * | A <- A+A+1	# (=1)
7 * | A <- A+A+1	# (=3)
8 * | A <- A+A 		# (=6)

9 * | MAR <- B

| If P0 is 1 (depressed): store SWT in <B>, else no op.
A 1 | SRAM <- SWT; CC=P1
A 0 | noop; CC=P1

| If P1 still depressed, stay in state 6
B 1 | ADR <- 0
C 1 | ADR <- A		# (=6)
D 1 | NOOP; LDOP

| If P1 released: fetch SRAM contents, return to state 5.
B 0 | B   <- SRAM
C 0 | ADR <- 0
D 0 | ADR <- A-1 	# (=5)
E 0 | noop; LDOP

0x80 = state80
|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
||									     ||
|| Increment 16-bit value in SRAM<1> (low order) and SRAM<2> (high order);   ||
||   Leave new low-order value in lights.				     ||
||									     ||
|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

0 * | A   <- 0; !REFR
1 * | A   <- A+1 	# (=1)
2 * | B   <- A+1 	# (=2)

3 * | MAR <- A
4 * | A   <- SRAM
5 * | SRAM<- A+1; CC

6 * | MAR <- B
7 * | B   <- A+1; CC=!C
8 * | A   <- SRAM
9 0 | SRAM <- A+1; CC
9 1 | SRAM <- A; CC

A * | ADR <- 0
B * | ADR <- 0
C * | noop; LDOP

0x81 = state81
|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
||									     ||
|| Examine DRAM location whose adr is in SRAM <1,2>:			     ||
||									     ||
|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

0 * | A   <- 0; !REFR
1 * | A   <- A+1 	# (=1)
2 * | B   <- A+1 	# (=2)

3 * | MAR <- A
4 * | A   <- SRAM
5 * | MAR <- B; !REFR
6 * | DRAM<- SRAM; !REFR
7 * | DRAM <- A; !REFR
8 * | B   <- DRAM; !REFR
9 * | ADR <- 0; !REFR
A * | ADR <- 0
B * | noop; LDOP


0x82 = state82
|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
||									     ||
|| Deposit switches in  DRAM location whose adr is in SRAM <1,2>:	     ||
||									     ||
|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

0 * | A   <- 0; !REFR
1 * | A   <- A+1	# (=1)
2 * | B   <- A+1 	# (=2)

3 * | MAR <- A
4 * | A   <- SRAM
5 * | MAR <- B; !REFR
6 * | DRAM<- SRAM; !REFR
7 * | DRAM<- A; !REFR
8 * | DRAM <- SWT; !REFR
9 * | A <- SWT; !REFR
A * | ADR <- 0
B * | ADR <- 0
C * | noop; LDOP


0xA0 = stateA0
|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
||									     ||
|| Examine uROM location whose adr is in SRAM <1,2>:			     ||
||	Leaves SRAM<1,2> in ADR.					     ||
||									     ||
|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

0 * | A   <- 0; !REFR
1 * | A   <- A+1	# (=1)
2 * | B   <- A+1 	# (=2)
3 * | MAR <- A
4 * | A   <- SRAM
5 * | MAR <- B
6 * | ADR <- SRAM
7 * | ADR <- A
8 * | B   <- uROM
9 * | ADR <- 0
A * | ADR <- 0
B * | noop; LDOP


0x83 = state83
|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
||									     ||
|| Fill DRAM with value in switches, starting at SRAM<1,2>.		     ||
||									     ||
|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

0 * | A   <- 0; !REFR
1 * | A   <- A+1 	# (=1)

2 * | MAR <- A+1 	# (=2)
3 * | B   <- SRAM
4 * | MAR <- A; !REFR  	# (=1)
5 * | DRAM<- SRAM; !REFR # <1>
6 * | DRAM<- B; !REFR	# (=SRAM<2>)
7 * | DRAM<- SWT; !REFR

8 * | B   <- A+1; !REFR	# (=2)
9 * | A   <- SRAM
A * | SRAM<- A+1; CC	

B * | MAR <- B; CC=!C
C * | A   <- SRAM
D 0 | SRAM <- A+1; CC; CC=!C
D 1 | SRAM <- A; CC; CC=!C

E * | ADR <- 0

F 1 | noop ## HUH?!?!? (No carry, No operation)
F 0 | ADR <- 0; LDOP

|| GENERAL REGISTER TESTING

0xF0 = stateF0
| Load GR1 with switches.
0 * | noop <- SWT; LDGR1; !REFR
1 * | ADR <- 0
2 * | ADR <- 0
3 * | noop; LDOP

0xF1 = stateF1
| Load GR2 with switches.
0 * | noop <- SWT; LDGR2; !REFR
1 * | ADR <- 0
2 * | ADR <- 0
3 * | noop; LDOP

0xF2 = stateF2
| Load lights from GR1.
0 * | B <- GR1; !REFR
1 * | ADR <- 0
2 * | ADR <- 0
3 * | noop; LDOP

0xF3 = stateF3
| Load lights from GR2.
0 * | B <- GR2; !REFR
1 * | ADR <- 0
2 * | ADR <- 0
3 * | noop; LDOP

|| PROGRAM COUNTER TESTING

0xE0 = stateE0
| Load PCL from switches
0 * | PCL <- SWT; !REFR
1 * | ADR <- 0
2 * | ADR <- 0
3 * | noop; LDOP

0xE1 = stateE1
| Load PCH from switches.
0 * | PCH <- SWT; !REFR
1 * | ADR <- 0
2 * | ADR <- 0
3 * | noop; LDOP

0xE2 = stateE2
| Drive PCL and PCH onto the bus, leaving PCL in the lights.
0 * | B <- PCL; !REFR
1 * | noop <- PCH
2 * | ADR <- 0
3 * | ADR <- 0
4 * | noop; LDOP

0xE3 = stateE3
| Drive PCL and PCH onto the bus, leaving PCH in the lights.
0 * | noop <- PCL; !REFR
1 * | B <- PCH
2 * | ADR <- 0
3 * | ADR <- 0
4 * | noop; LDOP

0xE4 = stateE4
| Does pcfetch1 into lights.
0 * | DRAM <- PCL
1 * | DRAM <- PCH
2 * | noop
3 * | B <- DRAM; LDOP; ADR+

|| REFRESH TESTING

0xD0 = stateD0
| Loop until P0 is pressed, then jump to 0xD1.  Does *not* software
| refresh any rows of DRAM (the hardware refresh is enabled).
0 * | A <- B; CC; CC=P0; !REFR

| P0 not pressed.
1 0 | ADR <- 0
2 0 | ADR <- A
3 0 | noop; LDOP

| P0 pressed.
1 1 | ADR <- 0
2 1 | ADR <- A+1
3 1 | noop; LDOP

0xD1 = stateD1
| Waits until P0 is released, then jumps to state 0.  Does *not*
| refresh any rows.
0 * | noop; CC; CC=P0; !REFR

| P0 not pressed.
1 0 | ADR <- 0
2 0 | ADR <- 0
3 0 | noop; LDOP

| P0 pressed.
1 1 | ADR <- 0
2 1 | ADR <- A
3 1 | noop; LDOP

|| Stack Pointer tests

0xC0 = stateC0
| Load SPa from switches
0 * | SPa <- SWT; !REFR
1 * | ADR <- 0
2 * | ADR <- 0
3 * | noop; LDOP

0xC1 = stateC1
| Load SPb from switches.
0 * | SPb <- SWT; !REFR
1 * | ADR <- 0
2 * | ADR <- 0
3 * | noop; LDOP

0xC2 = stateC2
| Display SPa in lights.
0 * | B <- SPa; !REFR
1 * | ADR <- 0
2 * | ADR <- 0
3 * | noop; LDOP

0xC3 = stateC3
| Display SPb in lights.
0 * | B <- SPb; !REFR
1 * | ADR <- 0
2 * | ADR <- 0
3 * | noop; LDOP

0xC4 = push1
| Push switches onto the stack.
0 * | ADR <- 0; !REFR
1 * | ADR <- 0; !REFR
2 * | DRAM <- SPa; !REFR
3 * | DRAM <- SPb; !REFR
4 * | DRAM <- SWT; !REFR
5 * | SPa <- SPa+1; CC=SP!C; !REFR
6 0 | SPb <- SPb+1; LDOP; ADR+
6 1 | NOOP; LDOP; ADR+

0xC5 = pop1
| Pop top of stack into lights.
0 * | ADR <- 0; !REFR
1 * | ADR <- 0; !REFR
2 * | SPa,DRAM <- SPa-1; CC=SP!C; !REFR
3 1 | SPb,DRAM <- SPb-1; !REFR
3 0 | DRAM <- SPb; !REFR
4 * | B <- DRAM; LDGR1; LDOP; ADR+; !REFR

|| IO tests

0x70 = state70
| Load IO register from switches.
0 * | I/OADR <- B
1 * | I/O <- SWT
2 * | ADR <- 0
3 * | ADR <- 0
4 * | noop; LDOP

0x71 = state71
| Load IO register from switches.
0 * | I/OADR <- B
1 * | I/O <- SWT
2 * | ADR <- 0
3 * | ADR <- 0
4 * | noop; LDOP

0x72 = state72
| Load IO register from switches.
0 * | I/OADR <- B
1 * | I/O <- SWT
2 * | ADR <- 0
3 * | ADR <- 0
4 * | noop; LDOP

0x73 = state73
| Load IO register from switches.
0 * | I/OADR <- B
1 * | I/O <- SWT
2 * | ADR <- 0
3 * | ADR <- 0
4 * | noop; LDOP

0x74 = state74
| Load IO register from switches.
0 * | I/OADR <- B
1 * | I/O <- SWT
2 * | ADR <- 0
3 * | ADR <- 0
4 * | noop; LDOP

0x75 = state75
| Load IO register from switches.
0 * | I/OADR <- B
1 * | I/O <- SWT
2 * | ADR <- 0
3 * | ADR <- 0
4 * | noop; LDOP

0x76 = state76
| Load IO register from switches.
0 * | I/OADR <- B
1 * | I/O <- SWT
2 * | ADR <- 0
3 * | ADR <- 0
4 * | noop; LDOP

0x77 = state77
| Load IO register from switches.
0 * | I/OADR <- B
1 * | I/O <- SWT
2 * | ADR <- 0
3 * | ADR <- 0
4 * | noop; LDOP

0x78 = state78
| Drive the switches to the I/O data line until P1 is pressed.
| First, check that Xmitter is ready
0 * | A <- 0
1 * | A <- A+A+1
2 * | A <- A+A
3 * | I/OADR <- A-1
4 * | A <- I/O
5 * | A <- A+A; CC; CC=N
6 0 | noop; LDOP
6 1 | NOOP
7 * | I/OADR <- 0
8 * | I/O <- SWT; CC; CC=P1
9 0 | noop; LDOP
A 1 | ADR <- 0
B 1 | ADR <- 0
C 1 | noop; LDOP

