|||||  Helper functions

0x90 = pop1(x)
| pops next byte into SRAM x
| 5 cycles
0 * | MAR <- uROM ; ADR+
1 * | DRAM,SPa <- SPa-1; LatchCC ; CC=!C
2 0 | DRAM,SPa <- SPb-1
2 1 | DRAM <- SPb
3 * | SRAM <- DRAM
4 * | OPCODE <- uROM ; ADR+

0x92 = hmult(x)
| GR1 = 0000 abcd
| GR2 = efgh 0000
| puts abcd*efgh in S<x>
| 24 cycles

0 * | A <- GR2
1 * | LDGR2 ; NOOP <- A+A
2 * | A <- A ; LatchCC ; CC=N
3 * | A <- GR1
4 0 | B <- 0
4 1 | B <- A

5 * | A <- GR2
6 * | LDGR2 ; NOOP <- A+A
7 * | A <- A ; LatchCC ; CC=N
8 * | A <- GR1
9 0 | A <- B
9 1 | A <- A+B
a * | B <- A+B

b * | A <- GR2
c * | LDGR2 ; NOOP <- A+A
d * | A <- A ; LatchCC ; CC=N
e * | A <- GR1
f 0 | A <- B
f 1 | A <- A+B  # WARP!

0x93 = hmult_2(x)
| Warped continuation of hmult (0x92)
0 * | B <- A+B

1 * | A <- GR2
2 * | A <- A ; LatchCC ; CC=N
3 * | A <- GR1
4 0 | A <- B
4 1 | A <- A+B
5 * | MAR <- uROM ; ADR+
6 * | SRAM <- A+B

7 * | OPCODE <- uROM ; ADR+

0x94 = lln(x,0x0F,y)
| load low nybble of x into low nybble of GR1
| load low nybble of y into high nybble of GR2
| 12 cycles
0 * | MAR <- uROM ; ADR+
1 * | A <- SRAM
2 * | MAR <- uROM ; ADR+
3 * | B <- SRAM
4 * | LDGR1 ; NOOP <- A&B

5 * | MAR <- uROM ; ADR+
6 * | A <- SRAM
7 * | A <- A+A
8 * | A <- A+A
9 * | A <- A+A
a * | LDGR2 ; NOOP <- A+A

b * | OPCODE <- uROM ; ADR+

0x95 = lhn(x,0x0F,y,0xF0)
| load low nybble of x into low nybble of GR1
| load high nybble of y into high nybble of GR2
| 11 cycles
0 * | MAR <- uROM ; ADR+
1 * | A <- SRAM
2 * | MAR <- uROM ; ADR+
3 * | B <- SRAM
4 * | LDGR1 ; NOOP <- A&B

5 * | MAR <- uROM ; ADR+
6 * | A <- SRAM
7 * | MAR <- uROM ; ADR+
8 * | B <- SRAM
9 * | LDGR2 ; NOOP <- A&B

a * | OPCODE <- uROM ; ADR+

0x96 = hhn(x,y,0xF0)
| load high nybble of x into low nybble of GR1
| load high nybble of y into high nybble of GR2
| 11 cycles

0 * | MAR <- uROM ; ADR+
1 * | A <- SRAM
2 * | A <- A ; LatchCC ; CC=N

3 0 | A <- A+A ; LatchCC ; CC=N
3 1 | A <- A+A+1 ; LatchCC ; CC=N
4 0 | A <- A+A ; LatchCC ; CC=N
4 1 | A <- A+A+1 ; LatchCC ; CC=N
5 0 | A <- A+A ; LatchCC ; CC=N
5 1 | A <- A+A+1 ; LatchCC ; CC=N
6 0 | LDGR1 ; NOOP <- A+A ; LatchCC ; CC=N
6 1 | LDGR1 ; NOOP <- A+A+1 ; LatchCC ; CC=N

5 * | MAR <- uROM ; ADR+
6 * | A <- SRAM
7 * | MAR <- uROM ; ADR+
8 * | B <- SRAM
9 * | LDGR2 ; NOOP <- A&B

a * | OPCODE <- uROM ; ADR+

0x97 = shiftl4(x)
| shift S<x> 4 bits to the left, filling in with zeros
| 7 cycles
0 * | MAR <- uROM ; ADR+
1 * | A <- SRA
2 * | A <- A+A
3 * | A <- A+A
4 * | A <- A+A
5 * | SRAM <- A+A
6 * | OPCODE <- uROM ; ADR+

0x98 = shiftr4(x,0x0F)
| shift S<x> 4 bits to the right, filling in with zeros
| Really:  rotate left 4 bits, mask out top four.
| 10 cycles
0 * | MAR <- uROM ; ADR+
1 * | A <- SRAM
2 * | A <- A ; LatchCC ; CC=N

3 0 | A <- A+A ; LatchCC ; CC=N
3 1 | A <- A+A+1 ; LatchCC ; CC=N
4 0 | A <- A+A ; LatchCC ; CC=N
4 1 | A <- A+A+1 ; LatchCC ; CC=N
5 0 | A <- A+A ; LatchCC ; CC=N
5 1 | A <- A+A+1 ; LatchCC ; CC=N
6 0 | A <- A+A ; LatchCC ; CC=N
6 1 | A <- A+A+1 ; LatchCC ; CC=N

7 * | B <- uROM ; ADR+
8 * | SRAM <- A&B

9 * | OPCODE <- uROM ; ADR+


0x99 = addr(x,y)
| S<y> += S<x>, latch CCs
| 6 cycles
0 * | MAR <- uROM ; ADR+
1 * | A <- SRAM
2 * | MAR <- uROM ; ADR+
3 * | B <- SRAM
4 * | SRAM <- A+B ; LatchCC
5 * | OPCODE <- uROM ; ADR+

0x9a = addrc(x,y)
| S<y> += S<x>+carry, latch CCs
| 6 cycles
0 * | MAR <- uROM ; ADR+
1 * | A <- SRAM
2 * | MAR <- uROM ; ADR+
3 * | B <- SRAM ; CC=!C
4 0 | SRAM <- A+B+1 ; LatchCC
4 1 | SRAM <- A+B ; LatchCC
5 * | OPCODE <- uROM ; ADR+

0x9b = cp(x,y)
| S<y> = S<x>
| 5 cycles
0 * | MAR <- uROM ; ADR+
1 * | A <- SRAM
2 * | MAR <- uROM ; ADR+
3 * | SRAM <- A
4 * | OPCODE <- uROM ; ADR+

0x9c = split(x,0x0F,y,0xF0)
| takes S<x> = abcd efgh and returns S<x> = 0000 abcd 
| and S<y> = efgh 0000
| 13 cycles
0 * | MAR <- uROM ; ADR+
1 * | A <- SRAM
2 * | A <- A ; LatchCC ; CC=N

3 0 | A <- A+A ; LatchCC ; CC=N
3 1 | A <- A+A+1 ; LatchCC ; CC=N
4 0 | A <- A+A ; LatchCC ; CC=N
4 1 | A <- A+A+1 ; LatchCC ; CC=N
5 0 | A <- A+A ; LatchCC ; CC=N
5 1 | A <- A+A+1 ; LatchCC ; CC=N
6 0 | A <- A+A ; LatchCC ; CC=N
6 1 | A <- A+A+1 ; LatchCC ; CC=N

7 * | B <- uROM ; ADR+
8 * | SRAM <- A&B
9 * | MAR <- uROM ; ADR+
a * | B < uROM ; ADR+
b * | SRAM <- A&B

c * | OPCODE <- uROM ; ADR+

||||| Arithmetic functions

0xa1 = add1()
| replace a1 on the stack with a1+b1
| 10 cycles
0 * | DRAM,SPa <- SPa-1 ; LatchCC ; CC=!C
1 0 | DRAM,SPb <- SPb-1
1 1 | DRAM <- SPb
2 * | A <- DRAM		# pop into B reg
3 * | DRAM <- SPa-1 ; LatchCC ; CC=!C
4 0 | DRAM <- SPb-1
4 1 | DRAM <- SPb
5 * | B <- DRAM		# top into A reg
6 * | DRAM <- SPa-1 ; LatchCC ; CC=!C
7 0 | DRAM <- SPb-1
7 1 | DRAM <- SPb
8 * | DRAM <- A+B	# A+B into top
9 * | OP <- uROM ; ADR+

0xa2 = add(x,y)
| add S<x>, S<y> and put result on stack, latch CC
| 10 cycles
0 * | MAR <- uROM ; ADR+
1 * | A <- SRAM
2 * | MAR <- uROM ; ADR+
3 * | B <- SRAM
4 * | DRAM <- SPa
5 * | DRAM <- SPb
6 * | DRAM <- A+B ; LatchCC
7 * | SPa <- SPa+1 ; CC=SP!C
8 0 | SPb <- SPb+1
8 1 | OPCODE <- uROM; ADR+
9 * | OPCODE <- uROM ; ADR+

0xa3 = addc(x,y)
| add S<x>, S<y>, carry and put result on stack, latch CC
| 10 cycles
0 * | MAR <- uROM ; ADR+
1 * | A <- SRAM
2 * | MAR <- uROM ; ADR+
3 * | B <- SRAM
4 * | DRAM <- SPa
5 * | DRAM <- SPb ; CC=!C
6 0 | DRAM <- A+B+1 ; LatchCC
6 1 | DRAM <- A+B ; LatchCC
7 * | SPa <- SPa+1 ; CC=SP!C
8 0 | SPb <- SPb+1
8 1 | OPCODE <- uROM; ADR+
9 * | OPCODE <- uROM ; ADR+


0xa4 = sub1()
| replace a1 on the stack with a1-b1
| 10 cycles
0 * | DRAM,SPa <- SPa-1 ; LatchCC ; CC=!C
1 0 | DRAM,SPb <- SPb-1
1 1 | DRAM <- SPb
2 * | A <- DRAM		# pop into B reg
3 * | DRAM <- SPa-1 ; LatchCC ; CC=!C
4 0 | DRAM <- SPb-1
4 1 | DRAM <- SPb
5 * | B <- DRAM		# top into A reg
6 * | DRAM <- SPa-1 ; LatchCC ; CC=!C
7 0 | DRAM <- SPb-1
7 1 | DRAM <- SPb
8 * | DRAM <- A-B	# A+B into top
9 * | OPCODE <- uROM ; ADR+

0xa5 = sub(x,y)
| S<x> - S<y> on stack, latch CC
| 10 cycles
0 * | MAR <- uROM ; ADR+
1 * | A <- SRAM
2 * | MAR <- uROM ; ADR+
3 * | B <- SRAM
4 * | DRAM <- SPa
5 * | DRAM <- SPb
6 * | DRAM <- A-B ; LatchCC
7 * | SPa <- SPa+1 ; CC=SP!C
8 0 | SPb <- SPb+1
8 1 | OPCODE <- uROM; ADR+
9 * | OPCODE <- uROM ; ADR+

0xa6 = subb(x,y)
| S<x> - S<y> - borrow on stack, latch CC
| 10 cycles
0 * | MAR <- uROM ; ADR+
1 * | A <- SRAM
2 * | MAR <- uROM ; ADR+
3 * | B <- SRAM
4 * | DRAM <- SPa
5 * | DRAM <- SPb
6 0 | DRAM <- A-B-1 ; LatchCC
6 1 | DRAM <- A-B ; LatchCC
7 * | SPa <- SPa+1 ; CC=SP!C
8 0 | SPb <- SPb+1
8 1 | OPCODE <- uROM; ADR+
9 * | OPCODE <- uROM ; ADR+
