#include <stdio.h>
#include "sim.h"

/* determines whether its argument has bits in the higher order byte
 * sets ALU_c_out and truncates if it does */
ctst(x)
   int x;
{
     if (x > 255) ALU_c_out = 0;
     return(x & 0xFF);
}

/*
 * computes and stores ALU output (data and flags) 
 * carry is set during the calculation by ctst() above.
 * Z is set before the operation if the two operands are equal.
 * N and ALU0 are the msb and the lsb of the result.
 */

simulate_alu(ctlbits, value, RegA, RegB)
   int ctlbits, *value, RegA, RegB;
{
     int m,cin,fcn,z,tmp, ALU_data;
     m = ctlbits & 0x01;
     fcn = (ctlbits >> 2) & 0x0F;
     cin = (ctlbits >> 1) & 0x01; 
     ALU_c_out = 1;		/* init: no carry out */
     
     switch (fcn)  {
     case 0: 
	  if (m) ALU_data = (~RegA&0xFF);
	  else 	if (cin) ALU_data = RegA;
	  else ALU_data = ctst(RegA + 1);
	  break;
     case 1: 
	  if (m) ALU_data = (~(RegA | RegB)&0xFF);
	  else { 	(tmp = (RegA | RegB));
			if (cin) ALU_data = tmp; 
			else ALU_data = ctst(tmp + 1); }
	  break;
     case 2:
	  if (m) ALU_data = (~RegA & RegB)&0xFF;
	  else {	tmp = RegA | (~RegB)&0xFF;
			if (cin) ALU_data = tmp;
			else ALU_data =  ctst(tmp+1); }
	  break;
     case 3:
	  if (m) ALU_data = 0x00;
	  else	if (cin) ALU_data = 0xFF;
	  else ALU_data = 0x00; 
	  break;
     case 4:
	  if (m) ALU_data = ~(RegA & RegB)&0xFF;
	  else {	tmp = ctst(RegA+((RegA&~RegB)&0xFF));
			if (cin) ALU_data = tmp;
			else ALU_data = ctst(tmp+1); }
	  break;
     case 5:
	  if (m) ALU_data = ~RegB&0xFF;
	  else {	tmp = ctst( (RegA|RegB) + ((RegA & ~RegB)&0xFF) );
			if (cin) ALU_data = tmp;
			else ALU_data = ctst(tmp+1); }
	  break;
     case 6:
	  if (m) ALU_data = ((RegA & ~RegB) | (~RegA & RegB))&0xFF;
	  else {	tmp = RegA + (RegB ^ 0xFF);
			if (cin) ALU_data = ctst(tmp);
			else ALU_data =  ctst(tmp+1); }
	  break;
     case 7:
	  tmp = (RegA & ~RegB)&0xFF;
	  if (m) ALU_data = tmp;
	  else	if (cin) ALU_data = ctst(tmp + 0xFF);
	  else ALU_data = tmp; 
	  break;
     case 8:
	  if (m) ALU_data = (~RegA | RegB) & 0xFF;
	  else {	tmp = ctst(RegA+(RegA&RegB));
			if (cin) ALU_data = tmp;
			else ALU_data = ctst(tmp+1); }
	  break;
     case 9:
	  if (m) ALU_data = (((RegA & RegB) | (~RegA & ~RegB))&0xFF);
	  else {	tmp = RegA + RegB;
			if (cin) ALU_data = ctst(tmp);
			else ALU_data = ctst(tmp+1); }
	  break;
     case 10:
	  if (m) ALU_data = RegB;
	  else {	tmp = (RegA | ~RegB) + (RegA & RegB);
			if (cin) ALU_data = ctst(tmp);
			else ALU_data = ctst(tmp+1); }
	  break;
     case 11: 
	  tmp = RegA & RegB;
	  if (m) ALU_data = tmp;
	  else	if (cin) ALU_data = ctst(tmp + 0xFF);
	  else ALU_data = tmp; 
	  break;
     case 12:
	  if (m) ALU_data = 0xFF;
	  else {	tmp = (RegA + RegA);
			if (cin) ALU_data = ctst(tmp);
			else ALU_data = ctst(tmp+1); }
	  break;
     case 13:
	  if (m) ALU_data = (RegA | ~RegB)&0xFF;
	  else {	tmp = ((RegA | RegB)+RegA);
			if (cin) ALU_data = ctst(tmp);
			else ALU_data = ctst(tmp+1); }
	  break;
     case 14:
	  if (m) ALU_data = RegA | RegB;
	  else {	tmp = ((RegA | ~RegB)+RegA);
			if (cin) ALU_data = ctst(tmp);
			else ALU_data = ctst(tmp+1); }
	  break;
     case 15:
	  if (m) ALU_data = RegA;
	  else	if (cin) ALU_data = ctst(RegA + 0xFF);
	  else ALU_data = RegA; 
	  break;
     }
     
     z = (ALU_data==0xFF)?1:0;
     ALU_flags = ((ALU_data & 0x01) << 3) | (( z & 0x01) << 2) |
	  ((ALU_c_out & 0x01) << 1) | ((ALU_data & 0x80) >> 7);
     *value = ALU_data;
}
