#include <stdio.h>
#include "display.h"
#include "sim.h"

/* Display.c contains procedures which update the display data
 * structure, put the various displays on the screen, and disable and
 * enable various combinations of display sections.
 */

/*
 * Main routine:
 * Update the display of *dd with value vv.
 */

UpDpy(dd, vv)
   register struct Display *dd;
   unsigned long vv;
{
     struct Display *ee;
     unsigned long v = vv;
     unsigned long oldv = dd->Value;
     
     v >>= dd->Shift;
     v &= dd->Mask;
     if (v != oldv)				/* No change.		*/
	  { dd->Value = v;
	    if (dd->Flags & F_ENAB) {		/* Disabled?		*/
		 dd->Value = v;
		 (*(dd->OutP))(dd);		/* nope, do the output	*/
	    }
       }
     if (ee = (dd->Addl))			/* Additional output	*/
	  UpDpy(ee, vv);
}

/*
 * Output routines.
 * Each is called with a single arg: the struct Display to be output.
 */

DpyHex(dd)	    	    		/* Hex */
   struct Display *dd;
{
     sprintf(DpyBuf, "0x%0*x", dd->Size, dd->Value);
     DpyOut(dd);
}

/* Kludge to display control bits in reasonable format */
DpyCtlRom(dd)
   struct Display *dd;
{
     char *cc, outbuf[BUFSIZ];
     int i = dd->Value, j = dd->Size;
     static int map[32] = {
	  0, 1, 2, 3, 5, 6, 7, 8, 11, 12, 14, 15, 16, 17, 19, 20, 23,
	  24, 26, 27, 28, 29, 31, 32, 35, 36, 37, 38, 39, 41, 42, 43};

     cc = DpyBuf+j;
     *cc-- = 0;
     while (j--) {
	  if (i & 1) *cc-- = '1';
	  else *cc-- = '0';
	  i >>= 1;
     }

     /* This could be a lot faster but I'm too tired.. */
     for (i=0; i<44; i++)
	  outbuf[i] = ' ';
     
     for (i=0; i<32; i++)
	  outbuf[map[i]] = DpyBuf[i];
     outbuf[44] = '\0';

     strcpy(DpyBuf, outbuf);
     
     DpyOut(dd);
}

DpyBin(dd)/* NC */			/* Binary register or value	*/
   struct Display *dd;
{
     register char *cc;
     register i = dd->Value, j = dd->Size;
     cc = DpyBuf+j;
     *cc-- = 0;
     while (j--) {
	  if (i & 1) *cc-- = '1';
	  else *cc-- = '0';
	  i >>= 1;
     }
     DpyOut(dd);		/* Display in reverse/normal video.	*/
}

DpyDec(dd)/* NC */			/* Decimal register or value	*/
 struct Display *dd;
 {	int v = dd->Value;
	sprintf(DpyBuf, "%3d", v & 0xFF);
	DpyOut(dd);		/* Display in reverse/normal video.	*/
 }

DpyTxt(dd)/* NC */			/* i[th] Text message		*/
 register struct Display *dd;
 {	char **tbl = (char **) dd->Aux0;
	int v = dd->Value;
	register i;
	sprintf(DpyBuf, "%s", tbl[v]);
	i = strlen(DpyBuf);
	while (i < dd->Size) DpyBuf[i++] = ' ';
	DpyBuf[i] = 0;
	DpyOut(dd);
 }

DpyLong(dd)/* NC */			/* Long message */
 register struct Display *dd;
 {
	sprintf(DpyBuf, "%*ld", dd->Size, dd->Value);
	DpyOut(dd);		/* Display in reverse/normal video.	*/
 }

DpyQuot(dd)			/* For use in time display - quotient */
 register struct Display *dd;
 {	char fr[4];
	int fract;
	long integer, temp1, temp2;

	integer = (dd->Value)/(dd->Aux0);
	temp1 = (dd->Value)*1000;
	temp2 = integer*1000;
	temp1 -= temp2;
	fract = temp1;

	fract = ((dd->Value)*1000)/(dd->Aux0);
	fr[3] = 0;
	fr[2] = (fract % 10)+'0';	fract = fract/10;
	fr[1] = (fract % 10)+'0';	fract = fract/10;
	fr[0] = (fract % 10)+'0';

	sprintf(DpyBuf, "%4ld.%s", integer, fr);
	DpyOut(dd);		/* Display in reverse/normal video.	*/
 }

/*
 * Add a display to the list:
 * Dpy returns a ptr to a DISPLAY structure
 */

struct Display *Dpy(x, y, outp, size, flags) /* NC */
   int (*outp)();
{
     register struct Display *dd = (struct Display *) malloc(sizeof *dd);
     dd->Value = 0;
     dd->x = x;	dd->y = y;
     dd->Size = size;
     dd->OutP = outp;
     dd->Flags = flags;
     dd->Addl = 0;
     dd->Next = DpyList;
     dd->Shift = 0;	dd->Mask = -1;
     return (DpyLast = DpyList = dd);
}

/* Add a sub-display to the last display.
 * The new display will use the same value, shifted & masked according to bits.
 */

struct Display *DpyAlso(x, y, outp, size, flags, bits)/* NC */
     int (*outp)();
 {	struct Display *dd = DpyList, *ee;
	register i;
	ee = Dpy(x, y, outp, size, flags);
	DpyList = dd;
	ee->Next = 0;
	ee->Addl = dd->Addl;
	dd->Addl = ee;
	for (i=0; !(bits & 1);) { bits >>= 1; i++; }
	ee->Shift = i;
	ee->Mask = bits;
	return ee;
 }

/* Add aux parameters to last-built display:
 */

DpyAux(a, b, c)/* NC */
 long a, b, c;
 {	DpyLast->Aux0 = a;
	DpyLast->Aux1 = b;
	DpyLast->Aux2 = c;
 }

/* Refresh display:
 */

DpyRef()
{
     register struct Display *dd = DpyList, *ee;
     unsigned long v;
     DpyClr();			/* Clear screen.		*/
     DpyFix();			/* Fixed display portion.	*/
     
     while (dd) {		/* Update all display fields	*/
	  v = dd->Value;
	  dd->Value = ~v;
	  UpDpy(dd, v);
	  if (ee = dd->Addl)
	  while (ee) {
	       ee->Value = ~(ee->Value); UpDpy(ee, v); 
	       ee = ee->Addl;
	  }
	  dd = dd->Next;
     }
}

/* Map fn onto all struct Displays
 */

DpyAll(fn)/* NC */
 int (*fn)();
 {	register struct Display *dd = DpyList, *ee;

	while (dd)
	 { fn(dd);
	   if (ee = dd->Addl)
		while (ee)	{ fn(ee);
				  ee = ee->Addl;
				}
	   dd = dd->Next;
	 }
 }

/* Fill a display field with given character:
 */

DpyField(dd, pat)/* NC */
 struct Display *dd;
 char pat;
 {	register char *cc;
	register i;
	cc = DpyBuf;
	for (i = dd->Size; i--; ) *cc++ = pat;
	*cc = 0;
	DpyOut(dd);
 }

/* Disable, Blank out a display field
 */
DpyOff(dd)/* NC */
 struct Display *dd;
 {	DpyField(dd, ' ');
	dd->Flags &= ~F_ENAB;
 }

/*
 *  Re-enable a display field
 */

DpyOn(dd)/* NC */
   struct Display *dd;
{
     dd->Flags |= F_ENAB;
     DpyRef1(dd);
     if (dd->Addl) return DpyOn(dd->Addl);
     /* XXX */
     else return NULL;
}   

/*
 * Refresh a display field
 */

DpyRef1(dd)/* NC */
   struct Display *dd;
{
     unsigned long v = dd->Value++;
     UpDpy(dd, v);
}

/* Disable all displays: blanks fields.
 */

DpyDisable()/* NC */
 {	DpyAll(DpyOff);
 }

/* Re-enable all displays:
 */

DpyEnable()/* NC */
 {	DpyAll(DpyOn);
 }


/* "display in normal or reverse video" */
DpyOut(dd)/* NC */
 struct Display *dd;
 {	if (dd->Flags & F_REV) DpyRev(dd->x, dd->y, DpyBuf);
	else DpyNrm(dd->x, dd->y, DpyBuf);
	DpyFlu();
 }

/* Convert integer to binary:
 */

char *bin(n,digits)/* NC */
  int n,digits;
  {	register int i, j=0;
	static char buf[100];

	for (i = 1 << (digits-1); i > 0; i >>= 1)
		buf[j++] = (n & i) ? '1':'0';
	buf[j] = 0;
	return buf;
 }

