
#include <stdio.h>
#include "sim.h"
#include "display.h"

/* Format.c contains procedures which set up the display data
 * structure and the help display, and which perform window
 * servicing functions.
 */

int	WindTop = 15,			/* Display window for user fns	*/
	WindBot = 20,
	WindY;				/* Current window line		*/

long Speed = 800000;		/* nanocycles/second */

char *LdSelTxt[] = {
     " Ld Op   ", " Ld Adr  ", " Ld RegA ", " Ld RegB ",
     " SRAM Wr ", " Ld MAR  ", " DRAM Wr ", " Ld IO   ",
     " Ld SPa  ", " Ld SPb  ", " Ld SPa&D", " Ld SPb&D",
     " Ld PClo ", " Ld PChi ", "!!! 14 !!", " (noop)  ",
};
char *BSelTxt[]  = {
     " SWITCHES", " uCd ROM ", "  SRAM   ", "  DRAM   ",
     "  Dr IO  ", "   ALU   ", "   SP    ", "  PClo   ",
     "  PChi   ", "  Reg 1  ", "  Reg 2  ", "!!! 11 !!",
     "!!! 12 !!", "!!! 13 !!", "!!! 14 !!", "!!! 15 !!",
};

/* creates display structures for all fields on the screen */
DpyInit()
{
     int r = F_ENAB+F_REV, x, y;
     
     TimDpy = Dpy(x=53, 0, DpyLong, 7, r);
     DpyAlso(x+=14, 0, DpyQuot, 8, r, -1);
     DpyAux(Speed, 0, 0); /* XXX My hack */
     UcntDpy = Dpy (x=53, y=1, DpyLong, 7, r);

     /* Line 1: */
     RADpy = Dpy(x=10, y=3, DpyBin, 8, r);
     DpyAlso(x+=11, y, DpyDec, 3, r, -1);
     AdrDpy = Dpy(x+=35, y, DpyBin, 13, r);
     RBDpy = Dpy(x=10, y=4, DpyBin, 8, r);
     DpyAlso(x+=11, y, DpyDec, 3, r, -1);
     URomDpy = Dpy(x+=40, y, DpyBin, 8, r);

     /* Line 2: */
     ALUoutDpy = Dpy(x=14, y+=2, DpyBin, 4, r);
     CCDpy = Dpy(x+=14, y, DpyBin, 8, r);
     MarDpy = Dpy(x+=33, y, DpyBin, 8, r);

     /* Line 3: */
     OpDpy = Dpy(x=10, y+=1, DpyBin, 8, r);
     PhDpy = Dpy(x+=20, y, DpyBin, 4, r);
     DregDpy = Dpy(x+=31, y, DpyBin, 8, r);

     /* Line 4: */
     CtlDpy = Dpy(x=18, y+=1, DpyCtlRom, 32, r);

     /* called with ctl_out_16 ... */
     DpyAlso(x-=1, y+=4, DpyTxt, 11, r, 0xF000000);
     DpyAux((int) BSelTxt, 0, 0);
     DpyAlso(x+=40, y, DpyTxt, 11, r, 0xF0000000);
     DpyAux((int) LdSelTxt, 0, 0);

     /* Line 5: */
     SwtDpy = Dpy(x=14, y-=2, DpyBin, 8, r);
     DpyAlso(x+=25, y, DpyBin, 1, r, 0x100);
     DpyAlso(x+=8, y, DpyBin, 1, r, 0x200);

     /* Line 6: */
     BusDpy = Dpy(x=38, y+=2, DpyBin, 8, r);

     /* Real Lines 4 and 5 */
     GR1Dpy = Dpy(33, 3, DpyBin, 8, r);
     DpyAlso(44, 3, DpyDec, 3, r, -1);
     GR2Dpy = Dpy(33, 4, DpyBin, 8, r);
     DpyAlso(44, 4, DpyDec, 3, r, -1);

     /* Real Line 6 */
     CCbitDpy = Dpy(46, 6, DpyBin, 1, r);

     /* Real Line 10: */
     SPoutDpy = Dpy(18, 9, DpyHex, 4, r);
     PCoutDpy = Dpy(30, 9, DpyHex, 4, r);

}	

/* Do the fixed portion of the display:
 */

int _pf_y;
pf(cc)
 char *cc;
 { 	DpyNrm(0, _pf_y++, cc);
 }

   DpyFix()
{
     _pf_y = 0;
     pf(
"The Most Excellent MAYBE Simulator    TIME (cycles): XXXXXXX (Sec) XXXXXXXX");
     pf(
"                                           (uInsts): XXXXXXX");
     ++_pf_y;
     pf(
"    REGA: XXXXXXXX = DDD   GR1 = XXXXXXXX = DDD    ADR: XXXXXXXXXXXXX");
     pf(
"    REGB: XXXXXXXX = DDD   GR2 = XXXXXXXX = DDD   UCODE ROM: XXXXXXXX");
     pf(
"              AFCN");
     pf(
"   ALU FLAGS: XXXX  CC REG: XXXXXXXX  CC BIT: X    SRAM MAR: XXXXXXXX");
     pf(
"  OPCODE: XXXXXXXX     PHASE: XXXX                 DRAM REG: XXXXXXXX");
     pf(
"     CTL ROM OUT: XXXXXXXXXXXXXXXX");
     pf(
"       Stack Ptr:         PC:");
     pf(
"    SWITCHES: XXXXXXXX             P0: X   P1: X");
     ++_pf_y;
     pf(
"                 XXXXXXXXXXX          XXXXXXXX           XXXXXXXXXXX");
     pf(
"    BUS:            DRIVE               DATA                LOAD");
}

/*
 * Clear display window:
 */
WindClr() /* NC */
{
     register i;
     for (i=WindTop; i <= WindBot; i++) DpyCLine(0, i);
     WindY = WindTop;
}

/*
 * Go to next window line:
 */
WindNext() /* NC */
{
     if (++WindY > WindBot) WindY = WindTop;
     DpyCLine(0, WindY);
}
