#include <stdio.h>
#include <sgtty.h>
#include <sys/ioctl.h>
#include <signal.h>

struct sgttyb tty;		/* where tty mode resides */

#define vt100

#define	INPX	10
#define	INPY	22

#define MSGX	0
#define MSGY	23

char MsgFlag = 0;		/* Is there a pending message?		*/

Msg(a0, a1, a2, a3, a4, a5, a6)
{
     char buf[100];
     ClrMsg();
     sprintf(buf, a0, a1, a2, a3, a4, a5, a6);
     DpyRev(MSGX, MSGY, buf);
     MsgFlag = 1;
     DpyFlu();
}

ClrMsg()
 {	if (!MsgFlag) return;
	DpyNrm(MSGX, MSGY,
	       "                                                    ");
	MsgFlag = 0;
	DpyFlu();
 }

/* Read a character from the console:
 */

RdChar()
 {	int ch;
	DpyMov(INPX, INPY);
	DpyFlu();
	ch = getchar();
	ClrMsg();
	return ch & 0x7F;
 }

/* See if a character is waiting to be read:
 */

KbHit()
 {	int res;
	ioctl(0, FIONREAD, &res);
	return res;
 }

/* Get user input line (terminated by ENTER).
 * Control chars other than SPACE serve as breaks.
 */

char UsrBuf[100];

char *UserInput(prompt)
   char *prompt;
 {	int i, j, x, y;
	char tbuf[3];
	tbuf[1] = 0;

	UsrClr();
	DpyRev(INPX, INPY, prompt);
	y = INPY;
	x = INPX + 1 + strlen(prompt);

	for (i=0; j=RdChar(); ) switch(j)
	 { case '\r':
	   case '\n':	UsrBuf[i] = 0; return UsrBuf;
	   case '\b':
	   case 0177:	if (i) { DpyNrm(--x, y, " ");
				 i--;
			       }
			continue;
	   case 'U'-64:
	  		while (i) { DpyNrm(--x, y, " ");
				 i--;
			       }
			continue;

	   default:	if (j < ' ') {	UsrBuf[i++] = j;
					UsrBuf[i] = 0;
					tbuf[0] = '^';
					tbuf[1] = j + 64;
					tbuf[2] = 0;
					DpyNrm(x, y, tbuf);
					return UsrBuf;
				     }
			tbuf[0] = j;
			DpyNrm(x++, y, tbuf);
			UsrBuf[i++] = j;
			continue;
	 }
 }

UsrClr()
 {	DpyCLine(INPX, INPY);
 }

DpyCLine(x, y)
   int x, y;
#ifdef vt100
{	DpyMov(x,y);
	printf("\033[K");
	fflush(stdout);
}
#else
 {	DpyMov(x, y);
	printf("\033K");
	fflush(stdout);
 }
#endif
DpyClr()
#ifdef vt100
{	printf("\033[2J");
}
#else
 {	putchar(033); putchar('E');
 }
#endif

DpyNrm(x, y, text)		/* Display text, normal video.	*/
   int x, y;
   char *text;
{
     DpyMov(x, y);
     printf("%s", text);
}

DpyRev(x, y, text)		/* Display text, reverse video	*/
   int x, y;
   char *text;
#ifdef vt100
{	DpyMov(x, y);
	printf("\033[7m%s\033[m",text);
}
#else
 {	DpyMov(x, y);
	printf("\033p%s\033q", text);
 }
#endif

DpyFlu()			/* Flush output			*/
 {	fflush(stdout);
 }

DpyMov(x, y)
#ifdef vt100
 {	printf("\033[%d;%dH", y+1, x+1);
	fflush(stdout);
 }
#else
 {	putchar(033);	putchar('Y');
	putchar(y+32);	putchar(x+32);
 }
#endif

/* reset tty mode and exit */
split()
  {	tty.sg_flags &= ~CBREAK;
	tty.sg_flags |= ECHO;
	ioctl(1,TIOCSETP,&tty);
	exit(0);
}

Init()
 {
	ioctl(1,TIOCGETP,&tty);
	tty.sg_flags |= CBREAK;
	tty.sg_flags &= ~ECHO;
	ioctl(1,TIOCSETP,&tty);

	/* make sure we clean up!!! */
	signal(SIGHUP,split);
	signal(SIGINT,split);
	signal(SIGQUIT,split);
 }

