
#include <stdio.h>

#include "sim.h"

extern char watflag;
int Watching = 0, display_stack = 0;

initdata()
 {	DpyFlag = 1;
	BreakPt = -1;
 }

char *ctlname = "ctlrom", *uname = "stemp", *dramname = NULL;

Usage()
{
     fprintf(stderr, "Usage: sim [-c ctlname] [-u uname] [-d dram]\n");
     exit(1);
}

main(argc,argv)
   int argc;
   char **argv;
{
     char line[100];
     int i, j, ch, arg;
     
     initdata();
     arg = 1;
     while (arg < argc) {
	  if (argv[arg][0] == '-') {
	       switch (argv[arg][1]) {
	       case 'c':
		    if (argv[arg+1])
			 ctlname = argv[++arg];
		    else
			 Usage();
		    break;
	       case 'u':
		    if (argv[arg+1])
			 uname = argv[++arg];
		    else
			 Usage();
		    break;
	       case 'd':
		    if (argv[arg+1])
			 dramname = argv[++arg];
		    else
			 Usage();
		    break;
	       }
	       arg += 1;
	  }
	  else Usage();
     }
     
     /* Read in symbol table (file "symbols") */
     RdSym("symbols");
     
     /* read in rom contents */

     strcpy(line, ctlname); strcat(line, ".3");
     read_rom(line,ctl_rom_3);
     
     strcpy(line, ctlname); strcat(line, ".2");
     read_rom(line,ctl_rom_2);
     
     strcpy(line, ctlname); strcat(line, ".1");
     read_rom(line,ctl_rom_1); 

     strcpy(line, ctlname); strcat(line, ".0");
     read_rom(line,ctl_rom_0); 

     strcpy(line, uname); strcat(line, ".bin");
     read_rom(line,ucode_rom);
     
     strcpy(line, uname); strcat(line, ".sym");
     RdSym(line);
     
     mapopen(uname);
     
     Init();
     if (dramname) {
	  strcpy(line, dramname);
	  strcat(line, ".bin");
	  load_dram(line);
     }
     
     DpyInit();
     SymFlag = 1;
     Help();				/* Initial Help display.	*/
     SetLevel(2);
     
top:

     /* Takes a while - initializing large arrays  */
     Reset();
     GoFlag = 0;	
     
     /* command loop */
     while (1) {
	  if (GoFlag && !KbHit()) {
	       if (Level < 3) {
		    if (UStep()) goto bkpt;
	       }
	       else
		    end_minor_cycle();
	       Update();
	       continue;
	  }

	  ch = RdChar();
	  switch (ch) {
	  case 'p':
	       DumpProfileInfo();
	       Msg("Profile information dumped.");
	       break;

	  case 'T':
	       display_stack = !display_stack;
	       break;

	  case 'P':
	       ResetProfileInfo();
	       Msg("Profile information reset.");
	       break;
	       
	  case 'l':
	       SetLevel(Level+1);
	       continue;
	       
	  case 'b':
	       i = Level;	/* Save display mode. */
	       SetLevel(0);	/* No display.	*/
	       Msg("Running to breakpoint");
	       while (!(j=UStep()) && !KbHit());
	       SetLevel(i);

	       if (j) {
	       bkpt:
		    Msg("Breakpoint at 0x%x.", ucode_mar);
		    Update();
	       }
	       else
		    Msg("stopped at 0x%x", ucode_mar);
	       break;
	       
	  case 'W':
	       Watching ^= 1;
	       if (!Watching) {
		    Msg("Not watching SRAM locations.");
		    break;
	       }
	       
	  case 'w':
	       watflag=0;
	       watch();
	       break;
	       
	  case 'L'-64:
	       DpyClr(); 
	       DpyRef();
	       Update();
	       break;
	       
	  case 'g':
	       if (! GoFlag) {
		    GoFlag = 1;
		    Msg("running...");
		    break;
	       }

	  case 'G'-64:
	  HALT:
	       SetLevel(Level);
	       GoFlag = 0;
	       Msg("stopped...");
	       break;
	       
	  case 'R':
	       goto top;
	       
	  case '9': case '8':  case '7':  case '6':  case '5':
	  case '4': case '3':  case '2':  case '1':  case '0':
	       UPDPY(SwtDpy, switches ^= 1 << (ch - '0'));
	       if (GoFlag == 1) Msg("running...");
	       break;
	       
	  case 'B':
	       SetBpt();
	       break;
	       
	  case 'S':
	       SymFlag ^= 1; 
	       if (SymFlag) Source(ucode_mar);
	       else ShowSrc(0, 0, 0);
	       break;
	       
	  case ' ':
	       if (Level < 3) UStep();
	       else end_minor_cycle();
	       if (GoFlag) goto HALT;
	       Update();
	       break;
	       
	  case 's':
	       if (UStep())
		    goto bkpt;
	       Update();
	       break;

	  case 'c':
	       end_minor_cycle();
	       Update();
	       break;
	       
	  case 'r':
	       sram_display();
	       watflag=0;
	       break;
	       
	  case 'y':
	       dram_display();
	       watflag=0;
	       break;
	       
	  case 'u':
	       rom_display();
	       watflag=0;
	       break;
	       
	  case '?':
	       Help();
	       break;
	       
	  case 'D'-0100: case 3: case 'q': case (EOF & 0x7f):
	       split();
	       break;
	       
	  default:
	       if (ch == '\n') ch = 0;
	       Msg(" Illegal command character: '%c' (%d, 0x%x) ", ch, ch, ch);
	       break;
	  }	
     }
}


/*
 *  Print HELP messages...
 */

int HelpX, HelpY;

h(mm)/* NC */
   char *mm;
{
     DpyNrm(HelpX, HelpY++, mm);
}

Help()
{
     DpyClr();
     HelpX = 0;	HelpY = 0;
     h("MAYBE Simulator: Command Character Summary");
     HelpX = 5;  HelpY++;
     
     h("    l   Simulation level (toggle)");
     h("        (microinstr vs. cycle)");
     
     h("SPACE   single step");
     h("    s   single microinstruction");
     h("    c   single cycle");
     
     h("  8,9   P0, P1 toggle");
     h("  0-7   Switches S0-7 toggle");
     HelpY++;
     h("    g   Start/stop running");
     h("   ^G   Stop execution");
     HelpY++;
     h("    ?   Display HELP info");
     h("   ^L   Refresh display");
     h("   ^C   Quit (return to UNIX)");
     
     HelpX = 44; HelpY = 2;
     h("    r   Examine Memory (SRAM)");
     h("    y   Examine Memory (DRAM)");
     h("    u   Examine microcode ROM");
     HelpY++;
     h("    U   Microcode level off/on");
     h("    B   set breakpoint adr");
     h("    b   step until breakpoint");
     h("    S   symbolic tag printout off/on");
     h("    W   watch selected SRAM on/off");
     
     HelpY++;
     h("    R   reset to power-up state");
     
     HelpX=0;  HelpY = 23;
     h("PRESS ANY KEY to return to simulation.");
     getchar();
     AllDpy();
     DpyRef();
}

error(a0, a1, a2, a3, a4, a5, a6)
{
     fprintf(stderr, a0, a1, a2, a3, a4, a5, a6);
     fprintf(stderr, "\r\n");
     split();
}

/*
 * Set display level:
 */

SetLevel(ll)
   int ll;
{	
     switch(ll) {
     case 0:
	  Msg("No display");
	  DpyDisable();
	  DpyFlag=0;
	  break;

     case 1:
	  Msg("Switch display only");
	  DpyDisable();
	  DpyOn(SwtDpy);
	  break;
	  
     default:
	  ll = 2;

     case 2:
	  Msg("Microinstruction level");
	  DpyDisable();
	  DpyFlag=1;
	  DpyOn(UcntDpy);
	  DpyOn(AdrDpy); DpyOn(RBDpy); DpyOn(RADpy);
	  DpyOn(GR1Dpy); DpyOn(GR2Dpy);
	  DpyOn(SwtDpy);
	  DpyOn(SPoutDpy);
	  DpyOn(PCoutDpy);
	  break;
	  
     case 3:
	  Msg("Single-cycle level");
	  DpyEnable();
	  break;
     }
     Level = ll;
     Update();
}

/*
 * Top-level display update.
 */

Update()
{
     switch (Level) {
     case 3:
	  AllDpy();
	  break;
     case 2:
	  UpDpy(AdrDpy, (unsigned long) ucode_mar);
	  UpDpy(RBDpy, (unsigned long) RegB);
     case 1:
	  UpDpy(SwtDpy, (unsigned long) switches);
     case 0:
	  break;
     }
     if ((Level>1) && SymFlag) Source(ucode_mar);
     if ((Level>1) && Watching) watch();
     if ((Level>1) && display_stack) display_TOS();
}
