#include <stdio.h>
#include <stack.h>
#include "profile.h"
#include "sim.h"

extern char *prof_string;
extern struct Symbol *Tags;
static Stack usubr_stack;

ResetProfileInfo()
{
     /* Initialize profiling arrays */
     bzero((char *) uinst_prof, sizeof(int)*256);
     bzero((char *) sinst_prof, sizeof(struct _sinst_info)*256);
     bzero((char *) uinst_seq, sizeof(int)*256*256);
     bzero((char *) sinst_seq, sizeof(int)*256*256);
     bzero((char *) usubr_prof, sizeof(struct _usubr_call)*8192);

     /* XXX Memory leak!! But who cares? */
     usubr_stack = NewStack(sizeof(int));
}

/*
 * Remember that new_op is only valid when LdOp is active (low)
 */
#define DISPATCH_OPCODE 128
#define CALL_OPCODE 1
#define RTN_OPCODE 2
#define SRAM_R0 0
DoProfile(old_op, new_op)
   int old_op, new_op;
{
     unsigned char new_s_opcode;

     uinst_prof[old_op] += 1;

     if (! LdOp) {
	  uinst_seq[old_op][new_op] += 1;

	  if (new_op == CALL_OPCODE)
	       DoMicroCall();
	  else if (new_op == RTN_OPCODE)
	       DoMicroRtn();
	  else if (new_op == DISPATCH_OPCODE) {
	       new_s_opcode = (unsigned char) sram[SRAM_R0];
	       
	       sinst_prof[new_s_opcode].calls += 1;
	       if (old_s_time != -1)
		    sinst_prof[new_s_opcode].ticks += (Time - old_s_time);
	       old_s_time = Time;

	       if (old_s_opcode != -1)
		    sinst_seq[old_s_opcode][new_s_opcode] += 1;
	       old_s_opcode = new_s_opcode;
	  }
     }
}

DumpProfileInfo()
{
     FILE *f;
     int i, j;
     char c;

     c = *UserInput("Dump uinst_count.prof? [y]");
     if (c == 'y' || c == 'Y' || c == '\0') {
	  Msg("Dumping uinst_count.prof...");
	  f = fopen("uinst_count.prof", "w");
	  for (i=0; i < 256; i++)
	       if (uinst_prof[i])
		    fprintf(f, "0x%02x %d\n", i, uinst_prof[i]);
	  fclose(f);
	  Msg("");
     }
     
     c = *UserInput("Dump uinst_seq.prof? [y]");
     if (c == 'y' || c == 'Y' || c == '\0') {
 	  Msg("Dumping uinst_seq.prof...");
	  f = fopen("uinst_seq.prof", "w");
	  
	  for (i=0; i < 256; i++) {
	       fprintf(f, "0x%02x --", i);
	       for (j=0; j < 256; j++)
		    if (uinst_seq[i][j])
			 fprintf(f, " 0x%02x %4d,", j, uinst_seq[i][j]);
	       fprintf(f, "\n");
	  }
	  fclose(f);
	  Msg("");
     }
     
     c = *UserInput("Dump sinst_count.prof? [y]");
     if (c == 'y' || c == 'Y' || c == '\0') {
	  Msg("Dumping sinst_count.prof...");
	  f = fopen("sinst_count.prof", "w");
	  for (i=0; i < 256; i++)
	       if (sinst_prof[i].calls)
		    fprintf(f, "0x%02x %3d %5d %4.3f\n", i,
			    sinst_prof[i].calls, sinst_prof[i].ticks,
			    (double) sinst_prof[i].ticks / (double)
			    sinst_prof[i].calls);
	  fclose(f);
	  Msg("");
     }
     
     c = *UserInput("Dump sinst_seq.prof? [y]");
     if (c == 'y' || c == 'Y' || c == '\0') {
	  Msg("Dumping sinst_seq.prof...");
	  f = fopen("sinst_seq.prof", "w");
	  
	  for (i=0; i < 256; i++) {
	       fprintf(f, "0x%02x --", i);
	       for (j=0; j < 256; j++)
		    if (sinst_seq[i][j])
			 fprintf(f, " 0x%02x %4d,", j, sinst_seq[i][j]);
	       fprintf(f, "\n");
	  }
	  fclose(f);
	  Msg("");
     }

     c = *UserInput("Dump usubr_calls.prof? [y]");
     if (c == 'y' || c == 'Y' || c == '\0') {
	  Msg("Dumping usubr_calls.prof...");
	  f = fopen("usubr_calls.prof", "w");
	  for (i=0; i<8192; i++)
	       if (usubr_prof[i].calls)
		    fprintf(f, "%-12s 0x%04x %3d %6d  %4.3f\n",
			    LookVal(i, Tags), i, usubr_prof[i].calls,
			    usubr_prof[i].ticks, (double) usubr_prof[i].ticks
			    / (double) usubr_prof[i].calls);
	  fclose(f);
	  Msg("");
     }
}

DoMicroCall()
{
     int subr_addr;

     subr_addr = (ucode_rom[ucode_mar+4] << 8) + ucode_rom[ucode_mar+3];
     if (subr_addr > 8191) {
	  fprintf(stderr, "Uhh.. subr_addr = %d\n", subr_addr);
	  exit(1);
     }

     StackPush(usubr_stack, &subr_addr);
     StackPush(usubr_stack, &Time);
}

DoMicroRtn()
{
     int subr_addr, start_time;

     /* If profiling info is reset, we might hit rtn() instructions */
     /* for which the call info is no longer on the stack */
     if (StackEmpty(usubr_stack))
	  return;

     start_time = (int) (* (int *)StackPop(usubr_stack));
     subr_addr = (int) (* (int *)StackPop(usubr_stack));
     usubr_prof[subr_addr].calls += 1;
     usubr_prof[subr_addr].ticks += (Time - start_time);
}
