/* Various profiling hacks for sim */

/*
 * Microinstructions -- each element contains the number of clock
 * ticks spent in that microinstruction (= with that opcode).
 * Algorithm: nano_count[<OPCODE>] += 1 each clock tick
 */
int uinst_prof[256];

/*
 * Stack instructions -- each element contains the number of calls to
 * a given stack instruction and the total number of clock ticks spent
 * in that instruction.  Algorithm: each time a new stack instruction
 * is loaded (= each time dispatch is called)
 * sinst_prof[OLD_SINST].calls is incremented by 1 and
 * sinst_prof[OLD_SINST].ticks is incremented by the number of ticks
 * since the last dispatch.
 */
struct _sinst_info {
     int calls, ticks;
};
struct _sinst_info sinst_prof[256];
int old_s_opcode = -1, old_s_time = -1;

/*
 * Microsequences -- each element uinst[x][y] contains the number of
 * times opcode y is executed immediately following opcode x.
 * Algorithm: each time the LdOp signal is raised,
 * uinst_seq[OLD_OPCODE][NEW_OPCODE] is incremented.
 */
int uinst_seq[256][256];

/*
 * Stack sequences -- same as uinst_seq except the increment occurs
 * when dispatch is called (instead of LdOp).
 */
int sinst_seq[256][256];

/*
 * Microsubroutines
 */
struct _usubr_call {
     int calls, ticks;
} usubr_prof[8192];
