#include <stdio.h>

/* simulator for MAYBE computer 
 * modified for fall 1984 -- Liz Bradley 8/84
 * sim.c contains all the set up overhead,
 * and a few procedures which should probably live elsewhere.
 * Other files which are part of the simulator:
 * 
 *  	format.c	display and data structure format
 * 	display.c	display enable and disable - mid level stuff
 * 	display.h	display - low level procedures
 * 	symbols.c	symbol table
 * 	machdep.c	machine dependent stuff
 */

#include "sim.h"
#include "selector.h"

int	ucode_adr = 0;			/* adr of current uinstr	*/

/*
 * Generate a random 8-bit byte:
 */
ranb() /* NC */
{
     return (rand()>>6) & 0xFF;
}

/* Reset things to a pwr-up state.
 * Clears CLOCK, ADR, PHASE;
 * Places the instruction in location 0 of Ucode ROM into OPCODE register.
 * Randomizes contents of RAM, other registers.
 */

Reset()
{
     register i;	
     for (i=0; i<SRAMSIZE; i++) sram[i] = ranb();
     
     /*
      * This takes FOREVER! Sorry, -saw.
      *	for (ll=0; ll<DRAMSIZE; ll++)
      * WDRAM(ranb(), ll);
      */
     sram_mar = ranb();
     RegA = ranb();	RegB = ranb();
     Time = 0;	UIcount = 0;
     ucode_hi = 0;	ucode_lo = 0;
     ucode_adr = 0;
     opcode = ucode_rom[ ucode_mar ];
     DRAMreg = ranb();
     CCReg = ranb();	
     phase = 0;
     set_ctl();
     alu_out();
     AllDpy();

     ResetProfileInfo();
     InitRefreshState();

     beg_minor_cycle();
}

/* assemble inputs of control rom, then determine and decode its outputs */
set_ctl()
{
     /* Uses previous CCsel to determine CCreg bit -- note that you */
     /* will lose if you select CCsel == 7 when the previous cycle did */
     /* not set SP_carry */
     if (CCsel != 7)
	  ctlin = ((opcode & 0xFF) << 5) + ((phase & 0x0F) << 1) +
	       ((CCReg >> CCsel) & 0x01);
     else
	  ctlin = ((opcode & 0xFF) << 5) + ((phase & 0x0F) << 1) +
	       (SP_carry & 0x01);

     /* Display the CC bit that affects the *current* choice of ctlin */
     UPDPY(CCbitDpy, (CCsel == 7) ? SP_carry & 0x01 : (CCReg >> CCsel) &0x01);

     ctl_out_3 = ctl_rom_3[ctlin];
     ctl_out_2 = ctl_rom_2[ctlin];
     ctl_out_1 = ctl_rom_1[ctlin]; 
     ctl_out_0 = ctl_rom_0[ctlin];

     CCsel = (ctl_out_0) & 0x07;
     LdGR2 = (ctl_out_0 >> 3) & 0x01;
     LdGR1 = (ctl_out_0 >> 4) & 0x01;
     Latch = (ctl_out_0 >> 5) & 0x01;
     NRefr = (ctl_out_0 >> 6) & 0x01;
     inca = (ctl_out_0 >> 7) & 0x01;
     sptr = (ctl_out_1) & 0x3F;
     alu = (ctl_out_2) & 0x3F;
     LdOp = (ctl_out_2 >> 6) & 0x01;
     src = (ctl_out_3) & 0x0F;
     dst = (ctl_out_3 >> 4) & 0x0F;

     UPDPY(CtlDpy, ctl_out_32);
}

int bflag, bus;

/*
 * Begin a minor cycle: cause data to propagate thru control ROM, etc.
 * May be called multiple times each cycle (if changes are made to inputs).
 * One call corresponds to one PHASE (nano)cycle.
 */

beg_minor_cycle()
{
     int cc_set = 0;
     
     /* calculate control signals */
     set_ctl();

     /* So we can produce a warning if it is used improperly */
     SP_carry = -1;
     
     /* compute bus drive and contents, set bflag to indicate bus drive */
     bflag = 1;
     switch (src) {
     case DSWT:
	  bus = switches & 0xFF;
	  break;
     case DUROM:
	  bus = ucode_rom[ucode_mar];
	  break;
     case DALU:
	  alu_out();
	  bus = ALU_data;
	  cc_set = 1;
	  break;
     case DSP:
	  sptr_out();
	  bus = SP_data;
	  SP_carry = ALU_c_out;
	  break;
     case DSRAM:
	  bus = sram[sram_mar];
	  break;
     case DDRAM:
	  bus = DRAMreg;
	  break;
     case DPCL:
	  bus = PClo;
	  break;
     case DPCH:
	  bus = PChi;
	  break;
     case DGR1:
	  bus = GR1;
	  break;
     case DGR2:
	  bus = GR2;
	  break;
     case DIO:			/* Punt on I/O stuff */
	  bus = 0;
	  break;
     default:
	  bflag = 0;
	  Msg("Unknown DRSEL value: %d.", src);
	  break;
     }

     /* If the ALU is not driven, the CC bits must be set anyway. */
     /* This is so things like A <- SRAM and a Latch CC on the next */
     /* cycle work. */
     if (! cc_set)
	  alu_out();

     /* Refresh */
     if (! NRefr)
	  RefreshRow();

     UPDPY(BusDpy, bus &= 0xFF);  
     UPDPY(MarDpy, sram_mar);
     UPDPY(AdrDpy, ucode_mar);
     UPDPY(URomDpy, ucode_rom[ucode_mar]);
}
   
/*
 * simulate the end of a minor cycle of micromachine: Causes active
 * (trailing) clock edge, loads registers.  Increments PHASE to start
 * new nanocycle.  Updates time and cycle counts.
 */
   
end_minor_cycle()
{
     /* PROF: Do uinst profiling before the opcode changes */
     DoProfile(opcode, ucode_rom[ucode_mar] & 0xFF);
     
     /* Load things that load off the bus */
     if (! bflag)
	  Msg("Bus not driven -- cannot load %d.", dst);
     else {
	  switch (dst) {
	  case LADR:
	       ucode_hi = ucode_lo;
	       ucode_lo = bus;
	       UPDPY(AdrDpy, ucode_mar);
	       break;
	  case LA:
	       UPDPY(RADpy, RegA = bus);
	       break;
	  case LB:
	       UPDPY(RBDpy, RegB = bus);
	       break;
	  case LSRAM:
	       sram[sram_mar] = bus;
	       break;
	  case LMAR:
	       UPDPY(MarDpy, sram_mar = bus);
	       break;
	  case LDRAM:
	       w_dram(bus);
	       break;
	  case LPCL:
	       PClo = bus;
	       break;
	  case LPCH:
	       PChi = bus;
	       break;
	  case LSPa:
	       SPa = bus;
	       break;
	  case LSPb:
	       SPb = bus;
	       break;
	  case LSPa_DRAM:
	       SPa = bus;
	       w_dram(bus);
	       break;
	  case LSPb_DRAM:
	       SPb = bus;
	       w_dram(bus);
	       break;
	  case LIOADR:
	  case LIO:
	  case LNOOP:
	       break;
	  default:
	       Msg("Unknown LDSEL value: %d.", dst);
	       break;
	  }

	  if (LdGR1)
	       GR1 = bus;
	  if (LdGR2)
	       GR2 = bus;
     }

     /* Now do things that don't load off the bus */

     /* Phase counter */
     phase += 1;
     if (phase == 0x0F)
	  pRCO = 1;
     else
	  pRCO = 0;
     phase &= 0x0F;
          
     /* CC register -- Latch is active low */
     if (! Latch)
	  CCReg = (pRCO << 7) + ((switches >> 4) & 0x30) + (ALU_flags & 0x0F);
     
     /* Program counter updates at end of tick, if driven */
     if (src == DPCL) {
	  PClo += 1;
	  if (PClo > 255) {
	       PCcarry = 1;
	       PClo &= 0xFF;
	  }
     }
     else if (src == DPCH) {
	  PChi += PCcarry;
	  PCcarry = 0;
     }
     else
	  PCcarry = 0;
     
     /* if LdOp asserted clear phase, otherwise just increment phase */
     /* -- LdOp is active low. */
     if (LdOp)
	  UPDPY(PhDpy, phase);
     else {
	  if (dst == LADR)
	       Msg("Warning!  Loading ADR and OPCODE in same cycle!");
	  
	  ucode_adr = ucode_mar;
	  UPDPY(OpDpy, opcode = (ucode_rom[ucode_mar]) & 0xFF);
	  UPDPY(PhDpy, phase = 0);
	  UPDPY(UcntDpy, UIcount+=1);        /* microinstruction count */
     }

     /* ADR+ */
     if (inca) {
	  if (dst == LADR)
	       Msg("Warning!  Loading and incrementing uADR!");
	  if (!(ucode_lo = 0xFF & (1+ucode_lo)))
	       ucode_hi++;
     }
     
     /* Address changes AFTER LdOp loads off the bus - trailing edge magic*/
     UPDPY(AdrDpy, ucode_mar);
     UPDPY(CCDpy, CCReg);
     UPDPY(TimDpy, ++Time);
     UPDPY(SPoutDpy, (SPb << 8) + SPa);
     UPDPY(PCoutDpy, (PChi << 8) + PClo);
     beg_minor_cycle();
}

/* read a file into a rom array */
read_rom(f,array)/* OK 9/3 */
   char *f;
   unsigned char *array;
{
     int fid,i;

     /* default contents of rom is zero */
     for (i = 0; i < ROMSIZE; i += 1)
	  array[i] = 0;
     
     if ((fid = open(f,0)) < 0) {
	  fprintf(stderr,"Cannot open %s for input\n",f);
	  exit(-1);
     }
     if (read (fid, array, 8192) != 8192) {
	  fprintf(stderr,"File %s incorrect size\n",f);
	  close(fid);
	  exit(-1);
     }
     
     close(fid);
}

dram_display()
{
     mem_display("DRAM", NULL, (unsigned long) DRAMSIZE,
		 (unsigned int) (DRAM_addr && 0xFFFF));
}

sram_display()
{
     mem_display("SRAM", sram, (unsigned long) SRAMSIZE,
		 (unsigned int) sram_mar & 0xFF);
}

rom_display()
{
     mem_display("ROM", ucode_rom, (unsigned long) ROMSIZE,
		 (unsigned int) (ucode_mar & 0xFF)<<3);
}

/*
 * Dirty hack: if array is NULL, then dram is used.
 * This is to make the IBM PC (and other small adr-space widgets) work.
 */

mem_display(name, array, size, dot)
   char *name, *array;
   unsigned long size;
   unsigned int dot;
{	char *inp, buf[100], *aa, oflag = 0;
	int adr = dot;
	int x=5, value, vflag;

	WindClr();
	/* So inp isn't "used before set" */
	inp = NULL;
	goto rdadr;

	/* Here with new address -- open a location.
	 */
newadr:	adr &= (size-1);		/* ASSUMES size is 2**n	*/
	WindNext();			/* Dont scroll iff same locn */
	oflag = 1;			/* Now theres an open location */

reshow:	sprintf(buf, "%s %s (%d.): ", name, bin(adr,16), adr);
	aa = buf + strlen(buf);
	value = 0xFF & (array? array[adr] : RDRAM(adr));
	sprintf(aa, "%s (%d.) ", bin(value,8), value);
	if (oflag) DpyRev(x, WindY, buf);
	else DpyNrm(x, WindY, buf);

	if (oflag) goto rdcont;		/* Read new contents. */

disp:	switch (*inp)			/* else, interpret command char */
	 {
	   case 'N'-64:	adr++; goto newadr;
	   case 'P'-64:	adr--; goto newadr;

	   case 'X'-64:
	   case 'x':	UsrClr(); return;

	   case 0:	oflag = 0;
	   default:	break;
	 }

rdadr:	sprintf(buf, "Type %s address (expression or ^X, ^N, ^P): ", name);
	inp = UserValue(buf, &value, &vflag);
	if (vflag < 0) goto rdadr;
	if (vflag > 0) { adr = value; goto newadr; }
	goto disp;

rdcont:	inp = UserValue(" Type new contents (expression or ^X, ^N, ^P): ",
			&value,	&vflag);
	if (vflag < 0) goto rdcont;

	DpyNrm(x, WindY, buf);		/* Show its not open anymore.	*/
	oflag = 0;

	if (vflag > 0)
	 {
	   if (array) array[adr] = value;
	   else WDRAM(adr, value);
	   goto reshow;
	 }

	goto disp;
 }


/*
 * Update all displayed values:
 */

AllDpy()
{
     UPDPY(RADpy, RegA);
     UPDPY(RBDpy, RegB);
     UPDPY(GR1Dpy, GR1);
     UPDPY(GR2Dpy, GR2);
     UPDPY(CCDpy, CCReg);
     UPDPY(OpDpy, opcode);
     UPDPY(AdrDpy, ucode_mar);
     UPDPY(URomDpy, ucode_rom[ucode_mar]);
     UPDPY(PhDpy, phase);
     UPDPY(BusDpy, bus);
     UPDPY(MarDpy, sram_mar);
     UPDPY(DregDpy, DRAMreg);
     UPDPY(SwtDpy, switches);
     UPDPY(TimDpy, Time);
     UPDPY(UcntDpy, UIcount);
     ctl_out_3 = ctl_rom_3[ctlin];
     ctl_out_2 = ctl_rom_2[ctlin];
     ctl_out_1 = ctl_rom_1[ctlin]; 
     ctl_out_0 = ctl_rom_0[ctlin];
     UPDPY(CtlDpy, ctl_out_32);
} 

/*
 * Microcode-level debugging aids:
 * Step to end of current microinstruction:
 * Returns 1 iff breakpoint reached (whence clears GoFlag)
 */
 
int UStep()
{
     do {
	  end_minor_cycle();  /* "end" calls "beg" minor cycle */
     } while (LdOp);   /* LdOp is active low */
     
     if ((ucode_mar == BreakPt) || ((opcode & 0xFF) == 0xFF)) {
	  GoFlag = 0;
	  return 1;
     }
     return 0;
}

/* Set a breakpoint:
 */

SetBpt()/* NC */
{
     char *inp = UserInput("Set breakpoint at: ");
     int flag, temp;

     if (*inp > ' ')  {
	  temp = Expr(&inp, &flag);
	  if (flag == 1) {
	       BreakPt = temp;
	       Msg("Breakpoint set at 0x%x", temp);
	       return;
	  }
     }
     Msg("Breakpoint cleared.");
     BreakPt = -1;
}

alu_out()
{
     simulate_alu(alu, &ALU_data, RegA, RegB);
     UPDPY(ALUoutDpy, ALU_flags);
}

sptr_out()
{
     simulate_alu(sptr, &SP_data, SPa, SPb);
     UPDPY(ALUoutDpy, ALU_flags);
}
