/* Symbol-table stuff for MAYBE simulator
 */

#include <stdio.h>
#include <ctype.h>
#include "sim.h"

#define isletter(CC) (isalpha(CC) || (CC) == '_' || (CC) == '.')

long RdNum();

struct Symbol
 {	struct Symbol *Next;		/* List-linked			*/
	char *Name;
	long Value;
	char type;			/* Where it came from.		*/
 } *Tags, *Ops;

/* Add a symbol to a symbol table:
 */

DefSym(sym, val, symtab)
   char *sym;
   struct Symbol **symtab;
{
     register struct Symbol *ss = (struct Symbol *) malloc(sizeof *ss);
     char *name = (char *) malloc(1+strlen(sym));

     if (ss == NULL || name == NULL) {
	  fprintf(stderr, "sim: Cannot add new symbol!\n");
	  exit(-1);
     }
     strcpy(name, sym);
     ss->Name = name;
     ss->Value = val;
     ss->Next = *symtab;
     *symtab = ss;
}

/*
 * Read a symbol table:
 */

int RdSym(filename)
   char *filename;
{
     FILE *fid;
     char buf[100], aval[40], code, name[30];
     register struct Symbol **symtab;

     if ((fid = fopen(filename, "r")) == NULL)
	  return 0;
     
     while (!feof(fid)) {
	  fgets(buf, 100, fid);
	  if (sscanf(buf, "%s %c %s", aval, &code, name) != 3)
	       continue;
	  switch(code) {
	  case ':':
	       symtab = &Tags;
	       break;
	  case '=':
	       symtab = &Ops;
	       break;
	  default:
	       continue;
	  }
	  DefSym(name, atob(aval), symtab);
     }

     return 1;
}


/* Look up a symbol, by value: returns pointer, or NULL.
 */

char *LookVal(val, symtab)
 register struct Symbol *symtab;
 {
	while (symtab)
	 { if (val == symtab->Value)
		return symtab->Name;
	   symtab = symtab->Next;
	 }
	return 0;
 }

/* Look up a tag, by value: returns pointer to char string output.
 * if max>0, specifies maximum offset tolerated.
 */

extern char SymFlag;

char *LookTag(val, max)
 {	register struct Symbol *symtab = Tags;
	char *best;
	static char buf[50];
	int off = 1000000, i;
	val &= 0xFF;

	if (SymFlag) for (; symtab; symtab = symtab->Next)
	 { if ((i = val - symtab->Value) < 0) continue;
	   if ((i < off) && (i <= max))
		{ off = i;
		  best = symtab->Name;
		}
	 }
	if (off == 1000000)
	 { sprintf(buf, "%d.", val);
	 }
	else
	 { if (off) sprintf(buf, "%s+%d.", best, off);
	   else sprintf(buf, "%s", best);
	 }
	return buf;
 }

/* Lookup a symbol, by name; returns value.
 * Looks first in tags, then in Ops.
 * Returns -1 iff none.
 */

LookSym1(name, symtab)
 char *name;
 register struct Symbol *symtab;
 {	while (symtab)
	 { if (!strcmp(name, symtab->Name))
		return (symtab->Value) & 0xFFFF;
	   symtab = symtab->Next;
	 }
	return -1;
 }

LookSym(name)
     char *name;
 {	register i;
	if ((i = LookSym1(name, Tags)) == -1) return LookSym1(name, Ops);
	return i;
 }

char *Op(val)
 {	register char *cc;
	static char buf[20];
	if (cc = LookVal(val, Ops)) return cc;
	sprintf(buf, "%d.", val);
	return buf;
 }

/* Print a symbolic microinstruction:
 */

/* ARGSUSED */
PrUinst(x, y, adr)
{
}


/*
 * Hack to watch all SRAM locations whose names begin with dot.
 * Modified (4/4/84) to watch bank 1 locations whose name begins with ..
 */

SymWatch()
 {	register struct Symbol *symtab = Ops;	/* = defined symbols only */

	while (symtab)
	 { if (((symtab->Name)[0] == '.') &&
		(symtab->Name)[1])
	     { if (symtab->Name[1] != '.')
		       wat4(symtab->Value, symtab->Name);
	       else
		       wat4(symtab->Value+256, symtab->Name);
	     }
	   symtab = symtab->Next;
	 }
 }

/*
 * Read a value from the user:
 * returns pointer to break char.
 * *flag set as for Expr, below.
 */

char *UserValue(prompt, value, flag)
   char *prompt;
   int  *value;
   int  *flag;
{
     char *inp;

     inp = (char *) UserInput(prompt);
     *flag = 0;
     *value = Expr(&inp, flag);
     return inp;
}

/*
 * Read an expression: updates **ss.
 * If anything read, sets *flag to 1.
 * On error, sets *flag to -1.
 */
Expr(ss, flag)
   register char **ss;
   int *flag;
{
     register Value, i;

     /* Ignore leading white */
     while (*ss && (**ss == ' '))
	  (*ss)++;			/* Ignore leading white	*/
     if (**ss < ' ')
	  return 0;			/* Nothing to read.	*/
     
     if (**ss == '-') {			/* Unary minus		*/
	  Value = 0;
	  goto binop;
     }

     if (isdigit(**ss)) {
	  Value = RdNum(ss, flag);
	  if (*flag < 1) return 0;
     }       
     else if (isletter(**ss)) {
	  char buf[100];
	  for (i=0; isletter((*ss)[i]) || isdigit((*ss)[i]); i++)
	       buf[i] = (*ss)[i];
	  buf[i] = 0;
	  if ((Value = LookSym(buf)) == -1) {
	       Msg("Undefined symbol: %s", buf);
	       return *flag = -1;
	  }
	  *ss += i;
	  *flag = 1;
     }
     else {
	  Msg("Bad expression syntax: %s", *ss);
	  return *flag = -1;
     }
     
binop:

     while (**ss && (**ss == ' '))
	  (*ss)++;			/* Ignore leading white	*/
     if (**ss < ' ')
	  return Value;			/* Nothing to read.	*/
     
     switch (**ss) {
     case '+':
	  (*ss)++;
	  if ((i = Expr(ss, flag)) == -1)
	       return *flag = -1;
	  if (!*flag) *flag = 1;
	  return Value+i;
     case '-':
	  (*ss)++;
	  if ((i = Expr(ss, flag)) == -1)
	       return *flag = -1;
	  if (!*flag) *flag = 1;
	  return Value-i;
     case 0:
	  return Value;
     default:
	  Msg("Bad operator in expression: %s", *ss);
	  return *flag = -1;
     }
     
}

long RdNum(ss, flag)
 register char **ss;
 int *flag;
 {	long value = 0;
	int i;

	if (**ss == '0') switch((*ss)[1])
	 { case 'x':	*ss += 2;
			while ((i = hexdigit(**ss)) >= 0)
			 { value <<= 4;
			   value += i;
			   (*flag) = 1;
			   (*ss)++;
			 }
			return value;
	   case 'b':	*ss += 2;
			while (((i = **ss) == '0') || (i == '1'))
			 { value += value;
			   value += i-'0';
			   (*flag) = 1;
			   (*ss)++;
			 }
			return value;
	   case 'd':	*ss += 2;
	   default:	break;

	 }

	while (((i = **ss) >= '0') && (i <= '9'))
	 { value *= 10;
	   value += i-'0';
	   (*flag) = 1;
	   (*ss)++;
	 }
	return value;
 }


/* Convert binary/decimal string to integer.
 * Binary unless trailing ".", whence decimal.
 */

int atob(s)
  register char *s;
  {	register char *p;
	register int i;

	while (*s<' ' && *s!=0) s += 1;

	for (p=s; *p; p++) if (*p == '.') return atoi(s);

	/* nothing but 0's and 1's -- treat as binary number */
	for (i = 0; *s; s += 1) i = (i << 1) + *s - '0';
	return(i);
}

hexdigit(ch)
   register char ch;
{
     if ((ch >= 'a') && (ch <= 'f')) return ch-'a'+10;
     if ((ch >= 'A') && (ch <= 'F')) return ch-'A'+10;
     if ((ch >= '0') && (ch <= '9')) return ch-'0';
     return -1;
} 
