#include <stdio.h>

/* pattern = contents */

#define ROMSIZE 8192	/* size of 2764 */
#define ADDRSIZE 13	/* log ROMSIZE = # of address bits */
#define DATASIZE 32	/* number of data bits */
#define LSIZE 500	/* max number of chars in input line */

long rom[ROMSIZE];	/* our internal copy of the 2764 contents */
int Line[ROMSIZE];	/* keep track of what input line set location */
char set[ROMSIZE+1];	/* indicates if locn should be programmed */
#define NOTSET 0
#define DEFAULT 1
#define SET 2

/* options and flags */
int debug = 0;		/* <>0 if debugging printout wanted */
int listing = 0;	/* <>0 if listing file should be produced */
char *ifile = NULL;	/* name of input file */
FILE *f;		/* handy file descriptor */

int lineno = 0;		/* input line number */
char line[LSIZE];	/* buffer for input line */
char *lptr;		/* pointer to current position on input line */
char *error;		/* pointer to error message, if any */

char addr[ADDRSIZE];	/* address pattern */
char contents[DATASIZE];  /* contents specification */
int Default;		/* <>0 if current line can be overridden later */

/* XXX used to be char */
short vars[128];		/* current values of pattern variables */

/* move lptr to next non-blank character */
skipnb()
{
     while (*lptr==' ' || *lptr=='\t' || *lptr=='\r')
	  lptr += 1;
     if (*lptr == '|')
	  while (*lptr != '\n')
	       lptr += 1;
}

/* read in pattern from input line, terminated by newline, comment, or = */
getpattern(array, size)
   char *array;
   int size;
{
     register int i, temp;
     
     i = size-1;
     while (1) {
	  /* get the next pattern element, always a single char */
	  skipnb();
	  if (*lptr=='|' || *lptr=='=' || *lptr=='\n') break;
	  
	  /* save element in array */
	  if (i < 0) {
	       error = "pattern has too many elements";
	       return;
	  }
	  array[i--] = *lptr;
	  vars[*lptr++] = -1;  /* no value for associated variable yet */
     }
     
     /*
      * make last element we read in end up in array[0], fill unspecified
      * part of pattern with "0"
      */
     for (temp = 0; temp < size-i-1; temp += 1)
	  array[temp] = array[temp+i+1];
     while (temp < size) array[temp++] = '0';
}

/* determine value of contents pattern */
value()
{
     register int i, v, element;
     
     for (i = DATASIZE, v = 0; i > 0; i -= 1) {
	  element = contents[i-1];
	  v <<= 1;
	  if (element == '0');
	  else if (element == '1') v += 1;
	  else if (element == '*') {
	       error = "* cannot appear in contents field";
	       goto err;
	  }
	  else if (vars[element] == -1) {
	       error = "variable in contents field did not appear in address";
	       goto err;
	  }
	  else v += vars[element];
     }
     return(v);
     
err:	fprintf(stderr,"*** error in %s, line %d: %s\n%s",
		ifile,lineno,error,line);
     return(0);
}

/* recursive routine to generate addresses that match address pattern */
match(index,acc)
   int index,acc;
{
     int element;
     
     if (index == 0) {
	  /* no more pattern, fill in location */
	  if (!Default && set[acc]==SET) {
	       fprintf(stderr,"*** error in %s, line %d: location %d filled more than once\n%s",
		       ifile,lineno,acc,line);
	  }
	  else {
	       rom[acc] = value();
	       set[acc] = Default ? DEFAULT : SET;
	       Line[acc] = lineno;
	       if (debug && !Default) {
		    fprintf(stderr,"  rom["); binout(stderr,acc,ADDRSIZE);
		    fprintf(stderr,"] = "); binout(stderr,rom[acc],DATASIZE);
		    if (Default) fprintf(stderr," (default)");
		    putc('\n',stderr);
	       }
	  }
	  return;
     }
     
     acc <<= 1;		/* get set to add another bit to accumulator */
     element = addr[index-1];
     index -= 1;
     if (element == '0') match(index,acc);
     else if (element == '1') match(index,acc+1);
     else if (element == '*') {
	  match(index,acc);	/* "*" matches both 0 and 1 */
	  match(index,acc+1);
     }
     else if (vars[element] != -1) match(index,acc+vars[element]);
     else {			/* new variable, try both values */
	  vars[element] = 0;
	  match(index,acc);
	  vars[element] = 1;
	  match(index,acc+1);
	  vars[element] = -1;
     }
}

/* output binary number to specified file */
binout(ofile,n,digits)
   FILE *ofile;
   int n,digits;
{
     register unsigned i;
     
     for (i = 1 << (digits-1); i > 0; i >>= 1)
	  putc((n&i) ? '1' : '0',ofile);
}

main(argc,argv)
   char **argv;
{
     register int i;
     int parse_ok = 1;
     char temp[100];

     printf("xprom: %d bit version, compiled by %s on %s.\n", DATASIZE,
	    WHO, DATE);
     
     /* process command line */
     for (i = 1; i < argc; i++) {
	  if (argv[i][0] == '-') {
	       switch (argv[i][1]) {

	       case 'l':
		    listing = 1;
		    break;

	       case 'd':
		    debug = 1;
		    break;
		    
	       default:
		    fprintf(stderr,"Unrecognized option: %s\n", argv[i]);
		    break;
	       }
	  }
	  else
	       if (ifile == NULL)
		    ifile = argv[i];
	       else
		    fprintf(stderr,"Too many files specified: %s\n", argv[i]);
     }
     
     /* make sure user specified all the filenames, etc. */
     if (ifile == NULL) {
	  fprintf(stderr,"No input file specified\n");
	  exit(-1);
     }
     else {
	  strcpy(temp,ifile);
	  strcat(temp,".prom");
	  if ((f = fopen(temp,"r")) == NULL) {
	       fprintf(stderr,"Cannot open %s for input\n",temp);
	       exit(-1);
	  }
     }
     
     /* user hasn't programmed any locations yet */
     for (i = 0; i < ROMSIZE; i++) {
	  set[i] = NOTSET;
	  rom[i] = -1;
     }
     
     /* process input, line by line */
     lineno = 0;
     while ((lptr = fgets(line,LSIZE,f)) != NULL) {
	  if (strlen(lptr) >= LSIZE) {
	       fprintf(stderr,"Input line too long, ignored...\n");
	       continue;
	  }
	  
	  error = NULL;			/* no errors yet! */
	  lineno += 1;
	  
	  skipnb();			/* skip to first non-blank char */
	  if (*lptr == '\n') continue;	/* blank or commment lines are okay */
	  getpattern(addr,ADDRSIZE);	/* read in address */
	  if (error != NULL) goto err;
	  
	  /*
	   * address is terminated by at least one "=".  A second "="
	   * indicates that this line is a default, ie, rom locations
	   * set by this line can be overridden by other specifications.
	   */
	  if (*lptr == '=') {
	       if (lptr[1] == '=') {
		    Default = 1; lptr += 2;
	       }
	       else {
		    Default = 0; lptr += 1;
	       }
	  }
	  else {
	       error = "= missing after address pattern";
	       goto err;
	  }
	  
	  getpattern(contents,DATASIZE); /* read in contents */
	  if (error != NULL) goto err;
	  
	  /* make sure input line contains only two fields */
	  skipnb();			/* skip to what had better be end */
	  if (*lptr != '\n') {
	       error = "Too many fields in input line";
	       goto err;
	  }
	  
	  if (debug) {
	       fprintf(stderr,"line %d: addr=",lineno);
	       for (i = ADDRSIZE; i > 0; i--) putc(addr[i-1],stderr);
	       fprintf(stderr," contents=");
	       for (i = DATASIZE; i > 0; i--) putc(contents[i-1],stderr);
	       putc('\n',stderr);
	  }
	  
	  /*
	   * find all the locations that match address pattern and fill them
	   * with the appropriate contents.
	   */
	  match(ADDRSIZE, 0);
	  continue;
	  
     err:
	  if (error != NULL) {
	       fprintf(stderr,"*** error in %s, line %d: %s\n%s",
		       ifile,lineno,error,line);
	       parse_ok = 0;
	  }
     }
     
     /* split now if input had some errors */
     if (!parse_ok) {
	  fprintf(stderr,"Input file had errors, exiting...\n");
	  exit(-1);
     } else fclose(f);	/* close input file */
     
     /* output a listing if requested to do so */
     if (listing) {
	  strcpy(temp,ifile);
	  strcat(temp,".list");
	  if ((f = fopen(temp,"w")) == NULL) {
	       fprintf(stderr,"Cannot open %s for output\n",temp);
	       exit(-1);
	  }
	  /*	  time(&t);	
	   *	  fprintf(f,"| This file compiled from %s at %s\n",ifile,ctime(&t));
	   */	  for (i = 0; i < ROMSIZE; i++) if (set[i] != NOTSET) {
		binout(f,i,ADDRSIZE); putc(' ',f);
		binout(f,rom[i],DATASIZE);
		fprintf(f,"  | %s by line %d\n",
			(set[i]==SET) ? "set" : "default",Line[i]);
	   }
	  fclose(f);
     }

     write_binfile(ifile, "0", 0);
     write_binfile(ifile, "1", 8);
     write_binfile(ifile, "2", 16);
     write_binfile(ifile, "3", 24);
     return(0);
}

write_binfile(name, ext, shift)
   char *name, *ext;
   int shift;
{
     FILE *f;
     int i;
     char temp[BUFSIZ];
     unsigned char buffer[ROMSIZE];

     sprintf(temp, "%s.%s", name, ext);
     f = fopen(temp, "w");
     if (f == NULL) {
	  fprintf(stderr, "xprom: Cannot open %s for output.\n", temp);
	  exit(1);
     }

     for (i = 0; i < ROMSIZE; i++)
	  buffer[i] = rom[i] >> shift;

     if (fwrite(buffer, 1, ROMSIZE, f) != ROMSIZE) {
	  fprintf(stderr, "xprom: Error writing %s.\n", temp);
	  exit(1);
     }

     fclose(f);
}

     
     
