/* ROM ucode for 6.111 lab #4 */
/* $Id: lab4.as,v 1.7 91/11/09 13:53:54 bjaspan Exp Locker: bjaspan $ */
/* $Source: /afs/athena.mit.edu/user/b/bjaspan/classes/6111/lab4/RCS/lab4.as,v $ */

# SPEC_FILE = lab4.sp;
# LIST_FILE = lab4.lst;
# MASK_COUNT = 8;
# SET_ADDRESS = 0;
# LOAD_ADDRESS = 0;

/* This must be first. */
start:
	JMP wait_loop;

/*************************************************************************/

/* Wait for the beginning of a block (a colon followed by an addr of */
/* zero) and then jump into read_loop to read it in. */
wait_loop:
	/* wait for a colon */
	JSB read_one_char;
	DO immed %h3a LDCMP;
	DO DRCHAR;
	if /EQUAL CJMP wait_loop;

	/* Get the # of data bytes, store it in the downcounter, prime
	   the checksum */
	JSB clear_checksum;
	JSB read_char;
	LDCT DRCHAR;
	
	/* If the starting address is zero, continue reading the
	   block, otherwise abort. */
	DO immed %h00 LDCMP;
	JSB read_char;
	DO DRCHAR;
	if /EQUAL CJMP wait_loop;
	
	JSB read_char;
	DO DRCHAR;
	if /EQUAL CJMP wait_loop;

	/* Beginning of block!  Read it in */
	DO CLRMAR;
	JMP jump_from_wait_loop;

/*************************************************************************/

/* Read in a block of data, jumping to wait_loop if the block contains */
/* an error.  Jump to play_song when done. */

read_loop:
	/* First char should be a colon */
	DO immed %h3a LDCMP;
	JSB read_one_char;
	DO DRCHAR;
	if /EQUAL CJMP read_loop_not_colon;

	/* Get the # of data bytes, store it in a downcounter, prime */
        /* the checksum  */
	JSB clear_checksum;
	JSB read_char;
	LDCT DRCHAR;

	/* Get the starting address and add to checksum.  Note that we */
	/* do *not* need to alter the MAR. */
	JSB read_char;
	JSB read_char;

jump_from_wait_loop:
	/* Check the AUX byte, jump to handle_aux if 1. */
	JSB read_char;
	DO immed %h01 LDCMP;
	DO DRCHAR;
	if /EQUAL CJMP read_data;
	JMP aux_read_data;

	/* Read the correct # of bytes, testing for 0 first */
read_data:
	JMP read_done_p;

read_data_loop:
	JSB read_char;
	DO DRCHAR LDSRAM;
	DO DRCHAR LDCMP;
	DO DRSRAM;
	if /EQUAL CJMP ldsram_failed;
	DO MARCNT;
read_done_p:
	RPCT read_data_loop;

	/* read the checksum and check it */
	JSB read_char;
	if /CKSUM_ZERO CJMP checksum_failed;

	/* ignore the CR/LF */
	JSB read_char;
	JSB read_char;

	/* read the next line in the block */
	JMP read_loop;

checksum_failed:
	JSB dump_memory;
	JMP wait_loop;


/*************************************************************************/

/* Duplicate of some code in read_loop to avoid an extra register bit. */
/* Checks the line containing an aux byte of 1, and then plays the song. */

	/* there should be 0 bytes, but cksum them anyway */
aux_read_data:
	JMP aux_read_done_p;

aux_read_data_loop:
	JSB read_char;
	DO DRCHAR LDSRAM MARCNT;
aux_read_done_p:
	RPCT aux_read_data_loop;

handle_aux_1:
	/* read the checksum and check it */
	JSB read_char;
	if /CKSUM_ZERO CJMP aux_checksum_failed;

	/* The data is correct, so see if this song is selected */
	JMP play_song;

aux_checksum_failed:
	JSB dump_memory;
	JMP wait_loop;

/*************************************************************************/

/* Check if the right song is selected, play it if so. */

play_song:

	/* Test to see if the byte in SRAM zero is a legal song number */
	DO CLRMAR;

test_song_0:
	DO immed %h00 LDCMP;
	DO DRSRAM;
	if /EQUAL CJMP test_song_1;
	JMP legal_song;

test_song_1:
	DO immed %h01 LDCMP;
	DO DRSRAM;
	if /EQUAL CJMP test_song_2;
	JMP legal_song;

test_song_2:
	DO immed %h02 LDCMP;
	DO DRSRAM;
	if /EQUAL CJMP test_song_3;
	JMP legal_song;

test_song_3:
	DO immed %h03 LDCMP;
	DO DRSRAM;
	if /EQUAL CJMP illegal_song;
	JMP legal_song;

legal_song:

	DO CLRMAR;
	JSB dump_memory;

	/* If the switches are %h0f, don't check the song number */
	DO DRSWT LDCMP;
	DO immed %h0f;
	if /EQUAL CJMP check_switches;
	DO MARCNT;
	JMP accept_song;

check_switches:
	DO DRSRAM MARCNT;
	/* start over if it is the wrong song */
	if /EQUAL CJMP wait_loop;

accept_song:
	/* skip addr 1 */
	DO MARCNT;

	/* set up end-of-song test */
	DO immed %hff LDCMP;

play_note:
	DO DRSRAM LDADDREG MARCNT;
	/* if the note just loaded is 0xFF, start over */
	if /EQUAL CJMP continue_current_song;
	JMP play_song;

continue_current_song:
	LDCT DRSRAM MARCNT;

play_wait:
	JSB freq_wait;
	JSB freq_wait;
	JSB freq_wait;
	JSB freq_wait;
	JSB freq_wait;
	JSB freq_wait;
	JSB freq_wait;
	JSB freq_wait;
	RPCT play_wait;
	JMP play_note;

/*************************************************************************/

/* read the next character from the UART, add it to cksum */

read_char:
	/* the new character is automatically read into the I/O pal */
	if /PE CJMP read_char_1;
	JMP uart_error;
read_char_1:
	if /FE CJMP read_char_2;
	JMP uart_error;
read_char_2:
	if /OR CJMP read_char_3;
	JMP uart_error;
read_char_3:
	if /DAVS CJMP read_char;
	DO;
	DO RDAV;

/* Read *one* character from the UART.  This does nasty things to the */
/* checksum if called directly. */
read_one_char:
	if /PE CJMP read_one_char_1;
	JMP uart_error;
read_one_char_1:
	if /FE CJMP read_one_char_2;
	JMP uart_error;
read_one_char_2:
	if /OR CJMP read_one_char_3;
	JMP uart_error;
read_one_char_3:
	if /DAVS CJMP read_one_char;
	DO DRCHAR LDADDREG;
	DO LDACCUM;
	DO RDAV;
	RTN;

/***********************************************************************/

/* random stuff for debugging */

dump_memory:
	DO CLRMAR;
	DO DRSRAM MARCNT;
	DO DRSRAM MARCNT;
	DO DRSRAM MARCNT;
	DO DRSRAM MARCNT;
	DO DRSRAM MARCNT;
	DO DRSRAM MARCNT;
	DO DRSRAM MARCNT;
	DO DRSRAM MARCNT;
	DO CLRMAR;
	RTN;

clear_checksum:
	DO CLRCKSUM;
	if /CKSUM_ZERO CJMP clear_checksum_failed;
	RTN;

clear_checksum_failed:
	JSB dump_memory;
	DO CLRMAR;
	DO immed %h0f LDSRAM;
	DO MARCNT;
	DO immed %hf0 LDSRAM;
	JMP display_error;

read_loop_not_colon:
	JSB dump_memory;
	DO CLRMAR;
	DO immed %h03 LDSRAM;
	DO MARCNT;
	DO immed %hC0 LDSRAM;
	JMP display_error;

ldsram_failed:
	JSB dump_memory;
	DO CLRMAR;
	DO immed %haa LDSRAM;
	DO MARCNT;
	DO immed %h77 LDSRAM;
	JMP display_error;

illegal_song:
	JSB dump_memory;
	DO CLRMAR;
	DO immed %h01 LDSRAM;
	DO MARCNT;
	DO immed %h80 LDSRAM;
	JMP display_error;

uart_error:
	DO RDAV;
	DO RDAV;
	DO RDAV;
	JSB dump_memory;
	DO CLRMAR;
	DO immed %h80 LDSRAM;
	DO MARCNT;
	DO immed %h01 LDSRAM;
	JMP display_error;

display_error:
	DO CLRMAR;
	DO DRSRAM LDADDREG MARCNT;
	JSB wait_half_second;
	DO DRSRAM LDADDREG MARCNT;
	JSB wait_half_second;
	JMP display_error;

/***********************************************************************/

/* Timing Code */

freq_wait:
	JSB wait_30; JSB wait_30; JSB wait_28;
	DO LDACCUM;
	RTN;

wait_30:
	DO; DO; 
wait_28:
        DO; DO; DO; DO; DO; DO; DO; DO; DO; DO;
	DO; DO; DO; DO; DO; DO; DO; DO; DO; DO; 
	DO; DO; DO; DO; DO; DO; DO;
	RTN;

wait_half_second:
	LDCT immed %hff;
	
wait_half_second_1:
	JSB freq_wait;
	JSB freq_wait;
	JSB freq_wait;
	JSB freq_wait;
	JSB freq_wait;
	JSB freq_wait;
	JSB freq_wait;
	JSB freq_wait;
	RPCT wait_half_second_1;
	RTN;
