#include <stdio.h>
#include <stdlib.h>
#include "MatrixMisc.H"
#include "Transform.H"
#include "View3D.h"

static const int debug = 0;

Matrix Perspective(const Matrix& m1);
Matrix PreClipTransform(Matrix&, Matrix&, Matrix&, double, double);
void PrintMatrix(char *msg, Matrix &m);

Solid View3D(Solid& s, Matrix& Peye_w, Matrix& Pfix_w, Matrix& UP_w,
	     double D, double F, double xt, double yt,
	     double xs, double ys)
{
     Solid s2(s);
     
     Matrix pc(Perspective(s.Vertices()*
			   PreClipTransform(Peye_w, Pfix_w, UP_w, D, F)));
     
     s2.Vertices() = pc*Scaler(xs, ys, 1.0)*Translator(xt, yt, 0.0);

     if (debug) {
	  PrintMatrix("Translated, rotated, normalized, perspected:", pc);
	  printf("\n");
	  PrintMatrix("Mapped to screen:", s2.Vertices());
     }

     return s2;
}

Matrix PreClipTransform(Matrix& Peye_w, Matrix& Pfix_w, Matrix& UP_w,
			double D, double F)
{
     Matrix xhat(1, 3), yhat(1, 3), zhat(1, 3);
     Matrix View(1, 3);
     View = Peye_w - Pfix_w;
     double S = D/2, DF = D/F;

     Matrix ViewHat(Unitize(View));
     zhat = -1.0*ViewHat;
     xhat = Unitize(CrossProd(UP_w, zhat));
     yhat = CrossProd(zhat, xhat);

     Matrix Rot(4, 4,
		xhat(1, 1), yhat(1, 1), zhat(1, 1), 0.0, 
		xhat(1, 2), yhat(1, 2), zhat(1, 2), 0.0, 
		xhat(1, 3), yhat(1, 3), zhat(1, 3), 0.0, 
		0.0,        0.0,        0.0,        1.0);
     Matrix N(4, 4,
	      1.0, 0.0, 0.0, 0.0,
	      0.0, 1.0, 0.0, 0.0,
	      0.0, 0.0, S/(D*(1-DF)), S/D,
	      0.0, 0.0, -S/(1-DF), 0.0);

     Matrix m(Translator(-Peye_w(1, 1), -Peye_w(1, 2), -Peye_w(1, 3)) *
	      Rot * N);

     if (debug) {
	  PrintMatrix("View:", View);
	  PrintMatrix("Rotation matrix:", Rot);
	  PrintMatrix("Normalization:", N);
	  PrintMatrix("Result:", m);
     }

     return m;
}

Matrix Perspective(const Matrix& m1)
{
     Matrix m2(m1);
     double v;
     int i;

     for (i = 1; i <= m1.Rows(); i++) {
	  v = m2(i, 4);
	  m2(i, 1) /= v;
	  m2(i, 2) /= v;
	  m2(i, 3) /= v;
	  m2(i, 4) = 1.0;
     }

     return m2;
}

void PrintMatrix(char *msg, Matrix &m)
{
     printf("%s\n", msg);
     m.Print();
}
