#include <stdio.h>
#include <math.h>
#include "halftone.h"
#include <X11/Xlib.h>
extern "C" {
#include <tgif/tgif.h>
#include <image/Image.h>
}
void PlotPoints(XPoint *, int);

#ifdef notdef /* __GNUG__ */
/* Image.h is too broken for cfront */
void FS_Halftone(Image *img)
   // Assumes the values in the image range in [0,255].
{
     const int h = ImgCols(img), w = ImgRows(img);
     double error, val;
     int i, j;

     // The bizarre breakdown in this function is just a hack to avoid
     // lots of unnecessary boundary-condition if statements.

     // Draw all the scanlines except the last one.
     for (i = 0; i < w-1; i++) {

	  // Draw all the points on the scanline except the last.
	  for (j = 0; j < h-1; j++) {
	       val = rint(RefImg(img, i, j) / 255.0) * 255.0;
	       error = RefImg(img, i, j) - val;
	       if (val < 255.0)
		    Plot(j, h - i);
	       SetImg(img, i+1, j, RefImg(img, i+1, j) + 0.375*error);
	       SetImg(img, i, j+1, RefImg(img, i, j+1) + 0.375*error);
	       SetImg(img, i+1, j+1, RefImg(img, i+1, j+1) + 0.25*error);
	  }

	  // Now draw the last one, and propagate the error down only.
	  val = rint(RefImg(img, i, j) / 255.0) * 255.0;
	  error = RefImg(img, i, j) - val;
	  if (val < 255.0)
	       Plot(j, h - i);
	  SetImg(img, i+1, j, RefImg(img, i+1, j) + 0.375*error);
     }

     // Now draw the last scan line, except the last point,
     // propagating only to the right.
     for (j = 0; j < h-1; j++) {
	  val = rint(RefImg(img, i, j) / 255.0) * 255.0;
	  error = RefImg(img, i, j) - val;
	  if (val < 255.0)
	       Plot(j, h - i);
	  SetImg(img, i, j+1, RefImg(img, i, j+1) + 0.375*error);
     }

     // Now draw the last point.
     val = rint(RefImg(img, i, j) / 255.0) * 255.0;
     error = RefImg(img, i, j) - val;
     if (val < 255.0)
	  Plot(j, h - i);
}
#endif /* notdef __GNUG__ */

void FS_Halftone(Matrix& m, double max)
{
     const int h = m.Cols(), w = m.Rows();
     const double thresh = max / 2.0;
     XPoint *pts;
     double error, vh, d;
     int y, x, cnt;

     pts = new XPoint[h];
     
     // The bizarre breakdown in this function is just a hack to avoid
     // lots of unnecessary boundary-condition if statements.

     // Draw all the scanlines except the last one.
     for (y = 1; y <= w-1; y++) {

	  cnt = 0;
	  
	  // Draw all the points on the scanline except the last.
	  for (x = 1; x <= h-1; x++) {
	       if (m(x, y) > thresh) {
		    error = m(x, y) - max;
	       } else {
		    error = m(x, y);
		    pts[cnt].x = x;
		    pts[cnt++].y = h - 1 - y;
	       }
	       vh = 0.375*error;
	       d = 0.25*error;
	       m(x, y+1) = m(x, y+1) + vh;
	       m(x+1, y) = m(x+1, y) + vh;
	       m(x+1, y+1) = m(x+1, y+1) + d;
	  }

	  // Now draw the last one, and propagate the error down only.
	  if (m(x, y) > thresh) {
	       error = m(x, y) - max;
	  } else {
	       error = m(x, y);
	       pts[cnt].x = x;
	       pts[cnt++].y = h - 1 - y;
	  }
	  m(x, y+1) = m(x, y+1) + 0.375*error;

	  PlotPoints(pts, cnt);
     }

     // Now draw the last scan line, except the last point,
     // propagating only to the right.
     for (cnt = 0, x = 1; x <= h-1; x++) {
	  if (m(x, y) > thresh) {
	       error = m(x, y) - max;
	  } else {
	       error = m(x, y);
	       pts[cnt].x = x;
	       pts[cnt++].y = h - 1 - y;
	  }
	  m(x+1, y) = m(x+1, y) + 0.375*error;
     }

     // Now draw the last point.
     if (m(x, y) <= thresh) {
	  pts[cnt].x = x;
	  pts[cnt++].y = h - 1 - y;
     }
	  
     PlotPoints(pts, cnt);

     delete [h] pts;
}

