#include <math.h>
#include "Matrix.h"
#include "transform.h"

Matrix Translator(Matrix& p, int i)
   // Returns an N dimentional matrix to translate by the ith row of p,
   // or NAM.
{
     if (i > p.Rows())
	  return Matrix::NAM();
     
     Matrix m(p.Cols(), p.Cols());
     int j;

     m.Identify();
     for (j = 1; j < p.Cols(); j ++)
	  m(p.Cols(), j) = p(i, j);

     return m;
}

Matrix Translator(int dx, int dy)
{
     Matrix m(3, 3,
	      1,   0, 0,
	      0,   1, 0,
	      dx, dy, 1);

     return m;
}

Matrix Translator(int dx, int dy, int dz)
{
     Matrix m(4, 4,
	      1, 0, 0, 0,
	      0, 1, 0, 0,
	      0, 0, 1, 0,
	      dx, dy, dz, 1);

     return m;
}

Matrix Translator(double dx, double dy, double dz)
{
     Matrix m(4, 4,
	      1.0, 0.0, 0.0, 0.0,
	      0.0, 1.0, 0.0, 0.0,
	      0.0, 0.0, 1.0, 0.0,
	      dx, dy, dz, 1.0);

     return m;
}

Matrix Scaler(int sx, int sy)
{
     Matrix m(3, 3,
	      sx,  0, 0,
	      0,  sy, 0,
	      0,   0, 1);

     return m;
}

Matrix Scaler(int sx, int sy, int sz)
{
     Matrix m(4, 4,
	      sx, 0, 0, 0,
	      0, sy, 0, 0,
	      0, 0, sz, 0,
	      0, 0, 0, 1);

     return m;
}

Matrix Scaler(double sx, double sy, double sz)
{
     Matrix m(4, 4,
	      sx, 0.0, 0.0, 0.0,
	      0.0, sy, 0.0, 0.0,
	      0.0, 0.0, sz, 0.0,
	      0.0, 0.0, 0.0, 1.0);

     return m;
}

Matrix Scaler(double sx, double sy)
{
     Matrix m(3, 3,
	      sx, 0.0, 0.0,
	      0.0, sy, 0.0,
	      0.0, 0.0, 1.0);

     return m;
}

Matrix Rotator(int theta, int xi, int yi)
{
     const double x = double(xi);
     const double y = double(yi);
     const double t = theta * PI / 180.0;
     const double ct = cos(t);
     const double st = sin(t);
     const double xt = x*(1.0 - ct) + y*st;
     const double yt = y*(1.0 - ct) - x*st;
     Matrix m(3, 3,
	      ct,  st, 0.0,
	      -st, ct, 0.0,
	      xt,  yt, 1.0);

     return m;
}

Matrix Rotator_X(double theta)
{
     const double t = theta * PI / 180.0;
     const double ct = cos(t);
     const double st = sin(t);
     Matrix m(4, 4,
	      1.0, 0.0, 0.0, 0.0,
	      0.0, ct, st, 0.0,
	      0.0, -st, ct, 0.0,
	      0.0, 0.0, 0.0, 1.0);

     return m;
}

Matrix Rotator_Y(double theta)
{
     const double t = theta * PI / 180.0;
     const double ct = cos(t);
     const double st = sin(t);
     Matrix m(4, 4,
	      ct, 0.0, -st, 0.0,
	      0.0, 1.0, 0.0, 0.0,
	      st, 0.0, ct, 0.0,
	      0.0, 0.0, 0.0, 1.0);

     return m;
}

Matrix Rotator_Z(double theta)
{
     const double t = theta * PI / 180.0;
     const double ct = cos(t);
     const double st = sin(t);
     Matrix m(4, 4,
	      ct, st, 0.0, 0.0,
	      -st, ct, 0.0, 0.0,
	      0.0, 0.0, 1.0, 0.0,
	      0.0, 0.0, 0.0, 1.0);

     return m;
}
