/*
 * $Id: math0.as,v 1.4 91/12/12 01:57:13 bjaspan Exp Locker: bjaspan $
 * $Source: /afs/athena.mit.edu/user/b/bjaspan/classes/6111/project/math/RCS/math0.as,v $
 */

# SPEC_FILE = math.sp;
# LIST_FILE = math0.lst;
# MASK_COUNT = 8;
# SET_ADDRESS = 0;
# LOAD_ADDRESS = 0;

/*
 * Memory map: Each page is 256 bytes.
 *
 * Page 0:
 * 		00-31	matrix A
 * 		32-63	matrix B
 * 		64-95	matrix C
 *
 * 		new transformation matrices are written into B, and
 * 		the CTM swaps between A and C (to avoid one copy)
 *
 * 		096	&CTM
 *		097	&NCTM
 * 		098-105	TMP1-TMP8
 *
 * 		106-111 eye point
 * 		112-117 fixatoint point
 *
 * 		118	number of vertices in object (in page 1)
 * 		119	number of edges in object
 * 		120-255	edge list
 *
 * 		68 edges (pairs of integers) can be stored
 *
 * Page 1:
 * 		00-255	matrix of vertices
 *
 * 		32 vertices can be stored
 *
 * Page 2:
 * 		00-255	transformed matrix of vertices
 */

start:
        /*
	 * Initialization:
	 * 1) set up default CTM, object, etc (ROM dump)
	 * 2) set TMP1-TMP8 to zero
	 * 3) reset ACK_O and DAV_O
	 */
	LDMAR CTM_ADDR;
	LDSRAM 00; do MARCNT;
	LDSRAM 64; do MARCNT;
	LDSRAM 0; do MARCNT;
	LDSRAM 0; do MARCNT;
	LDSRAM 0; do MARCNT;
	LDSRAM 0; do MARCNT;
	LDSRAM 0; do MARCNT;
	LDSRAM 0; do MARCNT;
	LDSRAM 0; do MARCNT;
	LDSRAM 0; do MARCNT;
	do CLR_T;
	LDREG r4 main_loop;
	jmp reset_ctm;

/* 
 * Wait for p0 to be pressed, execute block in CROM[swt]
 */
main_loop:
	if /p0 cjmp main_loop;
main_release_wait:
	if /p0 cjmp main_released;
	jmp main_release_wait;
main_released:

	LDREG r2 DRSWT;

	/* read a block opcode into B */
	LDA 1;
	LDMAR TMP1;
	LDREG r4 l5;
	jmp read_bytes;

l5:	
	LDB DRREG r1;

	/* Jump to the appropriate block handler.  The order here */
	/* depends on the opcode values defined by the protocol. */
	LDA OBJECT;
	latch;
	if /equal cjmp not_object;
	jmp object_block;

not_object:
	LDA aplus1;
	latch;
	if /equal cjmp not_eye;
	jmp eye_block;

not_eye:
	LDA aplus1;
	latch;
	if /equal cjmp not_fix;
	jmp fix_block;

not_fix:
	LDA aplus1;
	latch;
	if /equal cjmp not_identify;
	jmp identify_block;

not_identify:
	LDA aplus1;
	latch;
	if /equal cjmp not_transform;
	jmp transform_block;

not_transform:
	/* the block opcode is not defined */
	jmp bad_block;

/***********************************************************************
 * check_eob_marker -- read a byte, check that it is 0xff, and return
 * to the main loop.  This is used as a return addr from read_bytes.
 ***********************************************************************/
check_eob_marker:
	LDMAR TMP1;
	LDA 1;
	LDREG r4 check_eob_marker_1;
	jmp read_bytes;
check_eob_marker_1:
	LDA DRREG r1;
	LDB %hff;
	latch;
	if /equal cjmp bad_block;
	jmp render_object;

/***********************************************************************
 * reset_ctm -- set <CTM_ADDR> to the identity matrix, and return to r4
 ***********************************************************************/
reset_ctm:
	brk 5;
	LDMAR CTM_ADDR;
	LDMAR DRSRAM;
	LDA 32;
	LDB 1;
reset_ctm_loop:
	LDA subc;
	if carry cjmp reset_ctm_cont;
	jmp reset_ctm_done;
reset_ctm_cont:
	LDSRAM 0;
	do MARCNT;
	jmp reset_ctm_loop;
reset_ctm_done:
	LDMAR CTM_ADDR;
	LDA DRSRAM;
	LDB 10;
	LDMAR a;
	LDSRAM 1;
	LDA add;
	LDMAR a;
	LDSRAM 1;
	LDA add;
	LDMAR a;
	LDSRAM 1;
	LDA add;
	LDMAR a;
	LDSRAM 1;
	brk 5;
	jmp DRREG r4;

/***********************************************************************
 * identify_block -- read nothing, reset CTM to identity (well...)
 ***********************************************************************/
identify_block:
	LDREG r4 check_eob_marker;
	jmp reset_ctm;

/***********************************************************************
 * fix_block, eye_block -- read three bytes into EYE_POINT or FIX_POINT
 ***********************************************************************/
eye_block:
	LDMAR EYE_POINT;
	jmp eye_or_fix_block;

fix_block:
	LDMAR FIX_POINT;
	jmp eye_or_fix_block;

eye_or_fix_block:
	LDREG r4 check_eob_marker;
	LDA 3;
	jmp read_bytes;

/***********************************************************************
 * transform_block -- read in a transformation matrix, apply it to the 
 * CTM, and render the new image (well...)
 ***********************************************************************/
transform_block:
	/* read all transformations in matrix B */
	LDMAR MATRIX_B;
	LDA 32;
	LDREG r4 transform_block_1;
	jmp read_bytes;
transform_block_1:
	/* check end-of-block marker  */
	LDA 1;
	LDMAR TMP1;
	LDREG r4 transform_block_2;
	jmp read_bytes;
transform_block_2:
	LDA 1;
	LDB DRREG r1;
	add;
	if carry cjmp transform_object;
	jmp bad_block;

transform_object:
	brk 5;
	/* The transformation has been received and is correct. */
	/* Multiply it into the CTM_ADDR. */
	LDMAR CTM_ADDR;
	LDREG OP1 DRSRAM;
	LDMAR NCTM_ADDR;
	LDREG DST DRSRAM;
	LDREG OP2 MATRIX_B;
	LDMAR TMP1;
	LDSRAM 4;
	LDMAR TMP3;
	LDSRAM 0;
	do MARCNT;
	LDSRAM 0;
	do MARCNT;
	LDSRAM 0;
	LDREG r4 transform_multiply_rtn;
	jmp matrix_mult;
transform_multiply_rtn:
	/* swap CTM_ADDR and NCTM_ADDR */
	LDMAR CTM_ADDR;
	LDREG r1 DRSRAM;
	LDMAR NCTM_ADDR;
	LDREG r2 DRSRAM;
	LDSRAM DRREG r1;
	LDMAR CTM_ADDR;
	LDSRAM DRREG r2;
	brk 5;
	jmp render_object;

/***********************************************************************
 * object_block -- read in an object block and store it in memory.
 * Note that the vertex data in stored in page 1
 ***********************************************************************/
object_block:
	/* read in the number of vertices */
	LDMAR NUM_VERT;
	LDA 1;
	LDREG r4 obj_rd_vert;
	jmp read_bytes;
obj_rd_vert:
	/* read in the vertex data */
	LDA DRREG r1;
	LDMPR 1;
	LDMAR VERT_LIST;
	LDREG r4 obj_rd_vert_1;
	jmp read_bytes;
obj_rd_vert_1:
	/* read and test the 0xFF */
	LDA 1;
     	LDMPR 0;
	LDMAR TMP1;
	LDREG r4 obj_rd_vert_2;
	jmp read_bytes;
obj_rd_vert_2:
	LDA DRREG r1;
	LDB 1;
	add;
	if carry cjmp object_rd_ed;
	jmp bad_block;

object_rd_ed:
	/* read in the number of edges */
	LDMAR NUM_EDGE;
	LDA 1;
	LDREG r4 object_rd_ed_1;
	jmp read_bytes;
object_rd_ed_1:
	LDMAR EDGE_LIST;
	LDA DRREG r1;
	LDREG r4 object_rd_ed_2;
	jmp read_bytes;
object_rd_ed_2:

	/* Divide NUM_VERT by 8 and NUM_EDGE by 2 to make them accurate. */
	LDMAR NUM_VERT;
	LDA DRSRAM;
	LDB 0;
	LDX;
	LDA 0;
	LDB 32;
	LDY;
	LDSRAM DRPR_HI;

	LDMAR NUM_EDGE;
	LDA DRSRAM;
	LDB 0;
	LDX;
	LDA 0;
	LDB 128;
	LDY;
	LDSRAM DRPR_HI;

	/* XXX Skip the final FF for now.  Note that this will make */
        /* the machine fail when input arrives serially. */
	jmp render_object;

/***********************************************************************
 * render_object -- multiply the object by the CTM, storing it in page
 * 2, and then give it to the video unit
 ***********************************************************************/
render_object:
	/* TMP3:OP1 = 1:VERT_LIST */
	LDREG OP1 VERT_LIST;	
	LDMAR TMP3;
	LDSRAM 1;

	/* TMP4:OP2 = 0:<CTM_ADDR> */
	/* assume TMP4 is already 0 */
	LDMAR CTM_ADDR;
	LDREG OP2 DRSRAM;

	/* TMP5:DST = 2:VERT_LIST */
	LDREG DST VERT_LIST;
	LDMAR TMP5;
	LDSRAM 2;

	/* TMP1 = <NUM_VERT> */
	LDMAR NUM_VERT;
	LDREG r4 DRSRAM;
	LDMAR TMP1;
	LDSRAM DRREG r4;
	
	LDREG r4 render_rtn;
	jmp matrix_mult;

/***********************************************************************
 * read_bytes -- read a specified number of bytes into sequential SRAM
 * locations.
 *
 * Entry:
 * MAR contains starting address for block
 * A is the number of bytes to read
 * r2 is const rom addr
 * r4 is return addr
 *
 * Exit:
 * r1 contains last byte read
 * A = 0, B = 0
 * r2 is next const rom addr
 * MAR is address after the block read
 * r3 is destroyed
 ***********************************************************************/
read_bytes:
	/* store bytes to read in r3 */
	LDREG r3 a;

next_byte:
        /* bytes remaining = 0? */
	LDA DRREG r3;
	LDB 0;
	latch;
	if /equal cjmp block_not_done;
	jmp block_done;

block_not_done:
	/* no, decrement count by 1 */
	LDB 1;
	LDREG r3 subc;

	/* copy one byte */
	LDA DRREG r2;
	do LDSRAM DRCONST;
	LDREG r1 DRSRAM;
	do MARCNT;
	LDREG r2 add;
	jmp next_byte;

block_done:
	jmp DRREG r4;


bad_block:
	brk 10;
	jmp bad_block;

/*
 * Code warp entry points
 */
matrix_mult:
	jmp matrix_mult_entry_0;

render_rtn:
	jmp render_rtn_entry_0;

# SET_ADDRESS = f5;
render_rtn_entry_0:
	LDCSEG 2;
	/* warp to 2:f6 */

# SET_ADDRESS = f9;
render_rtn_return_0:
	jmp main_loop;

# SET_ADDRESS = fb;
matrix_mult_entry_0:
	LDCSEG 1;
	/* warp to 1:fc */

# SET_ADDRESS = fe;
matrix_mult_return_0:
	LDMAR TMP6;
	jmp DRSRAM;	
