/*
 * $Id: math1.as,v 1.4 91/12/12 01:57:39 bjaspan Exp $
 * $Source: /afs/athena.mit.edu/user/b/bjaspan/classes/6111/project/math/RCS/math1.as,v $
 */

# SPEC_FILE = math.sp;
# LIST_FILE = math1.lst;
# MASK_COUNT = 8;
# SET_ADDRESS = 0;
# LOAD_ADDRESS = 100;

/***********************************************************************
 * Matrix multiply: <DST> <-- <OP1>*<OP2>
 * Entry:
 * TMP3:OP1, TMP4:OP2 are factors
 * TMP5:DST is result
 * TMP1 is the number of rows (vertices) in OP1
 * r4 is return addr
 *
 * Exit:
 * OP1, OP2, DST, TMP2, TMP6 are destroyed
 ***********************************************************************/
matrix_mult_entry:
	/* save return addr in TMP6 */
	LDMAR TMP6;
	LDSRAM DRREG R4;

do_row:
	/* inner loop: TMP2 is number of cols in OP2 (always 4) */
	LDMAR TMP2;
	LDSRAM 4;

	/* decrement and test outer loop counter */
	LDMAR TMP1;
	LDA DRSRAM;
	LDB 1;
	LDSRAM subc;
	if carry cjmp next_col;
	jmp mult_done;
	
next_col:
	/* debugging code */
	DRREG OP1;
	DRREG OP2;
	DRREG DST;

	/* decrement and test inner loop counter */
	LDMAR TMP2;
	LDA DRSRAM;
	LDB 1;
	LDSRAM subc;
	if carry cjmp inner_not_done;
	jmp next_row;

inner_not_done:
	/* set the current element to zero */
	LDMAR TMP5;
	LDMPR DRSRAM;
	LDMAR DRREG DST;
	LDSRAM 0;
	do MARCNT;
	LDSRAM 0;
	LDMPR 0;

	LDREG r4 l1;
	jmp accum_prod;
l1:	LDREG r4 l2;
	jmp accum_prod;
l2:	LDREG r4 l3;
	jmp accum_prod;
l3:	LDREG r4 l4;
	jmp accum_prod;
l4:	

	/* dump DST for debugging */
	LDMAR TMP5;
	LDMPR DRSRAM;
	LDMAR DRREG DST;
	DRSRAM MARCNT;
	DRSRAM;
	LDMPR 0;

	/* update DST to next element */
	LDA DRREG DST;
	LDB 2;
	LDREG DST add;

	/* reset OP1 to beginning of row */
	LDA DRREG OP1;
	LDB 8;
	LDREG OP1 subc;
     
	/* update OP2 to use next column */
	LDA DRREG OP2;
	LDB 30;
	LDREG OP2 subc;

	jmp next_col;

next_row:
	/* update OP1 to use next row */
	LDA DRREG OP1;
	LDB 8;
	LDREG OP1 add;

	/* reset OP2 to beginning of matrix */
	LDA DRREG OP2;
	LDB 8;
	LDREG OP2 subc;

	jmp do_row;

mult_done:
	jmp matrix_mult_return_1;

/*
 * accum_prod -- perform one multiplication-addition in a matrix
 * multiplication.  <DST> <-- <DST> + prod
 * 
 * Algorithm:
 * (1) load factor from OP1 into A and B, load into multiplier X
 * (2) load factor from OP2 into A and B, load into multiplier Y
 * (3) load lb of <DST> into A, lb of prod into B, add and
 * store result into lb of <DST> *and* latch CCs.
 * (4) repeat (3) with hb, adding a carry-in if lb addition produced a
 * carry out
 * (5) increment OP1 by 2, OP2 by 8
 *
 * Note: This "subroutine" assumes its return address is in r4
 */
accum_prod:
	/* A, B <- <OP1> */
	LDMAR TMP3;
	LDMPR DRSRAM;
	LDMAR DRREG OP1;
	LDA DRSRAM MARCNT;
	LDB DRSRAM MARCNT;
	LDMPR 0;
	LDX;

	/* A, B <- <OP2> */
	LDMAR TMP4;
	LDMPR DRSRAM;
	LDMAR DRREG OP2;
	LDA DRSRAM MARCNT;
	LDB DRSRAM MARCNT;
	LDMPR 0;
	LDY;

	/* set all sram references to DST page */
	LDMAR TMP5;
	LDMPR DRSRAM;

	/* <dst+1> <- <dst+1> + low prod */
	LDMAR DRREG DST;
	do MARCNT;
	LDA DRSRAM;
	LDB DRPR_LO;
	LDSRAM add;

	/* <dst> <- <dst> + hi prod + carry */
	LDMAR DRREG DST;
	LDA DRSRAM;
	LDB DRPR_HI;
	if carry cjmp accum_with_carry;
	LDSRAM add;
	jmp done_with_accum;
accum_with_carry:
	LDSRAM addc;

done_with_accum:
	LDMPR 0;

	/* OP1 += 2, OP2 += 8 */
	LDA DRREG OP1;
	LDB 2;
	LDREG OP1 add;

	LDA DRREG OP2;
	LDB 8;
	LDREG OP2 add;
     
	jmp DRREG r4;

# SET_ADDRESS = fc;
# LOAD_ADDRESS = 1fc;
matrix_mult_entry_1:
	jmp matrix_mult_entry;

# SET_ADDRESS = fd;
# LOAD_ADDRESS = 1fd;
matrix_mult_return_1:
	LDCSEG 0;
	/* warp to 0:fe */
