#include <stdio.h>
#include <math.h>

#define DEBUG

/*
 * Memory module
 */
#define SRAM_SIZE 2048
#define DRSRAM sram[(mpr << 8) + mar]
#define LDSRAM sram[(mpr << 8) + mar]
#define MARCNT mar++
#define SKIP_ELEMENT MARCNT; MARCNT;
#define SKIP_ROW SKIP_ELEMENT SKIP_ELEMENT SKIP_ELEMENT SKIP_ELEMENT
     
int sram[SRAM_SIZE];		/* sram */
int mar;			/* memory address register */
int mpr;			/* memory page register */

/*
 * Math module
 */
#define LDX X = ((A << 8) + B)
#define LDY Y = ((A << 8) + B)
#define MULT ((((unsigned) X*Y)>>8) & 0xffff)
#define MULT_LO (MULT & 0xff)
#define MULT_HI ((MULT >> 8) & 0xff)
#define CARRY_MASK 0x1
#define CARRY_ACTIVE (!(alu_cc & CARRY_MASK))
#define EQUAL_MASK 0x2
#define EQUAL_ACTIVE (alu_cc & EQUAL_MASK)

#define add alu(9, 1)
#define addc alu(9, 0)
#define sub alu(6, 1)
#define subc alu(6, 0)

int A, B;			/* registers feeding multiplier and ALU*/
int X, Y;			/* internal 16 bit operands */

int alu_cc;			/* ALU cc bits */

/*
 * Register file
 */
int r1, r2, r3, r4;		/* four general registers */
#define OP1 r1
#define OP2 r2
#define DST r3

#define MATRIX_A 0
#define MATRIX_B 32
#define MATRIX_C 64
#define CTM 96
#define NCTM 97
#define TMP1 98
#define TMP2 99
#define TMP3 100
#define TMP4 101
#define TMP5 102
#define TMP6 103
#define TMP7 104
#define TMP8 105
#define EYE_POINT 106
#define FIX_POINT 112
#define NUM_VERT 118
#define NUM_EDGE 119
#define EDGE_LIST 120
#define VERT_LIST 0

int mat_a[] = {
     1, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 1, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 1, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 1, 0,
};
int mat_b[] = {
     1, 0, 0,   0,   0, 0,   2, 0,
     0, 0, 0,   222, 0, 128, 0, 0,
     0, 0, 255, 128, 0, 222, 0, 0,
     0, 0, 0,   0,   0, 0,   1, 0,
};

main()
{
     magic_sram_setup();
     print_matrix(MATRIX_A);
     print_matrix(MATRIX_B);

     OP1 = MATRIX_A;
     OP2 = MATRIX_B;
     DST = MATRIX_C;
     sram[TMP1] = 4;
     sram[TMP3] = 0;
     sram[TMP4] = 0;
     sram[TMP5] = 1;

     matrix_mult();

     print_matrix((1 << 8) + MATRIX_C);
}

/*
 * Matrix multiply: <DST> <-- <OP1>*<OP2>
 * number of rows (vertices) in <OP1> in TMP1
 * page for OP1 is TMP3, page for OP2 is TMP4, page for DST is TMP5
 */
matrix_mult()
{
do_row:
     /* inner loop: TMP2 is number of cols in OP2 (always 4) */
     mar = TMP2;
     LDSRAM = 4;

     /* decrement and test outer loop counter */
     mar = TMP1;
     A = DRSRAM;
     B = 1;
     LDSRAM = subc;
     if (CARRY_ACTIVE)
	  ;
     else
	  goto mult_done;
     
next_col:

     /* decrement and test inner loop counter */
     mar = TMP2;
     A = DRSRAM;
     B = 1;
     LDSRAM = subc;
     if (CARRY_ACTIVE)
	  ;
     else
	  goto next_row;
     
#ifdef DEBUG
     printf("row %d, col %d, dst %d\n", OP1, OP2, DST);
#endif
     
     /* set the current element to zero */
     mar = TMP5;
     mpr = DRSRAM;
     mar = DST;
     LDSRAM = 0; MARCNT;
     LDSRAM = 0;
     mpr = 0;

     accum_prod();
     accum_prod();
     accum_prod();
     accum_prod();

     /* update DST to next element */
     A = DST;
     B = 2;
     DST = add;

     /* reset OP1 to beginning of row */
     A = OP1;
     B = 8;
     OP1 = subc;
     
     /* update OP2 to use next column */
     A = OP2;
     B = 30;
     OP2 = subc;

     goto next_col;

next_row:
     /* update OP1 to use next row */
     A = OP1;
     B = 8;
     OP1 = add;

     /* reset OP2 to beginning of matrix */
     A = OP2;
     B = 8;
     OP2 = subc;

     goto do_row;

mult_done:
     return;
}

/*
 * Accumulate multiplier output: <DST> <-- <DST> + prod
 * Algorithm: load lb of <DST> into A, lb of prod into B, add and
 * store result into lb of <DST> *and* latch CCs.  repeat with hb,
 * adding a carry-in if lb addition produced a carry out
 */
accum_prod()
{
     /* A, B <- <OP1> */
     mar = TMP3;
     mpr = DRSRAM;
     mar = OP1;
     A = DRSRAM; MARCNT;
     B = DRSRAM; MARCNT;
     mpr = 0;
     LDX;

     /* A, B <- <OP2> */
     mar = TMP4;
     mpr = DRSRAM;
     mar = OP2;
     A = DRSRAM; MARCNT;
     B = DRSRAM; MARCNT;
     mpr = 0;
     LDY;

     mar = TMP5;
     mpr = DRSRAM;
     
     /* <dst+1> <= <dst+1> + low prod */
     mar = DST;
     MARCNT;
     A = DRSRAM;
     B = MULT_LO;
     LDSRAM = add;

     /* <dst> <- <dst> + hi prod + carry */
     mar = DST;
     A = DRSRAM;
     B = MULT_HI;
     if (CARRY_ACTIVE) goto add_with_carry;
     LDSRAM = add;
     goto done_with_accum;
add_with_carry:
     LDSRAM = addc;

done_with_accum:
     mpr = 0;
     A = OP1;
     B = 2;
     OP1 = add;

     A = OP2;
     B = 8;
     OP2 = add;

#ifdef DEBUG2
     printf("OP1 = %d, OP2 = %d, accum = %d.%d\n", OP1, OP2,
	    sram[(sram[TMP5] << 8) + DST], sram[(sram[TMP5] << 8) + DST+1]);
#endif
     return;
}

alu(op, cin)
   int op, cin;
{
     int c;

     /*
      * A subtraction will produce a carry output (the Cout pin will
      * be low == active) and the subtraction does *NOT* require a
      * borrow.  Example:  1 - 1 = 0, but no borrow was required so
      * Cout will be low (active).  0 - 1 = 255, and a borrow was
      * necessary so Cout will be high (inactive).
      */
     
     if (A == B)
	  alu_cc |= EQUAL_MASK;

     c = 0xff;
     switch (op) {
     case 6:
	  if (cin) c = A + ((B ^ 0xff) & 0xff);
	  else c = A + ((B ^ 0xff) & 0xff) + 1;
	  break;
     case 9:
	  if (cin) c = A + B;
	  else c = A + B + 1;
	  break;
     case 15:
	  if (cin) c = A - 1;
	  else c = A;
	  break;
     default:
	  printf("invalid alu op %d\n", op);
	  exit(1);
     }

     if (c > 255)
	  alu_cc &= (~CARRY_MASK);
     else
	  alu_cc |= CARRY_MASK;

     return (c & 0xff);
}

magic_sram_setup()
{
     int i;

     for (i=0; i<SRAM_SIZE; i++)
	  sram[i] = 0xbf;
     
     for (i=0; i<32; i++)
	  sram[MATRIX_A + i] = mat_a[i];
     for (i=0; i<32; i++)
	  sram[MATRIX_B + i] = mat_b[i];
}
     
print_matrix(index)
   int index;
{
     int i, j;

     for (i=0; i<4; i++) {
	  for (j=0; j<8; j += 2)
	       printf("%d.%d ", sram[index+8*i+j] & 0xff,
		      sram[index+8*i+j+1] & 0xff);
	  printf("\n");
     }
     printf("\n");
}
