#include <stdio.h>

/* Control bits */
int bits, cc, dr, ld, aluf, val, ctrl, reg, alu_cin;

/* hardware storage */
char A, B;
int output;
short X, Y;
int mpr, mar;
int pseg, pc, paddr;
int regs[4];
unsigned char sram[2048];
unsigned char rom2[2048], rom1[2048], rom0[2048];
unsigned char crom[256];
int bus;
int alu_cc;

/* Simulator control */
int print_state;

#define CARRY_MASK 0x1
#define CARRY_ACTIVE (!(alu_cc & CARRY_MASK))
#define EQUAL_MASK 0x2
#define EQUAL_ACTIVE (alu_cc & EQUAL_MASK)
#define MULT (((X*Y)>>8) & 0xffff)
#define MULT_LO (MULT & 0xff)
#define MULT_HI ((MULT >> 8) & 0xff)


#define DRROM 0
#define DRALU 1
#define DRREG 2
#define DRPR_LO 3
#define DRPR_HI 4
#define DRSRAM 5
#define DRSWT 6
#define DRIN 7

#define LDA 1
#define LDB 2
#define LDX 3
#define LDY 4
#define LDREG 5
#define LDSRAM 6
#define LDMAR 7
#define LDMPR 8
#define LDOUT 9
#define LDCSEG 10

#define MARCNT 1
#define Cin 2
#define DRCONST 3
#define CLR_T 4
#define ACK_T 5
#define DAV_T 6

#define assert 0
#define jmp 1
#define carry 2
#define equal 3
#define p0 4
#define ACK_I 5
#define DAV_I 6
#define brk 7

#define MATRIX_A 0
#define MATRIX_B 32
#define MATRIX_C 64
#define CTM_ADDR 96
#define NCTM_ADDR 97
#define TMP1 98
#define TMP2 99
#define TMP3 100
#define TMP4 101
#define TMP5 102
#define TMP6 103
#define TMP7 104
#define TMP8 105
#define EYE_POINT 106
#define FIX_POINT 112
#define NUM_VERT 118
#define NUM_EDGE 119
#define EDGE_LIST 120
#define VERT_LIST 0

#define OBJECT 1
#define EYE 2
#define FIX 3
#define IDENTIFY 4
#define TRANSFORM 5

init(argc, argv)
   int argc;
   char **argv;
{
     FILE *f2, *f1, *f0;
     char buf[BUFSIZ];
     int i;

     if (argc == 2) {
	  fprintf(stderr, "Usage: sim bin-name const-name\n");
	  exit(1);
     }

     strcpy(buf, argv[1]); strcat(buf, "2"); strcat(buf, ".bin");
     f2 = fopen(buf, "r");
     strcpy(buf, argv[1]); strcat(buf, "1"); strcat(buf, ".bin");
     f1 = fopen(buf, "r");
     strcpy(buf, argv[1]); strcat(buf, "0"); strcat(buf, ".bin");
     f0 = fopen(buf, "r");
     if (f2 == NULL || f1 == NULL || f0 == NULL) {
	  fprintf(stderr, "cannot open rom files, abort.\n");
	  exit(1);
     }

     fread(rom2, sizeof(char), 2048, f2);
     fread(rom1, sizeof(char), 2048, f1);
     fread(rom0, sizeof(char), 2048, f0);
     fclose(f2); fclose(f1); fclose(f0);

     strcpy(buf, argv[2]); strcat(buf, ".bin");
     f0 = fopen(buf, "r");
     if (f0 == NULL) {
	  fprintf(stderr, "cannot open const rom file, abort.\n");
	  exit(1);
     }
     fread(crom, sizeof(char), 256, f0);
     fclose(f0);

     for (i=0; i<2048; i++)
	  sram[i] = 0xbf;

     print_state = 0;
}

main(argc, argv)
   int argc;
   char **argv;
{
     char inp[BUFSIZ];
     
     init(argc, argv);

     while (1) {
	  paddr = (pseg << 8) + pc;
	  bits = (rom2[paddr] << 16) + (rom1[paddr] << 8) + (rom0[paddr]);
	  cc = (bits >> 21) & 0x7;
	  dr = (bits >> 16) & 0x7;
	  ld = (bits >> 12) & 0xf;
	  aluf = (bits >> 8) & 0xf;
	  val = (bits) & 0xff;
	  ctrl = (bits) & 0x7;
	  reg = (bits >> 19) & 0x3;

	  /* defaults */
	  alu_cin = 1;
	  bus = 0xbf;

	  /* combinational control signals */
	  if (cc == assert && dr != DRROM) {
	       
	       switch (ctrl) {
	       case 0:
		    break;
	       case Cin:
		    alu_cin = 0;
		    break;
	       case DRCONST:
		    bus = crom[A];
		    break;
	       case MARCNT:
	       case CLR_T:
	       case ACK_T:
	       case DAV_T:
		    break;
	       default:
		    fprintf(stderr, "unknown ctrl value %d\n", ctrl);
		    break;
	       }
	  }
	  
	  /* drive a value onto the bus */
	  if (cc != 0 || dr == 0 || ctrl != DRCONST) {
	       switch (dr) {
	       case DRROM:
		    bus = val;
		    break;
	       case DRALU:
		    bus = alu(aluf, alu_cin);
		    break;
	       case DRREG:
		    bus = regs[reg];
		    break;
	       case DRPR_HI:
		    bus = MULT_HI;
		    break;
	       case DRPR_LO:
		    bus = MULT_LO;
		    break;
	       case DRSRAM:
		    bus = sram[(mpr << 8) + mar];
		    break;
	       case DRSWT:
		    fprintf(stderr, "pc %x switch value: ", pc);
		    fflush(stderr);
		    fgets(inp, BUFSIZ, stdin);
		    sscanf(inp, "%x", &bus);
		    break;
	       case DRIN:
		    fprintf(stderr, "pc %x input value: ", pc);
		    fflush(stderr);
		    fgets(inp, BUFSIZ, stdin);
		    bus = atoi(inp);
		    break;
	       default:
		    fprintf(stderr, "unknown dr value %d\n", dr);
		    break;
	       }
	  }

	  /* things to print at "rising edge" */
	  if (print_state || cc == brk)
	       fprintf(stderr, "PC = %02x:%02x, bus = %02x, mar = %1x:%02x\n",
		       pseg, pc, bus, mpr, mar);

	  /* load the value from the bus */
	  switch (ld) {
	  case 0:
	       break;
	  case LDA:
	       A = bus;
	       break;
	  case LDB:
	       B = bus;
	       break;
	  case LDX:
	       X = (A << 8) + B;
	       break;
	  case LDY:
	       Y = (A << 8) + B;
	       break;
	  case LDREG:
	       regs[reg] = bus;
	       break;
	  case LDSRAM:
	       sram[(mpr << 8) + mar] = bus;
	       break;
	  case LDMAR:
	       mar = bus;
	       break;
	  case LDMPR:
	       mpr = bus;
	       break;
	  case LDOUT:
	       output = bus;
	       printf("%d\n", bus);
	       fflush(stdout);
	       break;
	  case LDCSEG:
	       pseg = bus;
	       break;
	  default:
	       fprintf(stderr, "unknown ld value %d\n", ld);
	       break;
	  }

	  /* register control signals */
	  if (cc == assert && dr != DRROM) {
	       switch (ctrl) {
	       case MARCNT:
		    mar += 1;
		    break;
	       }
	  }

	  /* compute the next program instruction */
	  switch (cc) {
	  case assert:
	       pc++;
	       break;
	  case jmp:
	       pc = bus;
	       break;
	  case carry:
	       if (CARRY_ACTIVE)
		    pc = bus;
	       else
		    pc++;
	       break;
	  case equal:
	       if (! EQUAL_ACTIVE)
		    pc = bus;
	       else
		    pc++;
	       break;
	  case p0:
	       fprintf(stderr, "pc %x p0: ", pc);
	       fflush(stderr);
	       fgets(inp, BUFSIZ, stdin);
	       if (atoi(inp) == -1) 
		    exit(0);
	       else if (atoi(inp))
		    pc++;
	       else
		    pc = bus;
	       break;
	  case brk:
	       handle_breakpoint();
	       break;
	  case ACK_I:
	       fprintf(stderr, "pc %x ack_i: ", pc);
	       fflush(stderr);
	       fgets(inp, BUFSIZ, stdin);
	       if (atoi(inp))
		    pc++;
	       else
		    pc = bus;
	       break;
	  default:
	       fprintf(stderr, "unknown cc value %d\n", cc);
	       break;
	  }
	  pc &= 0xff;
	  
     }
}

handle_breakpoint()
{
     switch (val) {
     case 0:
	  break;
     case 1:
	  print_object(2);
	  break;
     case 2:
	  print_object(1);
	  break;
     case 3:
	  print_matrix(sram[TMP5], regs[2] - 64, 8);
	  break;
     case 4:
	  print_matrix(0, MATRIX_B, 4);
	  break;
     case 5:
	  fprintf(stderr, "CTM = %d, NCTM = %d, CTM:",
		  sram[CTM_ADDR], sram[NCTM_ADDR]);
	  print_matrix(0, sram[CTM_ADDR], 4);
	  break;
     case 10:
	  print_state = 1;
	  break;
     case 11:
	  print_state = 0;
	  break;
     case 12:
	  fprintf(stderr, "sram[2:%02x] = %d.%d\n",
		  regs[2], sram[(2 << 8) + regs[2]],
		  sram[(2 << 8) + regs[2] + 1]);
	  pc++;
	  return;
     case 13:
	  fprintf(stderr, "output = %d\n", output);
	  break;
     case 15:
	  fprintf(stderr, "abs: B = %d, A = %d\n", B, A);
	  break;
     default:
	  fprintf(stderr, "Unknown breakpoint code, ignoring.\n");
	  break;
     }

     fprintf(stderr, "Breakpoint!\n");
     getchar();
     getchar();
     pc++;
}

print_matrix(page, offset, num)
   int page, offset, num;
{
     int i, j;

     fprintf(stderr, "\nMatrix in %03d:%03d, %d rows:\n",
	     page, offset, num);
     for (i = 0; i < num; i++) {
	  for (j = 0; j < 8; j+=2) {
	       fprintf(stderr, "%03d.%03d ",
		       sram[(page << 8) + i*8 + j + offset],
		       sram[(page << 8) + i*8 + j + offset + 1]);
	  }
	  fprintf(stderr, "\n");
     }
}

print_object(page)
{
     int verts, edges, i, j;

     verts = sram[118];
     fprintf(stderr, "\nObject in page %d:\n", page);
     for (i = 0; i < verts; i++) {
	  for (j = 0; j < 8; j+=2) {
	       fprintf(stderr, "%03d.%03d ", sram[(page << 8) + i*8 + j],
		       sram[(page << 8) + i*8 + j + 1]);
	  }
	  fprintf(stderr, "\n");
     }
     fprintf(stderr, "\n");

     edges = sram[119];
     fprintf(stderr, "Edge list:\n");
     for (i=0; i<2*edges; i+=2)
	  fprintf(stderr, "%02d %02d\n", sram[120+i], sram[120+i+1]);
     fprintf(stderr, "\n");
}
	  
alu(op, cin)
   int op, cin;
{
     int c;

     /*
      * A subtraction will produce a carry output (the Cout pin will
      * be low == active) and the subtraction does *NOT* require a
      * borrow.  Example:  1 - 1 = 0, but no borrow was required so
      * Cout will be low (active).  0 - 1 = 255, and a borrow was
      * necessary so Cout will be high (inactive).
      */
     
     c = 0xff;
     switch (op) {
     case 0:
	  if (cin) c = A;
	  else c = A+1;
	  break;
     case 6:
	  if (cin) c = A + ((B ^ 0xff) & 0xff);
	  else c = A + ((B ^ 0xff) & 0xff) + 1;
#if 0
	  if (print_state)
	       fprintf(stderr,
		       "sub: A = %d, B = %d, cin = %d, c = %d, cout = %d\n",
		       A, B, cin, c & 0xff, c > 255);
#endif
	  break;
     case 9:
	  if (cin) c = A + B;
	  else c = A + B + 1;
	  break;
     case 15:
	  if (cin) c = A - 1;
	  else c = A;
	  break;
     default:
	  fprintf(stderr, "invalid alu op %d\n", op);
	  exit(1);
     }

     if (c > 255)
	  alu_cc &= (~CARRY_MASK);
     else
	  alu_cc |= CARRY_MASK;

     if (c == 255)
	  alu_cc |= EQUAL_MASK;
     else
	  alu_cc &= (~EQUAL_MASK);

     return (c & 0xff);
}
	       
	       
	       
	  
