(define (free-identifiers exp)
  (match exp
      ((variable-reference->term~ sym) (set-singleton sym))
      ((abstraction->term~ sym term)
       (set-difference (free-identifiers term)
		       (set-singleton sym)))
      ((application->term~ term1 term2)
       (set-union (free-identifiers term1)
		  (free-identifiers term2)))
      ))

;;; Test cases for free-identifers
;; a variable reference, an application, an abstraction, and all three
(free-identifiers (parse-term 'a))
;Value: (a)
(free-identifiers (parse-term '(a b)))
;Value: (a b)
(free-identifiers (parse-term '(lambda x x)))
;Value: ()
(free-identifiers (parse-term '((lambda x (x a)) (c d))))
;Value: (a c d)

(define (substitute Tnew Isub exp)
  (match exp
      ((variable-reference->term~ sym) (if (sym=? Isub sym) Tnew sym))

      ((application->term~ operator operand)
       (application->term (substitute Tnew Isub operator)
			  (substitute Tnew Isub operand)))

      ((abstraction->term~ bound-var body)
       ;; This uses three lets so that free-id sets are only computed once
       
       (let ((free-body (free-identifiers body)))

	 ;; First, check for trivial cases
	 (if (or (sym=? bound-var Isub)
		 (not (set-member? Isub free-body)))
	     exp

	     ;; Next, see if we can get away without finding an Ifresh
	     ;; We know (1) Isub != bound-var, and Isub is a member of
	     ;; free-body
	     (let ((free-new (free-identifiers Tnew)))
	       (if (not (set-member? bound-var free-new))
		   (abstraction->term bound-var (substitute Tnew Isub body))

		   ;; The general case
		   (let ((fresh (symbol-not-in-set
				 (set-union (set-union free-body free-new)
					    (set-singleton Isub)))))
		     (abstraction->term
		      Ifresh
		      (substitute Tnew Isub
				  (substitute Ifresh bound-var body)))))))))
      ))

;;; Test cases for substitute
;; a variable reference, an application, a simple abstraction, and
;; more complex abstractions 

(unparse-term (normal-order-reduce (parse-term 'a)))
;Value: a
(unparse-term (normal-order-reduce (parse-term '(a b))))
;Value: (a b)
(unparse-term (normal-order-reduce (parse-term '((lambda x x) a))))
;Value: a
(unparse-term
 (normal-order-reduce
  (parse-term '((lambda y (y y)) (lambda x x)))))
;Value: (lambda x x)
