#include <stdio.h>
#include <tgif/tgif.h>
#include <math.h>

#ifdef PROFILE
int PointsOn, PointsOff, PointsPlotted, profile;
#endif


#define DRAWING "ps1-4.drawing"
#define THREEDEGREES 0.05236
#define PauseAndClear() ForceOutput(); WaitForButton(); Clear();

void problem1(p)
   Panel p;
{
     UsePanel(p);

     SimpleLine(10, 10, 200, 100, 1);
     PauseAndClear();

     SimpleLine(10, 10, 200, 100, 8);
     PauseAndClear();

     SymmetricLine(10, 10, 200, 100, 1);
     PauseAndClear();

     SymmetricLine(10, 10, 200, 100, 8);
     PauseAndClear();

     BresenhamLine(10, 10, 200, 100, 1);
     PauseAndClear();

     BresenhamLine(10, 10, 200, 100, 8);
     PauseAndClear();
}

void problem2(p, x0, y0, r)
   Panel p;
   int x0, y0;
   double r;
{
     int i;
     double t;

     UsePanel(p);

     for (i = 0, t = 0; i < 120; i += 1, t += THREEDEGREES)
	  SimpleLine(x0, y0, x0 + (int) rint(r*cos(t)),
		     y0 + (int) rint(r*sin(t)), 1);

     PauseAndClear();

     for (i = 0, t = 0; i < 120; i += 1, t += THREEDEGREES)
	  SymmetricLine(x0, y0, x0 + (int) rint(r*cos(t)),
			y0 + (int) rint(r*sin(t)), 1);

     PauseAndClear();

     for (i = 0, t = 0; i < 120; i += 1, t += THREEDEGREES)
	  BresenhamLine(x0, y0, x0 + (int) rint(r*cos(t)),
			y0 + (int) rint(r*sin(t)), 1);

     PauseAndClear();
}

#ifdef PROFILE
InitProfile() { PointsPlotted = PointsOn = PointsOff = 0; }

void problem3(p, x0, y0, r)
   Panel p;
   int x0, y0;
   double r;
{
     int total_on, total_off, total_plotted;
     int i;
     double t;

     profile = 1;
     
     UsePanel(p);

     total_on = total_off = total_plotted = 0;
     for (i = 0, t = 0; i < 120; i += 1, t += THREEDEGREES) {
	  InitProfile();
	  SimpleLine(x0, y0, x0 + (int) rint(r*cos(t)),
		     y0 + (int) rint(r*sin(t)), 1);
	  total_on += PointsOn;
	  total_off += PointsOff;
	  total_plotted += PointsPlotted;
     }

     printf("Simple density: %0.5f\n", total_plotted / (r*120));
     printf("Simple straightness: %0.5f\n\n", (double) total_on / (r*120));

     Clear();
     
     total_on = total_off = total_plotted = 0;
     for (i = 0, t = 0; i < 120; i += 1, t += THREEDEGREES) {
	  InitProfile();
	  SymmetricLine(x0, y0, x0 + (int) rint(r*cos(t)),
			y0 + (int) rint(r*sin(t)), 1);
	  total_on += PointsOn;
	  total_off += PointsOff;
	  total_plotted += PointsPlotted;
     }

     printf("Symmetric density: %0.5f\n", total_plotted / (r*120));
     printf("Symmetric straightness: %0.5f\n\n", (double) total_on / (r*120));
     
     Clear();

     total_on = total_off = total_plotted = 0;
     for (i = 0, t = 0; i < 120; i += 1, t += THREEDEGREES) {
	  InitProfile();
	  BresenhamLine(x0, y0, x0 + (int) rint(r*cos(t)),
			y0 + (int) rint(r*sin(t)), 1);
     	  total_on += PointsOn;
	  total_off += PointsOff;
	  total_plotted += PointsPlotted;
     }

     printf("Bresenham density: %0.5f\n", total_plotted / (r*120));
     printf("Bresenham straightness: %0.5f\n\n", (double) total_on / (r*120));
     
     Clear();

     profile = 0;
}
#endif /* PROFILE */

void problem4(p)
   Panel p;
{
     FILE *f;
     int x0, y0, x1, y1;

     Clear(p); Calibrate(p); Clear(p);
     
     f = fopen(DRAWING, "r");
     if (! f) {
	  fprintf(stderr, "Cannot open %s, exiting.\n", DRAWING);
	  exit(1);
     }

     while (! feof(f)) {
	  fscanf(f, "%d %d %d %d\n", &x0, &y0, &x1, &y1);
	  InchesToPixels(&x0, &y0);
	  InchesToPixels(&x1, &y1);
	  BresenhamLine(x0, y0, x1, y1, 1);
     }

     fclose(f);
     PauseAndClear();
}

main()
{
     Panel p;

     p = NewPanel(300, 300);

     problem1(p);
     problem2(p, 150, 150, 100.0);
#ifdef PROFILE
     problem3(p, 150, 150, 100.0);
#endif
     problem4(p);

     DoneWithPanel(p);
}


     
