/*
 * C++ class for MxN matrices.
 *
 * $Id: Matrix.C,v 1.5 90/10/21 16:17:58 bjaspan Exp Locker: bjaspan $
 * 
 * $Log:	Matrix.C,v $
 * Revision 1.5  90/10/21  16:17:58  bjaspan
 * applied diff containing Marc's changes
 * 
 * Revision 1.4  90/10/09  16:56:03  bjaspan
 * <math.h> is not needed
 * 
 * Revision 1.3  90/10/06  15:00:58  bjaspan
 * fixed printf field specification in member Print()
 * 
 * Revision 1.2  90/10/05  14:30:47  bjaspan
 * fixed Log comment character for .C
 * 
 * Revision 1.1  90/10/05  14:29:03  bjaspan
 * Initial revision
 */

/* An m x n matrix has m rows and n columns -- ie height x width */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "Matrix.hh"

Matrix *Matrix::M_NAM = Matrix::MakeNAM();

/* Constructors */

void Matrix::Init(int m, int n)
{
     M_m = m;
     M_n = n;
     nam = 0;
     d = new double[m*n];
}

Matrix::Matrix(int m, int n)
{
     Init(m, n);
}

Matrix::Matrix(int m, int n, double first ...)
{
     int i;
     va_list ap;
     
     Init(m, n);

     d[0] = first;
     va_start(ap, first);
     for (i=1; i < m*n; i += 1)
	  d[i] = va_arg(ap, double);
}

Matrix::Matrix(int m, int n, int first ...)
{
     int i;
     va_list ap;
     
     Init(m, n);

     d[0] = double(first);
     va_start(ap, first);
     for (i=1; i < m*n; i += 1)
	  d[i] = double(va_arg(ap, int));
}

Matrix::Matrix(int m, int n, int *array)
{
     int i;

     Init(m, n);
     for (i=0; i < (m*n); i += 1)
	  d[i] = double(array[i]);
}

Matrix::Matrix(int m, int n, double *array)
{
     int i;

     Init(m, n);
     for (i=0; i < (m*n); i += 1)
	  d[i] = array[i];
}

Matrix::Matrix(const Matrix& m1)
{
     int i;

     Init(m1.M_m, m1.M_n);
     for (i=0; i < (M_m*M_n); i += 1)
	  d[i] = m1.d[i];
}

Matrix::~Matrix()
{
     delete [M_m*M_n] d;
}

/* Overloaded operators */

int operator==(const Matrix& m1, const Matrix& m2)
{
     int i;

     if (m1.nam != m2.nam)
	  return 0;
     if (m1.nam == 1)
	  return 1;
     if (! m1.SameSize(m2))
	  return 0;

     for (i = 0; i < m1.M_m*m1.M_n; i++)
	  if (m1.d[i] != m2.d[i])
	       return 0;

     return 1;
}

Matrix operator+(const Matrix& m1, const Matrix& m2)
{
     if (m1.M_m != m2.M_m || m1.M_n != m2.M_n ||
	 m1 == Matrix::NAM() || m2 == Matrix::NAM())
	  return Matrix::NAM();

     Matrix m(m1.M_m, m1.M_n);
     int i;

     for (i=0; i < (m1.M_m*m1.M_n); i++)
	  m.d[i] = m1.d[i] + m2.d[i];

     return m;
}

Matrix operator*(const Matrix& m1, const Matrix& m2)
{
     if (m1.M_n != m2.M_m || m1 == Matrix::NAM() || m2 == Matrix::NAM())
	  return Matrix::NAM();

     int i, j;
     Matrix m(m1.M_m, m2.M_n);

     for (i=1; i <= m1.M_m; i += 1)
	  for (j=1; j <= m2.M_n; j += 1)
	       m(i, j) = Matrix::RowDotColumn(m1, i, m2, j);

     return m;
}

Matrix& Matrix::operator=(const Matrix& m)
{
     M_m = m.M_m;
     M_n = m.M_n;
     nam = m.nam;
     bcopy(m.d, d, M_m*M_n*sizeof(double));

     return(*this);
}

double Matrix::RowDotColumn(const Matrix& m1, int i, const Matrix& m2, int j)
{
     int a, b;
     double sum = 0;

     for (a = 1, b = 1; a <= m1.M_n; a += 1, b += 1)
	  sum += ((Matrix &) m1)(i, a) * ((Matrix &) m2)(b, j);

     return sum;
}

/* Member functions */

void Matrix::Print() const
{
     int i, j;

     if (nam) {
	  printf("NAM\n");
	  return;
     }
     
     for (i = 1; i <= M_m; i++, printf("\n"))
	  for (j = 1; j <= M_n; j++)
	       printf("%6.2f ", (*this)(i, j));
}

/* Special matrix constructors */
Matrix *Matrix::MakeNAM()
{
     Matrix *m = new Matrix(1,1);
     m->nam = 1;
     return m;
}

/* Matrix mutators */
void Matrix::Identify()	
{
     int i;
     
     if (M_m != M_n)
	  return;

     bzero((char *) d, sizeof(double)*M_m*M_n);
     for (i = 1; i <= M_m; i++)
	  (*this)(i, i) = 1;
}

     
