#ifndef __MATRIX_H__
#define __MATRIX_H__

#include <assert.h>  /* for abort() */

class Matrix {
public:
     /* Constructors */
     Matrix(int, int);
     Matrix(int, int, int *);
     Matrix(int, int, double *);
     Matrix(int, int, int ...);
     Matrix(int, int, double ...);
     Matrix(Matrix&);
     ~Matrix();

     /* Member functions */
     inline int Rows();
     inline int Cols();
     inline int SameSize(const Matrix&) const;
     inline double& operator()(int, int);
     void Print();

     /* Overloaded operators */
     friend Matrix operator+(const Matrix&, const Matrix&);
     friend Matrix operator*(const Matrix&, const Matrix&);
     friend int operator==(const Matrix&, const Matrix&);
     Matrix& operator=(const Matrix&);

     /* Special matrix creators */
     static inline const Matrix& NAM();

     /* Matrix mutators */
     void Identify();

private:
     void Init(int, int);

     static Matrix *M_NAM;
     static Matrix *MakeNAM();
     static double RowDotColumn(const Matrix&, int i, const Matrix&, int j);
     
     int M_m, M_n, nam;
     double *d;
};

inline int Matrix::Rows()
{
     return M_m;
}

inline int Matrix::Cols()
{
     return M_n;
}
	
inline double& Matrix::operator()(int m, int n)
{
     /* My kingdom for exception handling ... */
     if (m > M_m || n > M_n || m < 1 || n < 1)
	  abort();
     
     return d[M_n*(m-1) + n - 1];
}

inline const Matrix& Matrix::NAM()
{
     return *Matrix::M_NAM;
}

inline int Matrix::SameSize(const Matrix& m) const
{
     return (M_m == m.M_m && M_n == m.M_n);
}

/* DO NOT ADD ANYTHING AFTER THIS #endif */
#endif /* __MATRIX_H__ */
