; Values
(define-datatype value
  (unit->value)
  (integer->value int)
  (boolean->value bool)
  (symbol->value sym)
  (procedure->value (-> (delayed-exp-value) exp-value))
  (pair->value delayed-exp-value delayed-exp-value)
  (module->value (-> (environment) environment))
  )

(define (flk-eval exp env)
  (match exp 
    ((unit-literal->exp~) (value->exp-value (unit->value)))
    ((boolean-literal->exp~ b) (value->exp-value (boolean->value b)))
    ((integer-literal->exp~ n) (value->exp-value (integer->value n)))
    ((symbolic-literal->exp~ s) (value->exp-value (symbol->value s)))
    ((variable-reference->exp~ v) (lookup v env))
    ((abstraction->exp~ var body) (eval-abstraction var body env))
    ((vabstraction->exp~ var body) (eval-vabstraction var body env))
    ((application->exp~ rator rand) (eval-application rator rand env))
    ((conditional->exp~ test con alt) (eval-conditional test con alt env))
    ((recursion->exp~ var body) (eval-recursion var body env))
    ((pairing->exp~ left right) (eval-pair left right env))
    ((module->exp~ fields) (eval-module fields env))
    ((with->exp~ m b) (eval-with m b env))
    ))

(define eval-vabstraction
  (lambda (var body env)
    (value->exp-value
     (procedure->value
      (lambda (delayed-arg)
	(let ((v (undelay delayed-arg)))
	  (match v
	    ((value->exp-value~ _) (flk-eval body (extend-env var v env)))
	    (_ v))))
      ))))

(define eval-module
  (lambda (fields env)
    (value->exp-value
     (module->value
      (if (null? fields)
	  (lambda (current-env)
	    (lambda (sym)
	      (lookup sym current-env)))
	  
	  (match (car fields)
	    ((module-field~ sym body)
	     (lambda (current-env)
	       (extend-env sym
			   (delayed (lambda () (flk-eval body env)))
			   (with-module (eval-module (cdr fields) env)
					(lambda (m) (m current-env))))))
	    (_ (write-string "error in eval-module!\n"))))
      ))))

(define eval-with
  (lambda (module body env)
    (with-module (flk-eval module env)
		 (lambda (m)
		   (flk-eval body (m env))))))

(define with-module
  (lambda (exp-value return)
    ((ensure module->value~
	     return
	     (lambda () (error->exp-value
			 "Non-module occurs in with expression")))
     exp-value)))
