#include <stdio.h>
#include <stdlib.h>
#include "Spline.h"

extern "C" {
/* No need to #include <tgif/tgif.h> */
void Line(int, int, int, int);
}

Spline::Spline(Matrix pts_arg, int order_arg)
:pts(pts_arg)
{
     if (pts.Rows() < order_arg)
	  order = pts.Rows();
     else
	  order = order_arg; 
     knot_vec = new int[pts.Rows() + 2*order - 1];
}

Spline::Spline(const Spline& s)
:pts(s.pts)
{
     order = s.order;
     knot_vec = new int[pts.Rows() + 2*order - 1];
     bcopy(s.knot_vec, knot_vec, sizeof(int)*(pts.Rows() + 2*order - 1));
}

Spline::~Spline()
{
     delete [pts.Rows() + 2*order - 1] knot_vec;
}

void Spline::DrawOpen()
{
     MakeOpenKnot();
     Draw(pts.Rows()+1, (int) pts(1, 1), (int) pts(1, 2), (int)
	  (int) pts(pts.Rows(), 1), (int) pts(pts.Rows(), 2));
}

void Spline::DrawClosed()
{
     double x, y;

     if (pts.Rows() < 4)
	  return;
     
     MakeClosedKnot();
     Q(order-1, double(knot_vec[order-1]), x, y);
     Draw(pts.Rows()+2, (int) x, (int) y, (int) x, (int) y);
}

void Spline::Draw(const int m, const int x0, const int y0, const int
		  xn, const int yn)
{
     const double epsilon = 0.05;
     double t, x, y;
     int i, ox, oy;

     ox = x0; oy = y0;
     for (i = order-1; i <= m; i += 1) {
	  for (t = (double) knot_vec[i]; t < (double) knot_vec[i+1];
	       t += epsilon) {
	       Q(i, t, x, y);
	       Line(ox, oy, (int) x, (int) y);
	       ox = (int) x;  oy = (int) y;
	  }
     }

     Line(ox, oy, xn, yn);
}     
     

void Spline::Q(int i, double t, double& xp, double& yp)
{
     int j, idx;
     double blend;

     xp = yp = 0.0;
     for (j = i-(order-1); j <= i; j += 1) {
	  blend = B(j, order, t);

	  idx = (j % pts.Rows()) + 1;
	  xp += pts(idx, 1)*blend;
	  yp += pts(idx, 2)*blend;
     }
}

double Spline::B(int i, int k, double t)
{
     int d1, d2;
     double result = 0.0;
     
     if (k == 1) {
	  if (t >= knot_vec[i] && t < knot_vec[i+1])
	       return 1.0;
	  else
	       return 0.0;
     }

     if (d1 = knot_vec[i+k-1] - knot_vec[i])
	  result += (t - knot_vec[i]) / (double) d1 * B(i, k-1, t);

     if (d2 = knot_vec[i+k] - knot_vec[i+1])
	  result += (knot_vec[i+k] - t) / (double) d2 * B(i+1, k-1, t);

     return result;
}

void Spline::MakeOpenKnot()
   // Modifies knot_vec to draw an open spline
{
     const int n = pts.Rows();
     int i, *p;

     for (i = 0, p = knot_vec; i < order; i += 1)
	  *p++ = 0;
     for (; i < n; i += 1)
	  *p++ = i-order+1;
     for (; i < n+order; i += 1)
	  *p++ = n-order+1;
}

void Spline::MakeClosedKnot()
   // Modifies knot_vec to draw a closed spline
{
     const int n = pts.Rows();
     int i, *p;

     for (i = 0, p = knot_vec; i < n + 2*order - 1; i += 1)
	  *p++ = i;
}

