#include "Matrix.h"
#include "transform.h"
#include "CohSuthClip.h"

extern "C" {
#include <tgif/tgif.h>
}

/* C1 = foot coords, C2 = leg coords, C3 = body coords. */
Matrix C1_world = Scaler(5, 5) * Translator(120, 10);
Matrix C2_C1 = Translator(-11, 4);
Matrix C3_C2 = Translator(1, 18);

int BodyArray[] = {
     -4, -3, 1,
     -12, 1, 1,
     -12, 8, 1,
     -4, 12, 1,
     4, 12, 1,
     12, 8, 1,
     12, 1, 1,
     4, -3, 1,
};
int BodyPoints = 8;
Matrix Body(BodyPoints, 3, BodyArray);

int LegArray[] = {
     -1, -2, 1,
     -3, 16, 1,
     5, 16, 1,
     2, -2, 1,
};
int LegPoints = 4;
Matrix Leg(LegPoints, 3, LegArray);

int FootArray[] = {
     -11, 4, 1,
     0, 2, 1,
     0, 0, 1,
     -11, 0, 1,
};
int FootPoints = 4;
Matrix Foot(FootPoints, 3, FootArray);

void Pose(int x, int y, int foot_angle, int ankle_angle, int hip_angle)
{
     /* These equations merely define the interpretation of */
     /* ankle_angle and hip_angle; they have nothing to do with the */
     /* transformations.  (0, 90, 90) means "standing straight". */
     const int leg_angle = ankle_angle - 90;
     const int body_angle = hip_angle - 90;
      
     const Matrix FootTrans = Rotator(foot_angle, 0, 0)*Translator(x, y)*
	  C1_world; 
     const Matrix LegTrans = Rotator(leg_angle, 0, 0)*C2_C1*FootTrans;
     const Matrix BodyTrans = Rotator(body_angle, 0, 0)*C3_C2*LegTrans;
     
     CohSuthClipDraw(Foot*FootTrans);
     CohSuthClipDraw(Leg*LegTrans);
     CohSuthClipDraw(Body*BodyTrans);
}
