
;;; Search functions for 6.034 PS 3

(in-package 'user)

(defvar *graph*
	;; Although the links are bidirectional, we list them once and let
	;; extend handle the two-way nature.
	'((S A 100) (S B 60) (S D 20)
	  (A B 40) (A C 10) (A E 70)
	  (B C 10) (B D 30)))

; Basic Branch and Bound Search

(defun search-bb (start-node finish-node graph)
  (aux-search-bb (make-queue (make-path start-node)) finish-node graph))

; (search-bb 's 'e *graph*) should give you the answer

(defun aux-search-bb (queue finish-node graph)
  (format t "~% Sorted Queue is ~s" queue)
  (if queue
      (if (eq finish-node (last-node-in-path (queue-head queue)))
	  (let ((path (queue-head queue)))
	    (cons (path-length path) (reverse (path-in-reverse path))))
	  (aux-search-bb (add-and-sort-queue 
			   (extend (queue-head queue) graph)
			   (queue-tail queue))
			 finish-node graph))))

;;; Problem 1.
(defun search-bb+dp-sort (start-node finish-node graph)
  (aux-search-bb+dp-sort (make-queue (make-path start-node))
			 finish-node graph))

(defun search-bb+dp-merge (start-node finish-node graph)
  (aux-search-bb+dp-merge (make-queue (make-path start-node))
			  finish-node graph)) 

(defun aux-search-bb+dp-sort (queue finish-node graph)
  (format t "~% Sorted Queue is ~s" queue)
  (if queue
      (if (eq finish-node (last-node-in-path (queue-head queue)))
	  (let ((path (queue-head queue)))
	    (cons (path-length path) (reverse (path-in-reverse path))))
	  (aux-search-bb+dp-sort (dynamic-prune (add-and-sort-queue 
					 (extend (queue-head queue) graph)
					 (queue-tail queue)))
			 finish-node graph))))

(defun aux-search-bb+dp-merge (queue finish-node graph)
  (format t "~% Sorted Queue is ~s" queue)
  (if queue
      (if (eq finish-node (last-node-in-path (queue-head queue)))
	  (let ((path (queue-head queue)))
	    (cons (path-length path) (reverse (path-in-reverse path))))
	  (aux-search-bb+dp-merge (dynamic-prune (merge-queues
					    (extend (queue-head queue) graph)
					    (queue-tail queue)))
			 finish-node graph))))

(defun dynamic-prune (queue)
  ;; Prunes all but the cheapest path to every given node.  Assumes
  ;; the list is sorted, which implies the cheapest path to a node
  ;; is the first path to a node.
  (dynamic-prune-iter queue nil))

(defun dynamic-prune-iter (queue keepers)
  (if (null queue)
      keepers
    (let ((path (car queue)))
      (if (not (some-path-ends-at (last-node-in-path path) keepers))
	  (setq keepers (append keepers (list path))))
      (dynamic-prune-iter (cdr queue) keepers))))

(defun some-path-ends-at (state queue)
  (cond ((null queue) nil)
	((eq (last-node-in-path (car queue)) state) t)
	(t (some-path-ends-at state (cdr queue)))))

(defun merge-queues (new-paths queue)
  (merge 'list queue (sort new-paths #'shorterp) #'shorterp))

;;; End of problem 1.

(defun add-and-sort-queue (new-paths queue)
  (sort (append new-paths queue) #'shorterp))

(defun shorterp (path1 path2)
  (< (path-length path1) (path-length path2)))

(defun extend (path graph)
  (if graph
      (let* ((last-node (last-node-in-path path))
	     (arc (next-arc graph))
	     (source (source-node arc))
	     (destination (destination-node arc))
	     (cost (cost arc)))
	; checks both source and destination because
	; the graph is meant to be undirected.
	(cond ((eq last-node source) 
	       (if (member destination path)
		   (extend path (rest-of-arcs graph))
		   (cons (add-node-to-path destination cost path)
			 (extend path (rest-of-arcs graph)))))
	      ((eq last-node destination)
	       (if (member source path)
		   (extend path (rest-of-arcs graph))
		   (cons (add-node-to-path source cost path)
			 (extend path (rest-of-arcs graph)))))
	      (t (extend path (rest-of-arcs graph)))))))

; Many of the functions in the following data abstractions are 
; best written as macros but to avoid confusing you further, they 
; are given as functions.

; a graph is represented as a list of arcs.

(defun next-arc (graph) (car graph))
(defun rest-of-arcs (graph) (cdr graph))

; an arc is a triplet: (source-node destination-node cost)

(defun source-node (arc) (car arc))
(defun destination-node (arc) (cadr arc))
(defun cost (arc) (caddr arc))

;;; A (partial) path is represented as a list with path length as first
;;; element and the rest of the list is the path reversed.

; Some Data Abstractions 

(defun make-path (start-node) (list 0.0 start-node))

(defun path-length (path) (car path))

(defun path-in-reverse (path) (cdr path))

(defun last-node-in-path (path) (car (path-in-reverse path)))

; The queue (of partial paths) is represented as a list.

(defun add-node-to-path (node cost path)
  (cons (+ (path-length path) cost)
	(cons node (path-in-reverse path))))

(defun make-queue (q-element) (list q-element))

(defun queue-head (queue) (car queue))

(defun queue-tail (queue) (cdr queue))
