#include <stdio.h>
#include <math.h>
#include "Spline.h"

extern "C" {
#include <tgif/tgif.h>
}

void PlotToScale(int, int, int);
void MarkPoints(Matrix&);
void reshape(Matrix&, Event, int);

void problem2()
{
     Matrix empty(0, 2), m(0, 2), p(1, 2);
     Spline *s = new Spline(p, 4);
     Event e;
     int open, action;

     Clear();
     ForceOutput();
     
     while (1) {
	  printf("button 1 = non-periodic spline, button 2 = quit, "
		 "button 3 = periodic spline\n");
	  e = GetEvent();
	  action = EventButton(e);
	  if (action == 1)
	       open = 1;
	  else if (action == 2)
	       break;
	  else if (action == 3) {
	       open = 0;
	       action = 1;
	  }
	  
	  m = empty;
	  
	  printf("button 1 = continue spline, button 2 = reshape spline, "
		 "button 3 = finish spline\n");
	  while (1) {
	       if (action == 1) {
		    p(1, 1) = EventX(e);
		    p(1, 2) = EventY(e);
		    m = m & p;

		    delete s;
		    s = new Spline(m, 4);
		    Clear();
		    MarkPoints(m);
		    if (open)
			 s->DrawOpen();
		    else
			 s->DrawClosed();
		    ForceOutput();
	       } else if (action == 2)
		    reshape(m, e, open);
	       else if (action == 3)
		    break;

	       e = GetEvent();
	       action = EventButton(e);
	  }
     }
     
     delete s;
}

inline double sqr(double x) { return x*x; }
inline double distance(double x1, double y1, double x2, double y2)
{
     return (sqrt(sqr(x2-x1)+sqr(y2-y1)));
}

void reshape(Matrix& m, Event e, int open)
{
     int idx, i, Ex = EventX(e), Ey = EventY(e);
     double d, min_d = -1.0;

     // Find the "grabbed" point
     for (i = 1; i <= m.Rows(); i++) {
	  d = distance((double) Ex, (double) Ey, m(i, 1), m(i, 2));
	  if (min_d == -1.0 || d < min_d) {
	       idx = i;
	       min_d = d;
	  }
     }

     printf("button 1 = place point, button 3 = quit\n");
     while (1) {
	  e = GetEvent();
	  if (EventButton(e) == 1) {
	       m(idx, 1) = (double) EventX(e);
	       m(idx, 2) = (double) EventY(e);
	       Spline s(m, 4);
	       Clear();
	       MarkPoints(m);
	       if (open)
		    s.DrawOpen();
	       else
		    s.DrawClosed();
	  }
	  else
	       break;
     }
}

void MarkPoints(Matrix& m)
{
     int i;

     for (i = 1; i <= m.Rows(); i += 1)
	  PlotToScale(int(m(i, 1)), int(m(i, 2)), 5);
}

/* Stolen from problem set 1 */
void PlotToScale(int px, int py, int s)
{
     int x, y, l, r;

     /* s odd --> plot s/2 pixels on either side; */
     /* s even --> plot (s/2)-1 to the left, s/2 to the right */
     l = r = s / 2;
     if (! (s % 2)) l -= 1;

     for (x = px - l; x <= px + r; x += 1)
	  for (y = py - l; y <= py + r; y += 1)
	       Plot(x, y);
}
