; The rest of this file is for Problem 2 of Problem Set 3
; It also uses the queue abstraction above.

; The representations used in the following routines are:

; A position in the tray is represented by a pair (x-location . y-location)
; A state (board position) looks like 
;   (position-of-the-empty-tile . puzzle-array)
; A puzzle-array is a 3X3 array storing the numbered tiles in the respective 
; board positions.  The empty tile is represented as the symbol *.
; The position of the empty tile is redundantly kept for convenience.
; A (partial) path looks like
;  (accumulated-cost-so-far estimated-cost . path-in-reverse)
; in which path-in-reverse contains states. 

; note that :initial-contents fills the array up column-first
; that is, solution is correctly stored below.  Take a look by
; doing (print-state goal-state).

(defun search-A* (start-state finish-state)
  (aux-search-A* (make-queue (empty-path start-state)) finish-state))

(defun aux-search-A* (queue finish-state)
  (if debug
      (format t "~% Sorted Queue is length ~s, queue ~% ~s" 
	      (length queue) queue)
      (format t "~% Sorted Queue is length ~s" (length queue)))
  (if queue
      (let ((path (queue-head queue)))
	(if (same-state-p finish-state (last-state-in-path path))
	    (print-path path)
	    (aux-search-A* (dynamic-prune (merge-queues
					   (extend-state path)
					   (queue-tail queue)))
			   finish-state)))))

(defun extend-state (path)
  ;; Returns an unsorted list of all the new paths reachable from
  ;; path.  Each element will contain a new state consed on to the
  ;; front of (path-in-reverse-A* path) plus an updated accumulated
  ;; cost and estimated cost.
  (let ((states (next-states (last-state-in-path path))))
    (mapcar #'(lambda (state)
		(make-new-path state path))
	    states)))

(defun estimate-cost (state)
  ;; Returns the number of elements of state that are not in the
  ;; correct position.
  (let* ((state-array (cdr state))
	 (dim (array-dimensions state-array))
	 (estimate 0))
    (dotimes (i (car dim))
      (dotimes (j (second dim))
	(if (not (eq (aref state-array i j) (aref *solution* i j)))
	    (setq estimate (1+ estimate)))
	))
    estimate))

(defun estimate-manhattan-cost (state)
  ;; Returns the sum of the distances each piece is from its correct
  ;; position using the //Manhattan Distance// formula.
  (let* ((state-array (cdr state))
	 (dim (array-dimensions state-array))
	 (estimate 0))
    (dotimes (i (car dim))
      (dotimes (j (second dim))
	(let ((elem (aref state-array i j)))
	  (if (eq elem '*) 
	      (setq elem 0))
	  (setq estimate 
		(+ estimate (manhattan-distance 
			     (aref *correct-position-array* elem)
			     (cons i j)))))
	))
    estimate))

(defun manhattan-distance (p1 p2)
  ;; Returns the //Manhattan Distance// between the two points p1 and
  ;; p2, which should be represented as (x . y).
  (+ (abs (- (car p1) (car p2)))
     (abs (- (cdr p1) (cdr p2)))))

(defun empty-path (state)
  ;; Make a path to state from nowhere.
  (cons 0
	(cons 0
	      (cons state nil))))

(defun make-new-path (state path)
  ;; Makes a new path to state from path.
  (cons (1+ (cost-so-far path))
	(cons (estimate-cost state)
	      (cons state (path-in-reverse-A* path)))))

(defun copy-array (array)
  ; used by Next-States
  (let* ((dim (array-dimensions array))
	 (new (make-array dim)))
    (dotimes (i (car dim))
      (dotimes (j (second dim))
        (setf (aref new i j) (aref array i j))))
    new))
  

(defvar *solution*
	(make-array (list 3 3) :initial-contents '((1 8 7) 
						   (2 * 6)
						   (3 4 5)))
  "An array showing the desired board position.")

(defvar *goal-state* (cons '(1 . 1) *solution*)
  "The desired goal state.")

(defvar *P1*
	(cons '(1 . 2)	
	      (make-array (list 3 3) :initial-contents '((2 1 7) 
							 (8 6 *)
							 (3 4 5))))
  "An initial state, a board position.  Test Case 1.")

(defvar *P2*
	(cons '(2 . 1)
	      (make-array (list 3 3) :initial-contents '((2 7 6) 
							 (4 3 1)
							 (8 * 5))))
  "An initial state, a board position.  Test Case 2.")

; This array is used in conjunction with Next-States to produce
; the set of board positions reachable in one step from a given 
; board position.

(defvar *possible-moves-array*
	(make-array
	  (list 3 3) 
	  :initial-contents
	  '((((1 . 0) (0 . 1))	        ; * in (0 0)
	     ((1 . 1) (0 . 0) (0 . 2))	; * in (0 1)
	     ((1 . 2) (0 . 1)))	        ; * in (0 2)
	    (((0 . 0) (2 . 0) (1 . 1))	; * in (1 0)
	     ((0 . 1) (2 . 1) (1 . 0) (1 . 2)) 
	                                ; * in (1 1)
	     ((0 . 2) (2 . 2) (1 . 1)))	; * in (1 2)
	    (((1 . 0) (2 . 1))	        ; * in (2 0)
	     ((1 . 1) (2 . 0) (2 . 2))	; * in (2 1)
	     ((1 . 2) (2 . 1)))))       ; * in (2 2)
  "Explicitly stores the positions of the tiles that can move 
into the indexed position.")

(defvar *correct-position-array*
  (make-array (list 9)
	      :initial-contents
	      '((1 . 1) (0 . 0) (1 . 0) 
		(2 . 0) (2 . 1) (2 . 2) 
		(1 . 2) (0 . 2) (0 . 1)))
  "Explicity stores the correct positions of each tile.")

; some useful supporting functions

(defun copy-array (array)
  ; used by Next-States
  (let* ((dim (array-dimensions array))
	 (new (make-array dim)))
    (dotimes (i (car dim))
      (dotimes (j (second dim))
        (setf (aref new i j) (aref array i j))))
    new))

; If you use the representation of the state as suggested, this routine 
; prints out the board position nicely for ease of visualization.

(defun print-state (state)
  (let ((puzzle-array (cdr state)))
    (format t "~% ------------- ")
    (format t "~% | ~s | ~s | ~s |" (aref puzzle-array 0 0) 
	    (aref puzzle-array 1 0) (aref puzzle-array 2 0))
    (format t "~% ------------- ")
    (format t "~% | ~s | ~s | ~s |" (aref puzzle-array 0 1) 
	    (aref puzzle-array 1 1) (aref puzzle-array 2 1))
    (format t "~% ------------- ")
    (format t "~% | ~s | ~s | ~s |"  (aref puzzle-array 0 2)
	    (aref puzzle-array 1 2) (aref puzzle-array 2 2))
    (format t "~% ------------- ")))

(defun same-state-p (state1 state2)
  (and (equal (car state1) (car state2))
       (let* ((array1 (cdr state1))
	      (array2 (cdr state2)))
	 (block done
    	   (dotimes (i 3 t)
    	     (dotimes (j 3)
  	       (if (not (equal (aref array1 i j) (aref array2 i j)))
		   (return-from done nil))))))))

(defun cost-so-far (path) (first path)) 

(defun estimated-cost (path) (second path)) 

(defun last-state-in-path (path) (third path))

(defun path-in-reverse-A* (path) (cddr path))

; Given the current state, Next-States returns a list of possible next
; states.  You probably don't need to understand how this procedure
; works if you use the representations suggested.  You just need to know
; that given an appropriately represented current state (indicated
; earlier), Next-States return you a list of legal next states that the
; current state can get to.

(defun next-states (current-state)
  ; returns a list of new possible states
  (let* ((position-of-* (car current-state))
	 (puzzle-array (cdr current-state))
	 (new-x (car position-of-*))
	 (new-y (cdr position-of-*))
	 (moves (aref *possible-moves-array* new-x new-y)))
    ; moves contains opositions of tiles that can move into
    ; the current empty tile position
    (mapcar 
      #'(lambda (move) 
	  (let ((from-x (car move))
		(from-y (cdr move))
		(new-array (copy-array puzzle-array)))
	    ; move the numbered tile to the position originally occupied 
	    ; by the empty tile
	    (setf (aref new-array new-x new-y)
		  (aref puzzle-array from-x from-y))
	    ; indicate that the empty tile is now moved
	    (setf (aref new-array from-x from-y) '*)
	    ; move is the new position of the empty tile
	    (cons move new-array)))
      moves)))

(defun print-path (path)
  (format t "~% Total Cost = ~S" (cost-so-far path))
  (format t "~% Path is ")
  (dolist (x (reverse (path-in-reverse-A* path)))
    (format t "~%    | ")
    (format t "~%    V ")
    (print-state x)))

(defun make-queue (q-element) (list q-element))
(defun queue-head (queue) (car queue))
(defun queue-tail (queue) (cdr queue))

;;; Dynamic programming pruning stuff

(defun dynamic-prune (queue)
  ;; Prunes all but the cheapest path to every given node.  Assumes
  ;; the list is sorted, which implies the cheapest path to a node
  ;; is the first path to a node.
  (dynamic-prune-iter queue nil))

(defun dynamic-prune-iter (queue keepers)
  (if (null queue)
      keepers
    (let ((path (car queue)))
      (if (not (some-path-ends-at (last-state-in-path path) keepers))
	  (setq keepers (append keepers (list path))))
      (dynamic-prune-iter (cdr queue) keepers))))

(defun some-path-ends-at (state queue)
  (cond ((null queue) nil)
	((same-state-p (last-state-in-path (car queue)) state) t)
	(t (some-path-ends-at state (cdr queue)))))

(defun merge-queues (new-paths queue)
  (merge 'list queue (sort new-paths #'lower-cost-estimate) 
	 #'lower-cost-estimate))

(defun lower-cost-estimate (p1 p2)
  (< (+ (cost-so-far p1) (estimated-cost p1))
     (+ (cost-so-far p2) (estimated-cost p2))))

