#include <stdio.h>
#include "transform.h"
#include "Matrix.h"
#include "Hedron.h"
#include "CohSuthClip.h"

extern void problem2();

extern "C" {
#include <tgif/tgif.h>
}

int Shape1Data[] = {
     50, 50, -50, 1,
     50, -50, -50, 1,
     -50, -50, -50, 1,
     -50, 50, -50, 1,
     50, 50, 50, 1,
     50, -50, 50, 1,
     -50, -50, 50, 1,
     -50, 50, 50, 1,
};
int Shape1Points = 8;
int Shape1Edges[] = {
     1, 2,
     2, 3,
     3, 4,
     4, 1,

     5, 6,
     6, 7,
     7, 8,
     8, 5,

     1, 5,
     2, 6,
     3, 7,
     4, 8,
};
int Shape1NEdges = 12;

void DrawBox(int xl, int yl, int xh, int yh)
{
     Line(xl, yl, xh, yl);
     Line(xh, yl, xh, yh);
     Line(xh, yh, xl, yh);
     Line(xl, yh, xl, xl);
}

void DrawPolyhedron(Hedron& h)
{
     int i;

     for (i = 0; i < h.NumEdges(); i += 1)
	  Line(int(h.VertexCoord(2*i, 1)),
	       int(h.VertexCoord(2*i, 2)),
	       int(h.VertexCoord(2*i+1, 1)),
	       int(h.VertexCoord(2*i+1, 2)));
}

void problem1(int xl, int yl, int xh, int yh)
{
     void Pose(int, int, int, int, int);
     int x, y, f, a, h;

     CohSuthSetClip(0.05*(xh-xl), 0.05*(yh-yl), 0.95*(xh-xl), 0.95*(yh-yl));
     DrawBox(int(0.05*(xh-xl)), int(0.05*(yh-yl)),
	     int(0.95*(xh-xl)), int(0.95*(yh-yl)));
     Pose(0, 0, 0, 90, 90);
     ForceOutput();

     while (! feof(stdin)) {
	  printf("Pose: ");  fflush(stdout);
	  scanf("%d %d %d %d %d", &x, &y, &f, &a, &h);
	  Clear();
	  DrawBox(int(0.05*(xh-xl)), int(0.05*(yh-yl)),
		  int(0.95*(xh-xl)), int(0.95*(yh-yl)));
	  Pose(x, y, f, a, h);
	  ForceOutput();
     }

     /* Reset the EOF condition */
     clearerr(stdin);
     printf("\n");
}

void problem3()
{
     double a, b, c;
     char z;
     Matrix m(Shape1Points, 4, Shape1Data);
     Hedron h(m, Shape1NEdges, Shape1Edges);
     Matrix CTM(4, 4);

     CTM.Identify();

     while (1) {
	  Clear();
	  DrawPolyhedron(h*CTM);
	  ForceOutput();

	  if (scanf("%1s", &z) == EOF)
	       break;

	  switch (z) {
	  case 't':
	  case 'T':
	       scanf("%f %f %f", &a, &b, &c);
	       CTM = CTM * Translator(a, b, c);
	       break;
	  case 's':
	  case 'S':
	       scanf("%f %f %f", &a, &b, &c);
	       CTM = CTM * Scaler(a, b, c);
	       break;
	  case 'r':
	  case 'R':
	       scanf("%1s %f", &z, &a);
	       switch (z) {
	       case 'X':
	       case 'x':
		    CTM = CTM * Rotator_X(a);
		    break;
	       case 'Y':
	       case 'y':
		    CTM = CTM * Rotator_Y(a);
		    break;
	       case 'Z':
	       case 'z':
		    CTM = CTM * Rotator_Z(a);
		    break;
	       }
	       break;
	  case 'i':
	  case 'I':
	       CTM.Identify();
	       break;
	  }
     }

     /* Reset the EOF condition */
     clearerr(stdin);
     printf("\n");
}




main()
{
     Panel p;

     p = NewPanel(300, 300);
     SetFunction(DRAW);
     UsePanel(p);

     problem1(0, 0, 300, 300);
     problem2();
     problem3();

     DoneWithPanel(p);
     return 0;
}
