#include "MatrixUtil.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

Matrix CrossProd(Matrix& A, Matrix& B)
   // Returns the cross product AxB, or NAM if A and B are not both
   // 1x3 matrices (vectors).
{
     if (A.Rows() != 1 || B.Rows() != 1 || A.Cols() != 3 || B.Cols()
	 != 3) return Matrix::NAM();

     Matrix C(1, 3,
	      A(1,2)*B(1,3)-A(1,3)*B(1,2),
	      A(1,3)*B(1,1)-A(1,1)*B(1,3),
	      A(1,1)*B(1,2)-A(1,2)*B(1,1));
     return C;
}

double MatrixMag(Matrix& m, int i=1)
   // Returns the magnitude of the ith row of m or -1.0.
{
     if (i > m.Rows())
	  return (-1.0);

     int j;
     double tot = 0.0;

     for (j = 1; j <= m.Cols(); j++)
	  tot += m(i, j)*m(i, j);

     return sqrt(tot);
}

Matrix Unitize(Matrix& m, int i=1)
   // Returns a unit vector (1xN matrix) of the ith row of m, or NAM.
{
     if (i > m.Rows())
	  return Matrix::NAM();

     Matrix n(1, m.Cols());
     double mag = MatrixMag(m, i);
     int j;

     if (mag <= 0) {
	  fprintf(stderr, "Magnitude is zero. Lose!\n");
	  exit(1);
     }	  

     for (j = 1; j <= m.Cols(); j++)
	  n(1, j) = m(i, j) / mag;

     return n;
}

	      
