#include "Matrix.h"
#include "Hedron.h"
#include "transform.h"

extern "C" {
#include <tgif/tgif.h>
}

/* C1 = foot coords, C2 = leg coords, C3 = body coords. */
Matrix C1_world = Scaler(5, 5, 5);
Matrix C2_C1 = Translator(-11, 4, 0);
Matrix C3_C2 = Translator(1, 18, 0);

int BodyArray[] = {
     -4, -3, 4, 1,
     -12, 4, 4, 1,
     -12, 8, 4, 1,
     -4, 12, 4, 1,
     4, 12, 4, 1,
     12, 8, 4, 1,
     12, 4, 4, 1,
     4, -3, 4, 1,
     -4, -3, -4, 1,
     -12, 4, -4, 1,
     -12, 8, -4, 1,
     -4, 12, -4, 1,
     4, 12, -4, 1,
     12, 8, -4, 1,
     12, 4, -4, 1,
     4, -3, -4, 1,
};
int BodyPoints = 16;
Matrix BodyM(BodyPoints, 4, BodyArray);
int BodyEdges[] = {
     1, 2,
     2, 3,
     3, 4,
     4, 5,
     5, 6,
     6, 7,
     7, 8,
     8, 1,
     9, 10,
     10, 11,
     11, 12,
     12, 13,
     13, 14,
     14, 15,
     15, 16,
     16, 9,
     1, 9,
     2, 10,
     3, 11,
     4, 12,
     5, 13,
     6, 14,
     7, 15,
     8, 16,
};
Hedron Body(BodyM, 24, BodyEdges);

int LegArray[] = {
     -1, -2, 2, 1,
     -3, 16, 2, 1,
     5, 16, 2, 1,
     2, -2, 2, 1,
     -1, -2, -2, 1,
     -3, 16, -2, 1,
     5, 16, -2, 1,
     2, -2, -2, 1,
};
int LegPoints = 8;
Matrix LegM(LegPoints, 4, LegArray);
int LegEdges[] = {
     1, 2,
     2, 3,
     3, 4,
     4, 1,
     5, 6,
     6, 7,
     7, 8,
     8, 5,
     1, 5,
     2, 6,
     3, 7,
     4, 8,
};
Hedron Leg(LegM, 12, LegEdges);

int FootArray[] = {
     -11, 4, 2, 1,
     0, 2, 2, 1,
     0, 0, 2, 1,
     -11, 0, 2, 1,
     -11, 4, -2, 1,
     0, 2, -2, 1,
     0, 0, -2, 1,
     -11, 0, -2, 1,
};
int FootPoints = 8;
Matrix FootM(FootPoints, 4, FootArray);
int FootEdges[] = {
     1, 2,
     2, 3,
     3, 4,
     4, 1,
     5, 6,
     6, 7,
     7, 8,
     8, 5,
     1, 5,
     2, 6,
     3, 7,
     4, 8,
};
Hedron Foot(FootM, 12, FootEdges);

Hedron Pose(double x, double y, double z, double foot_orient, double
	    foot_angle, double ankle_angle, double hip_angle)
{
     const Matrix FootOrient = Rotator_Y(foot_orient)*
	  Translator(x, y, z)*C1_world; 
     const Matrix FootTrans = Rotator_Z(foot_angle)*FootOrient;
     const Matrix LegTrans = Rotator_Z(ankle_angle)*C2_C1*FootTrans;
     const Matrix BodyTrans = Rotator_Z(hip_angle)*C3_C2*LegTrans;

     return Foot*FootTrans + Leg*LegTrans + Body*BodyTrans;
}
