#include <stdio.h>
#include <stdlib.h>
#include "Matrix.h"
#include "MatrixUtil.h"
#include "Hedron.h"
#include "transform.h"
#include "Clipping.h"

const int WIDTH = 300;
const int HEIGHT = 300;

extern "C" {
#include <tgif/tgif.h>
}

int debug = 0;

Hedron Pose(double, double, double, double, double, double, double);

void PrintMatrix(char *msg, Matrix &m)
{
     printf("%s\n", msg);
     m.Print();
}

Hedron ClipHedron(Hedron& h, double xmin, double xmax, double ymin,
		  double ymax, double hither, double yon)
{
     Matrix clipper(2, 3,
		    xmin, ymin, hither,
		    xmax, ymax, yon);

     PrintMatrix("clipper", clipper);

     return h.Clip(clipper);
}


void DrawPolyhedron(Hedron& h)
{
     int i;

     for (i = 0; i < h.NumEdges(); i += 1)
          Line(int(h.VertexCoord(2*i, 1)),
               int(h.VertexCoord(2*i, 2)),
               int(h.VertexCoord(2*i+1, 1)),
               int(h.VertexCoord(2*i+1, 2)));
}

Matrix PreClipTransform(Matrix& Peye_w, Matrix& Pfix_w, Matrix& UP_w)
{
     Matrix xhat(1, 3), yhat(1, 3), zhat(1, 3);
     Matrix View(1, 3);
     View = Peye_w - Pfix_w;

     Matrix ViewHat(Unitize(View));
     xhat = Unitize(CrossProd(UP_w, ViewHat));
     yhat = CrossProd(ViewHat, xhat);
     zhat = -1.0*ViewHat;

     Matrix Rot(4, 4,
		xhat(1, 1), yhat(1, 1), zhat(1, 1), 0.0, 
		xhat(1, 2), yhat(1, 2), zhat(1, 2), 0.0, 
		xhat(1, 3), yhat(1, 3), zhat(1, 3), 0.0, 
		0.0,        0.0,        0.0,        1.0);
     
     Matrix V(Translator(-Peye_w(1, 1), -Peye_w(1, 2), -Peye_w(1, 3)) * Rot);
     if (debug) {
	  PrintMatrix("View:", View);
	  PrintMatrix("Rotation matrix:", Rot);
	  PrintMatrix("V:", V);
     }

     return V;
}

Hedron Perspective(const Hedron& h1, const double d)
{
     Hedron h2(h1);
     double v;
     int i;

     for (i = 1; i <= h2.Vertices().Rows(); i++) {
	  if (h2.Vertices()(i, 3) == 0.0) {
	       h2.Vertices()(i, 1) = HUGE;
	       h2.Vertices()(i, 2) = HUGE;
	  }
	  else {
	       v = h2.Vertices()(i, 3) / d;
	       h2.Vertices()(i, 1) /= v;
	       h2.Vertices()(i, 2) /= v;
	  }
     }

     return h2;
}

void problem2()
{
     Matrix CTM(4, 4);
     Matrix Peye(1, 3, 0, 0, -250);
     Matrix UP(1, 3, 0, 1, 0);
     Matrix Pfix(1, 3, 0, 0 ,0);
     Matrix V(4, 4);
     double border = 20.0, hither = 10.0, yon = 500.0;
     double x, y, z, foot_orient, foot_angle, ankle_angle, hip_angle;
     double a, b, c;
     char q;

     CTM.Identify();

     /* Initial Setup */
     x = y = z = 0.0;
     foot_orient = foot_angle = ankle_angle = hip_angle = 0.0;

     while (1) {
	  Clear();

	  double PerMag = MatrixMag(Peye - Pfix);
	  V = PreClipTransform(Peye, Pfix, UP);

	  Hedron world(Pose(x, y, z, foot_orient, foot_angle,
			    ankle_angle, hip_angle) * CTM);
	  Hedron eye(world*V);
	  Hedron perspect(Perspective(eye, PerMag));
	  Hedron clipped(ClipHedron(perspect,
				    -(WIDTH/2.0-border),WIDTH/2.0-border,
				    -(HEIGHT/2.0-border),HEIGHT/2.0-border,
				    hither, yon));

	  DrawPolyhedron(clipped * Translator(150, 150, 0));
	  ForceOutput();

	  if (scanf("%1s", &q) == EOF)
	       break;

	  switch (q) {
	  case 'p':
	  case 'P':
	       scanf("%f %f %f %f", &foot_orient, &foot_angle,
		     &ankle_angle, &hip_angle);
	       break;
	  case 'c':
	  case 'C':
	       scanf("%f %f %f", &a, &b, &c);
	       border = a; hither = b; yon = c;
	       break;
	  case 'e':
	  case 'E':
	       scanf("%f %f %f", &a, &b, &c);
	       Peye(1, 1) = a; Peye(1, 2) = b; Peye(1, 3) = c;
	       break;
	  case 'f':
	  case 'F':
	       scanf("%f %f %f", &a, &b, &c);
	       Pfix(1, 1) = a; Pfix(1, 2) = b; Pfix(1, 3) = c;
	       break;
	  case 'u':
	  case 'U':
	       scanf("%f %f %f", &a, &b, &c);
	       UP(1, 1) = a; UP(1, 2) = b; UP(1, 3) = c;
	       break;
	  case 't':
	  case 'T':
	       scanf("%f %f %f", &a, &b, &c);
	       CTM = CTM * Translator(a, b, c);
	       break;
	  case 's':
	  case 'S':
	       scanf("%f %f %f", &a, &b, &c);
	       CTM = CTM * Scaler(a, b, c);
	       break;
	  case 'r':
	  case 'R':
	       scanf("%1s %f", &q, &a);
	       switch (q) {
	       case 'X':
	       case 'x':
		    CTM = CTM * Rotator_X(a);
		    break;
	       case 'Y':
	       case 'y':
		    CTM = CTM * Rotator_Y(a);
		    break;
	       case 'Z':
	       case 'z':
		    CTM = CTM * Rotator_Z(a);
		    break;
	       }
	       break;
	  case 'i':
	  case 'I':
	       CTM.Identify();
	       break;
	  }
     }

     /* Reset the EOF condition */
     clearerr(stdin);
     printf("\n");
}
     
main()
{
     Panel p;

     p = NewPanel(WIDTH, HEIGHT);
     SetFunction(DRAW);
     UsePanel(p);

     problem2();

     DoneWithPanel(p);
     return 0;
}
