#include <stdio.h>
#include <stdlib.h>
#include "Matrix.h"
#include "MatrixUtil.h"
#include "Hedron.h"
#include "transform.h"
#include "Clipping.h"

const int WIDTH = 300;
const int HEIGHT = 300;

extern "C" {
#include <tgif/tgif.h>
}

int debug = 0;

void PlotToScale(int px, int py, int s)
{
     int x, y, l, r;

     /* s odd --> plot s/2 pixels on either side; */
     /* s even --> plot (s/2)-1 to the left, s/2 to the right */
     l = r = s / 2;
     if (! (s % 2)) l -= 1;

     for (x = px - l; x <= px + r; x += 1)
	  for (y = py - l; y <= py + r; y += 1)
	       Plot(x, y);
}

void PrintMatrix(char *msg, Matrix &m)
{
     printf("%s\n", msg);
     m.Print();
}

Hedron ClipHedron(Hedron& h, double xmin, double xmax, double ymin,
		  double ymax, double hither, double yon)
{
     Matrix clipper(2, 3,
		    xmin, ymin, hither,
		    xmax, ymax, yon);

     return h.Clip(clipper);
}


void DrawPolyhedron(Hedron& h)
{
     int i;

     for (i = 0; i < h.NumEdges(); i += 1)
          Line(int(h.VertexCoord(2*i, 1)),
               int(h.VertexCoord(2*i, 2)),
               int(h.VertexCoord(2*i+1, 1)),
               int(h.VertexCoord(2*i+1, 2)));
}

Matrix PreClipTransform(Matrix& Peye_w, Matrix& Pfix_w, Matrix& UP_w)
{
     Matrix xhat(1, 3), yhat(1, 3), zhat(1, 3);
     Matrix View(1, 3);
     View = Peye_w - Pfix_w;

     Matrix ViewHat(Unitize(View));
     xhat = Unitize(CrossProd(UP_w, ViewHat));
     yhat = CrossProd(ViewHat, xhat);
     zhat = -1.0*ViewHat;

     Matrix Rot(4, 4,
		xhat(1, 1), yhat(1, 1), zhat(1, 1), 0.0, 
		xhat(1, 2), yhat(1, 2), zhat(1, 2), 0.0, 
		xhat(1, 3), yhat(1, 3), zhat(1, 3), 0.0, 
		0.0,        0.0,        0.0,        1.0);
     
     Matrix V(Translator(-Peye_w(1, 1), -Peye_w(1, 2), -Peye_w(1, 3)) * Rot);
     if (debug) {
	  PrintMatrix("View:", View);
	  PrintMatrix("Rotation matrix:", Rot);
	  PrintMatrix("V:", V);
     }

     return V;
}

Hedron Perspective(const Hedron& h1, const double d)
{
     Hedron h2(h1);
     double v;
     int i;

     for (i = 1; i <= h2.Vertices().Rows(); i++) {
	  if (h2.Vertices()(i, 3) == 0.0) {
	       h2.Vertices()(i, 1) = HUGE;
	       h2.Vertices()(i, 2) = HUGE;
	  }
	  else {
	       v = h2.Vertices()(i, 3) / d;
	       h2.Vertices()(i, 1) /= v;
	       h2.Vertices()(i, 2) /= v;
	  }
     }

     return h2;
}

Hedron GetPoints()
{
     Event e;
     int x, y, i, j;
     Matrix points(0, 4), hack(1, 4, 0, 0, 0, 1);
     Clear();

     i = 0;
     while (1) {
	  e = GetEvent();
	  if (EventButton(e) == 1) {
	       i++;
	       x = EventX(e);   y = EventY(e);
	       hack(1, 1) = double(x);
	       hack(1, 2) = double(y);
	       points = points & hack;
	       PlotToScale(x, y, 5);
	  }
	  else if (EventButton(e) == 3)
	       break;
     }

     // Move the matrix to (0, 0, 0)
     points = points * Translator(-double(WIDTH)/2.0,
				  -double(HEIGHT)/2.0, 0.0);
     
     Hedron h(points, i);
     for (j = 0; j < i; j++) {
	  h.Edges()[2*j] = j+1;
	  h.Edges()[2*j+1] = j+2;
     }
     h.Edges()[2*i-1] = 1;

     printf("Points Hedron:\n");
     h.Print();
     return h;
}

void problem3()
{
     Matrix CTM(4, 4);
     Matrix Peye(1, 3, 0, 0, -250);
     Matrix UP(1, 3, 0, 1, 0);
     Matrix Pfix(1, 3, 0, 0 ,0);
     Matrix V(4, 4);
     double border = 20.0, hither = 10.0, yon = 500.0;
     double a, b, c;
     char q;
     int i, k, cnt;

     CTM.Identify();

     /* Initial Setup */
     Hedron points(GetPoints());
     Hedron complete(points);
     int pts = points.Vertices().Rows();
     for (i = 20, cnt = 1; i <= 360; cnt += 1, i += 20) {
	  complete = complete + points * CTM;
	  CTM = CTM * Rotator_Y(20);
	  for (k = 0; k < pts; k++)
	       complete.AddEdge(cnt*pts+k+1, (cnt-1)*pts+k+1);
     }
     for (k = 0; k < pts; k++)
	  complete.AddEdge((cnt-1)*pts+k+1, k+1);
     
     while (1) {
	  Clear();

	  double PerMag = MatrixMag(Peye - Pfix);
	  V = PreClipTransform(Peye, Pfix, UP);

	  Hedron world(complete*CTM);
	  Hedron eye(world*V);
	  Hedron perspect(Perspective(eye, PerMag));
	  Hedron clipped(ClipHedron(perspect,
				    -(WIDTH/2.0-border),WIDTH/2.0-border,
				    -(HEIGHT/2.0-border),HEIGHT/2.0-border,
				    hither, yon));

	  DrawPolyhedron(clipped * Translator(150, 150, 0));
	  ForceOutput();

	  if (scanf("%1s", &q) == EOF)
	       break;

	  switch (q) {
	  case 'c':
	  case 'C':
	       scanf("%f %f %f", &a, &b, &c);
	       border = a; hither = b; yon = c;
	       break;
	  case 'e':
	  case 'E':
	       scanf("%f %f %f", &a, &b, &c);
	       Peye(1, 1) = a; Peye(1, 2) = b; Peye(1, 3) = c;
	       break;
	  case 'f':
	  case 'F':
	       scanf("%f %f %f", &a, &b, &c);
	       Pfix(1, 1) = a; Pfix(1, 2) = b; Pfix(1, 3) = c;
	       break;
	  case 'u':
	  case 'U':
	       scanf("%f %f %f", &a, &b, &c);
	       UP(1, 1) = a; UP(1, 2) = b; UP(1, 3) = c;
	       break;
	  case 't':
	  case 'T':
	       scanf("%f %f %f", &a, &b, &c);
	       CTM = CTM * Translator(a, b, c);
	       break;
	  case 's':
	  case 'S':
	       scanf("%f %f %f", &a, &b, &c);
	       CTM = CTM * Scaler(a, b, c);
	       break;
	  case 'r':
	  case 'R':
	       scanf("%1s %f", &q, &a);
	       switch (q) {
	       case 'X':
	       case 'x':
		    CTM = CTM * Rotator_X(a);
		    break;
	       case 'Y':
	       case 'y':
		    CTM = CTM * Rotator_Y(a);
		    break;
	       case 'Z':
	       case 'z':
		    CTM = CTM * Rotator_Z(a);
		    break;
	       }
	       break;
	  case 'i':
	  case 'I':
	       CTM.Identify();
	       break;
	  }
     }

     /* Reset the EOF condition */
     clearerr(stdin);
     printf("\n");
}
     
main()
{
     Panel p;

     p = NewPanel(WIDTH, HEIGHT);
     SetFunction(DRAW);
     UsePanel(p);

     problem3();

     DoneWithPanel(p);
     return 0;
}
