;;;; the material in this file is drawn from copyrighted programs
;;;; in lisp, third edition, by patrick h. winston and berthold
;;;; k. p. horn, and in the san marco lisp explorer (r).

;;;; from san marco lisp explorer (r), ch28.lsp
;;;; copyright (c) 1984, 1985, 1987, 1988, 1989 san marco associates


;;this portion of the file contains an interpreter for a
;;transition-tree natural language interface.

(defvar *debug* nil)
(defvar *debug-call* nil)

;;;; tree definition and interpretation

(defmacro define-tree (&rest args)
  (let ((name-of-tree (first args))
	(tree-description (second args)))
    `(defun ,name-of-tree (word-list)
      (when *debug*
	(format t "~%~s entered with ~s" ',name-of-tree word-list))
      (multiple-value-bind 
	    (result binding remaining-word-list)
	  (interpret-tree ',tree-description word-list)
	(when *debug*
	  (if result
	      (format t "~%~s suceeded with ~s remaining" 
		      ',name-of-tree remaining-word-list)
	      (format t "~%~s failed with  ~s remaining" 
		      ',name-of-tree remaining-word-list)))
	(values result binding remaining-word-list)))))

(defun interpret-tree (tree word-list &optional a-list)
  (cond ((endp tree) (values t nil word-list))
        ((eq (first tree) (first word-list))
         (interpret-tree (rest tree) (rest word-list) a-list))
        ((eq '> (first tree))
         (multiple-value-bind
             (result binding word-list)
             (funcall (second tree) word-list)
           (when result
             (interpret-tree (rest (rest tree))
                            word-list
                            (cons (list (second tree) binding)
                                  a-list)))))
        ((eq 'rtn (first tree))
         (values t
                 (evaluate-forms (rest tree) a-list)
                 word-list))
        ((and (eq 'if-end-rtn (first tree))
              (endp word-list))
         (values t (evaluate-forms (rest tree) a-list) nil))
        ((eq 'brnchs (first tree))
         (interpret-branches (rest tree) word-list a-list))
        (t (values nil nil word-list))))

(defun interpret-branches (branches word-list a-list)
  (if (endp branches)
      (values nil nil word-list)
      (progn
	(when *debug* (format t "~% starting on ~s" (first branches)))
	(multiple-value-bind (result binding words-left-over)
	    (interpret-tree (first branches) word-list a-list)
	  (cond (result
		 (when *debug* (format t "~% suceeded on ~s" (first branches)))
		 (values result binding words-left-over))
		(t
		 (when *debug* (format t "~% failed on ~s" (first branches)))
		 (interpret-branches (rest branches)
				     word-list
				     a-list)))))))

;;;; auxiliaries

(defun evaluate-forms (forms a-list)
  (eval `(let ,(make-let-variables a-list)
              ,@forms)))

(defun make-let-variables (a-list)
  (mapcar #'(lambda (pair)
              `(,(first pair) ',(second pair))) a-list))


;;;; top-level interface

(defun run-interface ()
  (print 'sentence>)
  (do ((input (read-sentence) (read-sentence)))
      ((equal input '(quit))
       (format t "~&ok---goodbye.")
       (values))
    (when input
      (unless (interface input)
	(format t "~&sorry---i can't understand that.~
              ~&enter quit if you want to stop."))
      (print 'sentence>))))

(defun read-sentence ()
  (with-input-from-string
    (input (string-trim ".?!" (read-line)))
    (do ((word (read input nil)
               (read input nil))
         (sentence nil))
        ((not word) (return (reverse sentence)))
      (push word sentence))))

;;;; from san marco lisp explorer (r), ch28.dta
;;;; copyright (c) 1984, 1985, 1987, 1988, 1989 san marco associates

#|
this portion of the file contains test data forms for the natural
language interpreter.
|#

;;;; transition-tree grammar

(define-tree interface
  (brnchs
   (count > objects if-end-rtn
	  (db-call `(db-count ,objects)))
   (how many > objects are there if-end-rtn
	(db-call `(db-count ,objects)))
   (identify the > attributes of > objects if-end-rtn
	     (db-call `(db-show (db-project ,objects over
					    ,@attributes))))
   (identify > objects if-end-rtn
	     (db-call `(db-show ,objects)))
   (what > attributes are > objects if-end-rtn
	 (db-call `(db-show (db-project ,objects over ,@attributes))))
   ))

(define-tree attributes
  (brnchs (> attribute > attributes rtn (cons attribute
                                              attributes))
          (and > attribute          rtn (list attribute))
          (> attribute              rtn (list attribute))))

(define-tree attribute
  (brnchs (class        rtn 'class)
          (classes      rtn 'class)	;added.
          (color        rtn 'color)
          (colors       rtn 'color)	;added.
          (size         rtn 'size)
          (sizes        rtn 'size)	;added.
          (weight       rtn 'weight)
          (weights      rtn 'weight)	;added.
          (position     rtn 'peg)
          (positions    rtn 'peg)	;added.
          (location     rtn 'peg)
          (locations    rtn 'peg)))	;added.

(define-tree attribute-values
  (brnchs (> attribute-value > attribute-values
             rtn (append attribute-value attribute-values))
          (and > attribute-value rtn attribute-value)
          (> attribute-value rtn attribute-value)))

(define-tree attribute-value
  (brnchs (large        rtn '(size eql large))
          (medium       rtn '(size eql medium))
          (small        rtn '(size eql small))
          (long         rtn '(size eql long))
          (short        rtn '(size eql short))
          (black        rtn '(color eql black))
          (blue         rtn '(color eql blue))
          (red          rtn '(color eql red))
          (yellow       rtn '(color eql yellow))
          (gray         rtn '(color eql gray))
          (heavy        rtn '(weight eql heavy))
          (light        rtn '(weight eql light))))

(define-tree objects
  (brnchs (> determiner > objects rtn objects)
          (> attributed-object and > objects
	       rtn `(db-union ,attributed-object ,objects))
	  (> attributed-object
	     rtn attributed-object)
	  ))

(define-tree attributed-object
    (brnchs (> attribute-values > object
	       rtn `(db-select ,object with ,@attribute-values))
	    (> object
	       rtn object)
	    ))

(define-tree determiner
  (brnchs (a) (the)))

(define-tree object
  (brnchs (saw          rtn '(db-select tools with class eql saw))
          (saws         rtn '(db-select tools with class eql saw))
          (hammer       rtn '(db-select tools with class eql hammer))
          (hammers      rtn '(db-select tools with class eql hammer))
          (wrench       rtn '(db-select tools with class eql wrench))
          (wrenches     rtn '(db-select tools with class eql wrench))
          (screwdriver  rtn '(db-select tools with class eql screwdriver))
          (screwdrivers rtn '(db-select tools with class eql
			      screwdriver))
	  ))

;;; End of transistion network

;;;; san marco lisp explorer (r), ch30.lsp
;;;; copyright (c) 1984, 1985, 1987, 1988, 1989 san marco associates

#|
this portion of the file contains a set of access for an
instructional relational database system.  note that
all the records for each database are kept in random
access memory.  in commercial-grade relational
database systems, records are normally stored on disk
memory and are brought into random access memory one
at a time or in small batches.

some of the procedures listed are drawn from the problem
solutions.  several are modified so that the symbol and is
ignored in argument lists.
|#

;;;; database interface

(defun db-call (arg) 
  (if *debug-call*
      (format t "~%~s" arg))
  (eval arg))

;;;; relation access procedures

(defmacro db-make-relation (&rest args)
  (let ((relation (first args))
	(fields-and-records (rest args)))
    `(setf ,relation
	   ',fields-and-records)))

(defmacro db-extract-value (&rest args )
  (let ((relation (first args))
	(projections (rest (rest args))))
    `(db-extract-value-aux ,relation ',(remove-and projections))))

(defun db-extract-value-aux (relation projections)
  (let ((result (db-project-aux (first relation)
                                (rest relation)
                                projections)))
    (when (rest result)
      (format t "~&i'm taking the first of more than one record."))
    (when (rest (first result))
      (format t "~&i'm taking the first of more than one field."))
    (first (first result))))

(defun db-count (relation)
  (format t "~&~a" (length (rest relation)))
  (values))

;;;; relation manipulators

(defun db-union (&rest records)
  (cons (first (first records))
        (remove-duplicates (apply #'append
                                  (mapcar #'rest records))
                           :test #'equal)))

(defmacro db-project (&rest args)
  (let ((relation (first args))
	(projections (rest (rest args))))
    `(cons ',(remove-and projections)
	   (db-project-aux (first ,relation) (rest ,relation)
			   ',(remove-and projections)))))

(defun db-project-aux (fields records projections)
  (remove-duplicates
    (mapcar
      #'(lambda (record)
          (mapcar
            #'(lambda (projection)
		(nth (position projection fields) record))
            projections))
      records)
    :test #'equal))

(defmacro db-select (&rest args)
  (let ((relation (first args))
	(triples (rest (rest args))))
    `(cons (first ,relation)
	   (db-select-aux (first ,relation)
			  (rest ,relation)
			  ',(remove-and triples)))))

(defun db-select-aux (fields records triples)
  (if (endp triples)
      records
      (remove-if-not
        #'(lambda (record)
            (funcall (second triples)
                     (nth (position (first triples) fields)
                          record)
                     (third triples)))
        (db-select-aux fields records (nthcdr 3 triples)))))

(defmacro db-join (&rest args)
  (let ((db1 (first args))
	(db2 (second args))
	(triples (rest (rest (rest args)))))
    `(cons (append (first ,db1) (first ,db2))
	   (db-join-aux (first ,db1)
			(first ,db2)
			(rest ,db1)
			(rest ,db2)
			',(remove-and triples)))))

(defun db-join-aux (fields1 fields2 db1 db2 triples &aux result)
  (dolist (line1 db1 (reverse result))
    (dolist (line2 db2)
      (do* ((triples triples (nthcdr 3 triples)))
           ((endp triples)
            (setf result (cons (append line1 line2) result)))
        (unless (funcall (second triples)
                         (nth (position (first triples) fields1)
                              line1)
                         (nth (position (third triples) fields2)
                              line2))
          (return))))))

;;;; auxiliary

(defun remove-and (list)
  (cond ((endp list) nil)
	((eql 'and (first list)) (rest list))
	(t (cons (first list) (remove-and (rest list))))))

;;;; relation displayer

(defun db-show (relation)
  (let ((widths (db-find-field-widths relation)))
    (db-show-record widths (first relation))
    (db-show-record
        widths
        (mapcar #'(lambda (width)
                    (make-string width :initial-element #\-))
                widths))
    (dolist (record (rest relation) (values))
      (db-show-record widths record))))

(defun db-show-record (widths fields)
  (format t "~&|")
      (do ((widths widths (rest widths))
           (fields fields (rest fields)))
	  ((endp widths))
	;;rewritten to work with common lisp subsets:
	(format t " ~a~a |"
		(first fields)
		(make-string
		  (- (first widths)
		     (length (format nil "~a" (first fields))))
		  :initial-element #\ ))))

(defun db-find-field-widths (relation &aux result)
  (setf result
        (mapcar #'(lambda (field)
                    (length (format nil "~a" field)))
                (first relation)))
  (dolist (record (rest relation) result)
    (setf result 
          (mapcar #'(lambda (number symbol)
                      (max number
                           (length (format nil "~a" symbol))))
                  result
                  record))))

;;;; San Marco LISP Explorer (R), CH30.DTA
;;;; Copyright (c) 1984, 1985, 1987, 1988, 1989 San Marco Associates

#|
This portion of the file contains test data for various relational
database procedures.
|#

;;;; Test Data For Relational Database Procedures

(db-make-relation tools
  (class 	color 	size 	weight)
  (saw		black	medium	heavy)
  (hammer	blue	large	heavy)
  (wrench	gray	small	light)
  (wrench	gray	large	heavy)
  (screwdriver	blue	long	light)
  (screwdriver	black	long	light)
  (screwdriver	red	long	heavy)
  (screwdriver	red	short	light))

