command_t = cluster is
    create, is_depth, is_exit, is_load, is_move, is_new, is_undo,
    is_write
    
    %%
    %% Overview:
    %% A command_t is a one of depth, exit, load, move, new, undo, 
    %% or write.  It has no semantic value, and is used only as an
    %% identifier.
    %%
    %% Abstraction function:
    %% A(r) = up(r)
    %%
    %% Rep invariant:
    %% true
    %%
    
    rep = oneof[depth, exit, load, move, new, undo, write: null]
    
    create = proc (s: string) returns (command_t) signals (bad_command)
	% Effects: Returns a command_t set to the tag s if s is one
	% of "depth", "exit", "load", "move", "new", "undo", or
	% "write"; signals bad_command otherwise.
	end create
    
    is_depth = proc (command: command_t) returns (bool)
	% Effects: Returns true of tag of command is depth, false otherwise.
	end is_depth

    is_exit = proc (command: command_t) returns (bool)
	% Effects: Returns true of tag of command is exit, false otherwise.
	end is_exit
    
    is_load = proc (command: command_t) returns (bool)
	% Effects: Returns true of tag of command is load, false otherwise.
	end is_load
    
    is_move = proc (command: command_t) returns (bool)
	% Effects: Returns true of tag of command is move, false otherwise.
	end is_move
    
    is_new = proc (command: command_t) returns (bool)
	% Effects: Returns true of tag of command is new, false otherwise.
	end is_new
    
    is_undo = proc (command: command_t) returns (bool)
	% Effects: Returns true of tag of command is undo, false otherwise.
	end is_undo

    is_write = proc (command: command_t) returns (bool)
	% Effects: Returns true of tag of command is write, false otherwise.
	end is_write
    
    end command_t
