coord_t = cluster is 
    parse, unparse, get_x, get_y, forward_neighbors, is_edge, is_corner,
    equal

    %%
    %% Overview:
    %% A coord_t is an immutable data type with an x and y coordinate 
    %% where a coordinate is an integer between 1 and 8 inclusive.
    %%
    %% Rep invariant:
    %% 1 |*$\le$*| rep.x, rep.y |*$\le$*| 8
    %%
    %% Abstraction function:
    %% A(r) = up(r)
    %%
    
    rep = record[x, y: int]
    
    create = proc (x, y: int) returns (coord_t) signals (bad_coord)
	% Effects: If 1 |*$\le$*| x, y |*$\le$*| 8, creates a coord_t
	% with x and y and returns it; otherwise signals bad_coord.
	end create
    
    parse = proc(s: string) returns (coord_t) signals (parse_error)
	% Parses. 
	end parse
    
    unparse = proc(p: coord_t) returns(string)    
	% Unparses.
	end unparse
    
    get_x = proc(p: coord_t) returns(int)
	% Effect: Returns the x coordinate of coord p.
	end get_x
    
    get_y = proc(p: coord_t) returns(int)
	% Effect: Returns the y coordinate of coord p.
	end get_y
    
    forward_neighbors = iter(p: coord_t) yields(coord_t)
	% Effect: Returns all of the coords diagonally adjacent to the current
	% coord.
	end forward_neighbors
    
    is_edge = proc(p: coord_t) returns(bool)
	% Effect: Returns true iff the coord_t is at the edge of the board.
	end is_edge
    
    is_corner = proc(p: coord_t) returns(bool)
	% Effect: Returns true iff the coord_t is at the corner of the board.
	end is_corner
    
    equal = proc(p1, p2: coord_t) returns(bool)
	% Effect: Returns true iff p1 and p2 are the same coord.
	end equal
    
    end coord_t
