display_t = cluster is 
    create, score, board, turn, level, status, clear_status, error,
    clear_error, clear_editor, get_next_command

    %%
    %% Overview:
    %% The display cluster is the program's only interface to the
    %% terminal and keyboard devices.  It has functions to display
    %% certain parts or all of a game cluster on the screen, and
    %% a single function get_next_command that uses a popup menu and
    %% line editor to get commands from the user.
    %% 
    %% Rep invariant:
    %% rep.kbd is the keyboard device, rep.scr is the terminal device
    %%
    %% Abstraction function:
    %% I'm not sure it makes sense to have an abstraction function
    %% for this cluster.
    %%
    
    rep = struct[kbd: keyboard, scr: screen]
    
    create = proc () returns (display_t)
	% Effects: Initializes the screen, but does not actually 
	% draw anything.  Returns the display.  Can not be called 
	% more than once per process.
	end initialize
    
    score = proc (disp: display_t, game: game_t)
	% Modifes: terminal device
	% Effects: Redraws the score field on the screen from game.
	end score
    
    board = proc (disp: display_t, game: game_t)
	% Modifes: terminal device
	% Effects: Redraws the board on the screen from game.
	end board
    
    turn = proc (disp: display_t, game: game_t)
	% Modifes: terminal device
	% Effects: Redraws whose turn it is on the screen from game.
	end turn
    
    level = proc (disp: display_t, game: game_t)
	% Modifes: terminal device
	% Effects: Redraws the level on the screen from game.
	end level
    
    status = proc (disp: display_t, msg: string)
	% Modifes: terminal device
	% Effects: Redraws the status message msg on the screen.
	end status
    
    clear_status = proc (disp: display_t)
	% Modifes: terminal device
	% Effects: Clears the status message on the screen.
	end clear_status
    
    error = proc (disp: display_t, msg: string)
	% Modifes: terminal device
	% Effects: Displays the error message msg on the screen.
	end error
    
    clear_error = proc (disp: display_t)
	% Modifes: terminal device
	% Effects: Clears the error message on the screen.
	end clear_error
    
    clear_editor = proc (disp: display_t, game: game_t)
	% Modifes: terminal device
	% Effects: Clears the editor area on the screen.
	end clear_editor
    
    redraw = proc (disp: display_t, game: game_t)
	% Modifes: terminal device
	% Effects: Redraws the entire screen from game, clearing the
	% status and error areas.
	end redraw
    
    get_next_command = proc (disp: dispay_t, reuse: bool)
	returns (command_t, string)
	% Modifes: terminal device
	% Effects: Gets the next command from the user.
	% Returns the command and the string argument if any;
	% the string is "" is there isn't a argument.
	%
	% If reuse is true, the previous command is displayed;
	% otherwise the previous command is destroyed.
	end get_next_command
        
    end display_t

