game_t = cluster is
    create, set_next_turn, get_next_turn, set_level, get_level,
    move, undo, parse, unparse
    
    %%
    %% Overview:
    %% Game_t is the top level representation of a game of Cromwell's
    %% Checkers.  It contains all the state information for an actual
    %% game.  It is a record that contains the player whose turn it is
    %% next, a lookahead level for the computer, scores for black and
    %% white, a board, and a history list of all turns.
    %% 
    %% Rep invariant:
    %% Sheesh, I don't know.
    %%
    %% Abstraction function:
    %% I don't know this, either.
    %% 
    
    rep = record[next_turn_color: color_t, level, black_score, 
    		white_score: int, board: board_t, history: history_t]
    
    create = proc () returns (game_t)
	% Effects:
	%  1) creates a board and initializes it to starting position
	%  2) inits level to default_level, next_turn_color to black,
	%     history to the empty list
	%  3) returns it.
	end create
    
    set_next_turn = proc (g: game_t, color: color_t) returns ()
	% Effects: Changes next_turn_color to next player.
	end next_turn
    
    get_next_turn = proc (g: game) returns (color_t)
	% Effects: Returns the player color that will be used for 
	% the next command that uses a color.
	end get_next_turn
    
    set_level = proc (g: game_t, l: int) returns ()
	% Effects: Sets the level to l.
	end set_level
    
    get_level = proc (g: game_t) returns (int)
	% Effects: Returns the current level of the search.
	end get_level
    
    move = proc (g: game_t, turn: turn_t)
	% Effects: 
	%   1) Performs turn (board$move)
	%   2) For each move in turn, updates the score and sets the 
	%      who field
	%   3) Adds the move to the history list (history$push)
	%   4) Updates next_turn_color
	end move
    
    undo = proc (g: game_t) signals (history_list_empty)
	% Effects: Pops the last element of the history list, and undoes
	% it. This means: restore pieces to original positions, put 
	% jumped pieces back on board, decrement appropriate score, 
	% change next_turn_color.
	end undo
    
    unparse = proc (g: game_t) returns (string)
	% Unparses.
	end unparse
    
    parse = proc (s: string) returns (game_t) signals (parse_error)
	% Parses.
	end parse

    end game_t
