history_t = cluster is
    create, push, pop, peek, turns_new_to_old, is_empty, parse, unparse

    %%
    %% Overview:
    %% A history is a stack of turns.  It is used by the undo command.
    %% It also has the property that the top element (returned by peek or
    %% pop) is the last turn made.  Elements are pushed and popped from/to
    %% the top of the array (addh and remh, not addl and reml.)
    %%
    %% Rep invariant:
    %% Given a game and a history list, undoing each turn on the history
    %% list produces a game with the initial board configuration and
    %% score.  This means that each turn is legal, that the who field
    %% of each turn in sequence alternates, and that for each move in each
    %% turn in the history, either 1) from and to name adjacent coordinates
    %% and jumped is false, or 2) from and to name diagonal coordinates 
    %% separated by exactly one other coordinate and jumped is true.
    %% 
    %% Abstraction function:
    %% A(r) = up(r)
    %% 
        
    rep = array[turn_t]
    
    create = proc () returns (history_t)
	% Effect: Returns an empty history stack
	end create
    
    push = proc (hist: history_t, turn: turn_t)
	% Effects: Pushes turn onto the top of the stack
	end push

    pop = proc (hist: history) returns (turn_t) signals (bounds)
	% Effects: Pops a turn of the top of the stack and returns it, or
	% signals bounds if it is empty.
	end pop

    peek = proc (hist: history) returns (turn_t) signals (bounds)
	% Effects: returns the turn on the top of the stack WITHOUT 
	% POPPING IT, or signals bounds if it is empty.
	end peek
    
    turns_new_to_old = iter (hist: history_t) yields (turn_t)
	% Requires: that hist is not modified in the loop
	% Effects: Yields each turn in hist in the order in which they 
	% would be popped.
	end moves
    
    is_empty = proc (hist: history_t) returns (bool)
	% Effects: Returns true if hist is empty, false otherwise.
	end is_empty

    parse = proc (s: string) returns (history_t) signals (parse_error)
	% Parses.
	end parse

    unparse = proc (hist: history_t) returns (string)
	% Unparses.
	end unoparse
    
    end history_t
