move_t = cluster is
    create, set_from, set_to, set_jumped, set_who, get_from, get_to, 
    get_jumped, get_who, parse, unparse
    
    %%
    %% Overview:
    %% A move contains the following information: the from and to
    %% coordinates of the move, whether the move jumped an opponent's
    %% piece, and who made the move.
    %%
    %% Rep invariant:
    %% from and to name coordinates that are either diagnolly
    %% adjacent or diagnol with one coordinate between them.  Furthermore,
    %% to is "forward" of from relative to the who field.
    %%
    %% Abstraction function:
    %% A(r) = up(r)
    %%
    
    rep = record[from, to: coord_t, jumped: bool, who: color_t]
    
    create = proc (from, to: coord_t) returns (move_t) signalls (bad_coord)
	% Effects: Returns a move_t with from and to filled in and with
	% jumped and who unset, if from and to a legal; otherwise,
	% signals bad_coord.
	end create
    
    set_from = proc (move: move_t, from: coord_t)
	% Effects: Sets the from field of move to from.
	end set_from

    set_who = proc (move: move_t, who: color_t)
	% Effects: Sets the who field of move to who.
	end set_who
    
    set_to = proc (move: move_t, to: coord_t)
	% Effects: Sets the to field of the coord_t to to.
	end set_to
    
    set_jumped = proc (move: move_t, jumped: bool)
	% Effects: Sets the jumped field of move to jumped.
	end set_jumped
    
    get_from = proc (move: move_t) returns (coord_t)
	% Effects: Returns the from field of move.
	end get_from
    
    get_to = proc (move: move_t) returns (coord_t)
	% Effects: Returns the to field of move.
	end get_to

    get_jumped = proc (move: move_t) returns (bool)
	% Effects: Returns the jumped field of move.
	end get_jumped

    get_who = proc (move: move_t) returns (color_t)
	% Effects: Returns the who field of move.
	end get_who

    parse = proc (s: string) returns (move_t) signals (parse_error)
	% Parses.
	end parse
    
    unparse = proc (move: move_t) returns (string)
	% Unparses.
	end unparse
    
    end move_t
