turn_t = cluster is
    create, enq, deq, peek, set_who, get_who, moves, parse, unparse
    
    %%
    %% Overview:
    %% A turn is a queue of moves paired with the player who made them.
    %% Moves are added and the low end and removed at the high end of
    %% the array.
    %%
    %% Rep invariant:
    %% The who field of a turn is the same as the who field of all of
    %% its moves.
    %%
    %% Abstraction function:
    %% A(r) = up(r)
    %% 

    rep = record[moves: array[move_t], who: color_t]
    
    create = proc (who: color_t) returns (turn_t)
	% Effects: Returns a turn with no moves and a who field set
	% to who.
	end create

    enq = proc (turn: turn_t, move: move_t)
	% Effects: Adds move to the queue turn.
	end enq
    
    deq = proc (turn: turn_t) returns (move_t) signals (bounds)
	% Effects: Dequeues a move from turn and returns it if
	% turn is not empty; signals bounds otherwise.
	end deq
    
    peek = proc (turn: turn_t) retuns (move_t) signals (bounds)
	% Effects: Returns the move that would be returned by deq 
	% but does not dequeue it if turn is not empty; signals
	% bounds otherwise.
	end peek
    
    set_who = proc (turn: turn_t, who: color_t)
	% Effects: Sets the who field of the turn to who.
	end set_who
    
    get_who = proc (turn: turn_t) returns (color_t)
	% Effects: Returns the who field of turn.
	end get_who
    
    moves = iter (turn: turn_t) yields (move_t)
	% Requires: turn is not modified in the loop
	% Effects: Yields each move in turn in the order in which 
	% they were enqueued.
	end moves

    unparse = proc (turn: turn_t) returns (string)
	% Unparses.
	end unparse
    
    parse = proc (s: string) returns (turn_t) signals (parse_error)
	% Parses.
	end parse
    
    end turn_t
