#include "Solid.h"

Solid::Solid(const Matrix& pts, int init = 10)
   :vertices(pts),faces(sizeof(int *), init),colors(sizeof(double), init)
{
     num_faces = 0;
}

Solid::Solid(const Solid& s)
   :vertices(s.vertices), faces(sizeof(int *), 10),
	colors(sizeof(double), 10)
{
     int i, l, *p;

     for (i=0; i < s.faces.Size(); i++) {
	  l = s.FaceLength(i) + 1;
	  p = new int[l];
	  bcopy((char *) *((int **) s.faces[i]), (char *) p, sizeof(int)*l);
	  faces += (char *) &p;
     }

     num_faces = s.num_faces;
     colors = s.colors;
}

Solid::~Solid()
{
     int i;

     for (i=0; i < faces.High(); i++)
	  delete *((int **)faces[i]);
}

Solid& Solid::operator=(const Solid& s)
{
     int i;
     
     /* Clean up the faces */
     for (i=0; i<num_faces; i++)
	  delete *((int **) faces[i]);

     faces = s.faces;
     vertices = s.vertices;
     num_faces = s.num_faces;

     return *this;
}

void Solid::AddFace(double color, int n, const int *v)
{
     int *lv = new int[n+1];
     bcopy((char *) v, (char *) lv, sizeof(int)*n);
     lv[n] = 0;
     faces += (char *) &lv;
     num_faces += 1;
     colors += (char *) &color;
}

int Solid::FaceLength(int i) const
{
     int *face = *((int **) faces.Get(i));
     int *p = face;

     while (*p++)
	  ;
     return (p - face - 1);
}

Solid operator*(const Solid& s1, const Matrix& m)
{
     Solid s2(s1);

     s2.Vertices() = s2.Vertices() * m;
     return s2;
}

void Solid::Print() const
{
     int i, j;
     
     printf("Vertices:\n");
     vertices.Print();

     for (i=0; i < num_faces; i++) {
	  printf("Face %d, length %d:", i, FaceLength(i));
	  for (j = 0; j < FaceLength(i); j++)
	       printf(" %d",  (*((int **) faces[i]))[j]);
	  printf("\n");
     }
}
