#include <stdio.h>
#include <math.h>
#include "halftone.h"
#include "Matrix.h"

extern "C" {
#include <tgif/tgif.h>
#include <image/Image.h>
}

inline double sinusoid(int i, int j, int w, int h)
{
     return (sin(double(i)*10.0*PI/double(w))*
	     sin(double(j)*10.0*PI/double(h)) + 1) * 127.5;
}

void SinImage(Image *img, int width, int height)
{
     int i, j;

     printf("Creating sinusoid matrix ...");
     fflush(stdout);

     for (i = 0; i < width; i++)
	  for (j = 0; j < height; j++)
	       SetImg(img, i, j, sinusoid(i, j, width, height));

     printf(" done.\n");
}

main(int argc, char **argv)
{
     const int x = 200, y = 200;
     Panel p;
     Image *img;

     p = NewPanel(x, y);
     UsePanel(p);

     if (argc == 1) {
	  img = NewImg(x, y);
	  SinImage(img, x, y);
	  FS_Halftone(img);
	  WaitForButton();
	  FreeImg(img);
     }
     else {
	  if (! ReadImg(argv[1], &img)) {
	       fprintf(stderr, "Cannot open image file.\n");
	       exit(1);
	  }
	  FS_Halftone(img);
	  WaitForButton();
	  FreeImg(img);
     }

     DoneWithPanel(p);
}
