#include <stdio.h>
#include <stdlib.h>
#include "Matrix.h"
#include "MatrixUtil.h"
#include "Solid.h"
#include "transform.h"
#include "View3D.h"
#include "render.h"

extern "C" {
#include <tgif/tgif.h>
}

const int WIDTH = 300;
const int HEIGHT = 300;
const int debug = 0;

int CubeColors[] = { 100, 130, 156, 168, 210, 252 };
int CubePoints[] = {
     -50, -50, -50, 1,
     -50, 50, -50, 1,
     50, 50, -50, 1,
     50, -50, -50, 1,
     -50, -50, 50, 1,
     -50, 50, 50, 1,
     50, 50, 50, 1,
     50, -50, 50, 1
};

int front[] = { 1, 4, 3, 2 };
int left[] = { 1, 2, 6, 5 };
int top[] = { 2, 3, 7, 6 };
int right[] = { 3, 4, 8, 7 };
int back[] = { 5, 6, 7, 8 };
int bottom[] = { 1, 5, 8, 4 };

int *CubeFaces[6] = {
     front, left, top, right, back, bottom,
};

Matrix pyrpts(5,4,
	      -50,0,-50,1,
	      50,0,-50,1,
	      50,0,50,1,
	      -50,0,50,1,
	      0,100,0,1);

int p1[] = {1, 2, 3, 4};
int p2[] = {2, 1, 5};
int p3[] = {3, 2, 5};
int p4[] = {4, 3, 5};
int p5[] = {1, 4, 5};

int *pyrfaces[] = {p1, p2, p3, p4, p5};
int pyrnfaces[] = {4, 3, 3, 3, 3};


#define DLine(a,b,c,d) Line(int(a), int(b), int(c), int(d))
void DrawSolid(const Solid& s)
{
     int i, j;

     for (i = 0; i < s.Faces(); i++) {
	  const int *p = s.Face(i);
	  for (j = 0; j < s.FaceLength(i)-1; j++) {
	       DLine(s.Vertices()(p[j], 1), s.Vertices()(p[j], 2), 
		     s.Vertices()(p[j+1], 1), s.Vertices()(p[j+1], 2));
	  }
	  DLine(s.Vertices()(p[j], 1), s.Vertices()(p[j], 2),
		s.Vertices()(p[0], 1), s.Vertices()(p[0], 2));
     }
}

void ParseParamFile(char *filename, Matrix *Peye, Matrix *UP,
		    Matrix *Pfix)
   // Fills in all arguments after the first.
{
     FILE *f;
     double a, b, c;
     int i, j, k, l;

     f = fopen(filename, "r");
     if (f == NULL) {
	  fprintf(stderr, "Cannot open parameter file %s.\n", filename);
	  exit(1);
     }

     /* Eye point */
     fscanf(f, "%f %f %f", &a, &b, &c);
     (*Peye)(1, 1) = a; (*Peye)(1, 2) = b; (*Peye)(1, 3) = c;
     if (debug)
	  printf("Eye point:\t %6.3f %6.3f %6.3f\n", a, b, c);

     /* UP vector */
     fscanf(f, "%f %f %f", &a, &b, &c);
     (*UP)(1, 1) = a; (*UP)(1, 2) = b; (*UP)(1, 3) = c;
     if (debug)
	  printf("UP vector:\t %6.3f %6.3f %6.3f\n", a, b, c);

     /* Fixation point */
     fscanf(f, "%f %f %f", &a, &b, &c);
     (*Pfix)(1, 1) = a; (*Pfix)(1, 2) = b; (*Pfix)(1, 3) = c;
     if (debug)
	  printf("Fixation point:\t %6.3f %6.3f %6.3f\n", a, b, c);

     fclose(f);
}

void problem2(char *filename)
{
     Matrix CTM(4, 4), Peye(1, 3), UP(1, 3), Pfix(1, 3), V(4, 4), *points;
     double S, F, d_s, d_f, border = 20.0, hither = 10.0, yon = 500.0;
     double a, b, c;
     char z, mode = 0;
     int i;

     /* Get the viewing situation */
     ParseParamFile(filename, &Peye, &UP, &Pfix);

     /* Hardcoded because of some weird pointer bug */
     points = new Matrix(8, 4, CubePoints);
     Solid cube(*points);
     for (i = 0; i < 6; i++)
	  cube.AddFace(CubeColors[i], 4, CubeFaces[i]);

     Solid pyr(pyrpts*Translator(0.0,0.0,-100.0));
     for (i=0; 	i<5; i++)
	  pyr.AddFace(i * 220/8.0+20,pyrnfaces[i],pyrfaces[i]);
     Solid solid(cube+pyr);


     if (debug) {
	  printf("Solid:\n");
	  solid.Print();
     }	  
     
     CTM.Identify();

     while (1) {
	  Clear();

	  if (debug) {
	       printf("Transformed solid:\n");
	       View3D(solid*CTM, Peye, Pfix, UP, hither, yon, 
		      150.0, 150.0, 150.0, 150.0).Vertices().Print();
	  }

	  if (mode)
	       RenderSolid(View3D(solid*CTM, Peye, Pfix, UP, hither, yon, 
				  150.0, 150.0, 150.0, 150.0), 300, 300);
	  else
	       DrawSolid(View3D(solid*CTM, Peye, Pfix, UP, hither, yon, 
				150.0, 150.0, 150.0, 150.0));

	  ForceOutput();

	  if (scanf("%1s", &z) == EOF)
	       break;

	  switch (z) {
	  case 'm':
	  case 'M':
	       mode = !mode;
	       break;
	  case 'c':
	  case 'C':
	       scanf("%f %f %f", &a, &b, &c);
	       border = a; hither = b; yon = c;
	       break;
	  case 'e':
	  case 'E':
	       scanf("%f %f %f", &a, &b, &c);
	       Peye(1, 1) = a; Peye(1, 2) = b; Peye(1, 3) = c;
	       break;
	  case 'f':
	  case 'F':
	       scanf("%f %f %f", &a, &b, &c);
	       Pfix(1, 1) = a; Pfix(1, 2) = b; Pfix(1, 3) = c;
	       break;
	  case 'u':
	  case 'U':
	       scanf("%f %f %f", &a, &b, &c);
	       UP(1, 1) = a; UP(1, 2) = b; UP(1, 3) = c;
	       break;
	  case 't':
	  case 'T':
	       scanf("%f %f %f", &a, &b, &c);
	       CTM = CTM * Translator(a, b, c);
	       break;
	  case 's':
	  case 'S':
	       scanf("%f %f %f", &a, &b, &c);
	       CTM = CTM * Scaler(a, b, c);
	       break;
	  case 'r':
	  case 'R':
	       scanf("%1s %f", &z, &a);
	       switch (z) {
	       case 'X':
	       case 'x':
		    CTM = CTM * Rotator_X(a);
		    break;
	       case 'Y':
	       case 'y':
		    CTM = CTM * Rotator_Y(a);
		    break;
	       case 'Z':
	       case 'z':
		    CTM = CTM * Rotator_Z(a);
		    break;
	       }
	       break;
	  case 'i':
	  case 'I':
	       CTM.Identify();
	       break;
	  }
     }

     /* Reset the EOF condition */
     clearerr(stdin);
     printf("\n");
}
     
main(int argc, char **argv)
{
     Panel p;

     if (argc < 2) {
	  fprintf(stderr, "Usage: ps5-2 filename.\n");
	  exit(1);
     }

     p = NewPanel(WIDTH, HEIGHT);
     SetFunction(DRAW);
     UsePanel(p);

     problem2(argv[1]);

     DoneWithPanel(p);
     return 0;
}
