#include <stdio.h>
#include <math.h>
#include "Matrix.H"
#include "MatrixMisc.H"
#include "halftone.h"

extern "C" {
#include <tgif/tgif.h>
int keypressed();
}

inline double DotProd3(Matrix& m1, Matrix& m2)
{
     return m1(1, 1)*m2(1, 1)+m1(1, 2)*m2(1, 2)+m1(1, 3)*m2(1, 3);
}

void RenderSphere(double r, double z0, const Matrix& lsource, const
		  Matrix& viewpoint, double Ia, double Ip, double ka,
		  double ks, double kd, double n, int W, int H, int D,
		  int S) 
{
     Matrix N(1, 3), L(1, 3), R(1, 3), V(1, 3), C(1, 3), I(W, H);
     double x, y, z, xmin, xmax, ymin, ymax, NdotL, RdotV, bright;
     double Iaka = Ia*ka, Ipkd = Ip*kd, Ipks = Ip*ks;
     int b, xp, yp;

     for (int i = 1; i <= W; i++)
	  for (int j = 1; j <= H; j++)
	       I(i, j) = 0.0;

     C(1, 1) = 0.0; C(1, 2) = 0.0; C(1, 3) = z0;

     L = Unitize(lsource - C);
     V = Unitize(viewpoint - C);

     for (y = -r; y <= r; y++) {
	  xmax = sqrt(sqr(r) - sqr(y));
	  for (x = -xmax; x <= xmax; x++) {
	       z = z0 - sqrt(xmax*xmax - x*x);

	       N(1, 1) = x / r;
	       N(1, 2) = y / r;
	       N(1, 3) = (z - z0) / r;
	       
	       NdotL = DotProd3(N, L);
	       R = 2.0*NdotL*N - L;
	       RdotV = DotProd3(R, V);
	       
	       if (NdotL < 0.0) 
		    NdotL = 0.0;
	       
	       if (RdotV < 0.0)
		    RdotV = 0.0;
	       
	       bright = Iaka + Ipkd*NdotL + Ipks*pow(RdotV, n); 
	       
	       xp = int(W*D*x/(2.0*S*z)) + W/2;
	       yp = int(H*D*y/(2.0*S*z)) + H/2;
	       
	       I(xp, yp) = bright;
	  }
     }
     
     FS_Halftone(I, 100.0);
}

void problem1(int W, int H, int D, int S)
{
     Matrix VP(1, 3, 0, 0, 0);
     Matrix LS(1, 3, -100, 100, 0);
     double Ia = 50.0, Ip = 100.0, ka = 0.4, kd = 0.5;
     double ks = 0.5, r = 70.0, z0 = 150.0, n = 2.0; 
     double a, b, c;
     char z;

     while (1) {
	  Clear();
	  RenderSphere(r, z0, LS, VP, Ia, Ip, ka, ks, kd, n,
		       W, H, D, S);
	  ForceOutput();

	  if (scanf("%1s", &z) == EOF)
	       break;

	  switch (z) {
	  case 'G':
	  case 'g':
	       break;
	  case 'l':
	  case 'L':
	       scanf("%f %f %f", &a, &b, &c);
	       LS(1, 1) = a;
	       LS(1, 2) = b;
	       LS(1, 3) = c;
	       break;
	  case 'z':
	  case 'Z':
	       scanf("%d", &z0);
	       break;
	  case 'n':
	  case 'N':
	       scanf("%f", &n);
	       break;
	  case 'i':
	  case 'I':
	       scanf("%1s %f", &z, &a);
	       switch (z) {
	       case 'a':
	       case 'A':
		    Ia = a;
		    break;
	       case 'p':
	       case 'P':
		    Ip = a;
		    break;
	       }
	       break;
	  case 'k':
	       scanf("%1s %f", &z, &a);
	       switch (z) {
	       case 'a':
		    ka = a;
		    break;
	       case 's':
		    ks = a;
		    break;
	       case 'd':
		    kd = a;
		    break;
	       }
	       break;
	  }
     }
}

main()
{
     Panel p;

     p = NewPanel(150, 150);
     UsePanel(p);
     SetFunction(DRAW);

     problem1(150, 150, 100, 150);

     DoneWithPanel(p);
}
