#include <stdio.h>
#include <stdlib.h>
#include "Matrix.H"
#include "MatrixMisc.H"
#include "LightSolid.H"
#include "CTM.H"
#include "Transform.H"
#include "View3D.h"
#include "render.h"

extern "C" {
#include <tgif/tgif.h>
}

const int WIDTH = 300;
const int HEIGHT = 300;
const int debug = 1;

int CubePoints[] = {
     /* Vertices */
     -50, -50, -50, 1,
     -50, 50, -50, 1,
     50, 50, -50, 1,
     50, -50, -50, 1,
     -50, -50, 50, 1,
     -50, 50, 50, 1,
     50, 50, 50, 1,
     50, -50, 50, 1
};
int CubeNormalPoints[] = {
     1, 0, 0, 1,
     0, 1, 0, 1,
     0, 0, 1, 1,
     -1, 0, 0, 1,
     0, -1, 0, 1,
     0, 0, -1, 1,
};
int CubeNormals[] = { 1, 2, 3, 4, 5, 6 };
double CubeKa[] = { 0.4, 0.4, 0.4, 0.4, 0.4, 0.4 };
double CubeKd[] = { 0.5, 0.5, 0.5, 0.5, 0.5, 0.5 };
double CubeKs[] = { 0.5, 0.5, 0.5, 0.5, 0.5, 0.5 };

int front[] = { 1, 4, 3, 2 };
int left[] = { 1, 2, 6, 5 };
int top[] = { 2, 3, 7, 6 };
int right[] = { 3, 4, 8, 7 };
int back[] = { 5, 6, 7, 8 };
int bottom[] = { 1, 5, 8, 4 };

int *CubeFaces[6] = {
     front, left, top, right, back, bottom,
};

Matrix pyrpts(5, 4,
	      -50, 0, -50, 1,
	      50, 0, -50, 1,
	      50, 0, 50, 1,
	      -50, 0, 50, 1,
	      0, 100, 0, 1);
Matrix pyrnrms(5, 4,
	       0.0, -1.0, 0.0, 1.0,	// bottom
	       0.894, 0.447, 0.0, 1.0,	// right
	       -0.894, 0.447, 0.0, 1.0,	// left
	       0.0, 0.447, -0.894, 1.0,	// front
	       0.0, 0.447, 0.894, 1.0); // back
	       

int p1[] = {1, 2, 3, 4};
int p2[] = {2, 1, 5};
int p3[] = {3, 2, 5};
int p4[] = {4, 3, 5};
int p5[] = {1, 4, 5};

int *pyrfaces[] = {p1, p2, p3, p4, p5};
int pyrnormals[] = { 1, 4, 2, 5, 3};
int pyrnfaces[] = {4, 3, 3, 3, 3};

double PyrKa[] = { 0.2, 0.2, 0.2, 0.2, 0.2, 0.2 };
double PyrKd[] = { 0.8, 0.8, 0.8, 0.8, 0.8, 0.8 };
double PyrKs[] = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };

double DegToRad(double d)
{
     return d * PI / 180.0;
}

#define DLine(a,b,c,d) Line(int(a), int(b), int(c), int(d))
void DrawSolid(const Solid& s)
{
     int i, j;

     for (i = 0; i < s.NumFaces(); i++) {
	  SolidFace *face = s.Face(i);
	  int *p = face->VIndxs();
	  for (j = 0; j < face->NumVerts()-1; j++) {
	       DLine(s.Vertices()(p[j], 1), s.Vertices()(p[j], 2), 
		     s.Vertices()(p[j+1], 1), s.Vertices()(p[j+1], 2));
	  }
	  DLine(s.Vertices()(p[j], 1), s.Vertices()(p[j], 2),
		s.Vertices()(p[0], 1), s.Vertices()(p[0], 2));
     }
}

void problem2(char *filename)
{
     CTM ctm;
     double S, F, d_s, d_f, border = 20.0, hither = 10.0, yon = 500.0;
     double Ia = 50.0, Ip = 100.0, n = 2.0;
     double a, b, c;
     char z, mode = 0;
     int i;

     /* the viewing situation */
     Matrix Peye(1, 3, 0, 0, -350);
     Matrix UP(1, 3, 0, 1, 0);
     Matrix Pfix(1, 3, 0, 0, 0);

     /* the lighting situation */
     Matrix lsource(1, 3, -100, 100, -350);

     Matrix *points = new Matrix(8, 4, CubePoints);
     Matrix *normals = new Matrix(6, 4, CubeNormalPoints);
     
     Solid cube(*points, *normals);
     for (i = 0; i < 6; i++) {
	  SolidFace face(4, CubeFaces[i], 1, &CubeNormals[i],
			 CubeKa[i], CubeKd[i], CubeKs[i]);
	  cube.AddFace(face);
     }

     Solid pyr(pyrpts*Translate(0.0, 0.0, -100.0), pyrnrms);
     for (i = 0; i < 5; i++) {
	  SolidFace face(pyrnfaces[i], pyrfaces[i], 1, &pyrnormals[i],
			 PyrKa[i], PyrKd[i], PyrKs[i]);
	  pyr.AddFace(face);
     }

     Solid solid(cube + pyr);

     if (debug) {
	  printf("Solid:\n");
	  solid.Print();
     }	  
     
     while (1) {
	  Clear();

	  if (debug) {
	       printf("Transformed solid:\n");
	       View3D(solid*ctm, Peye, Pfix, UP, hither, yon, 
		      150.0, 150.0, 150.0, 150.0).Print();
	  }

	  if (mode) {
	       Solid LitSolid(ConstLightSolid(solid*ctm, Ia, lsource,
					      Peye, Ip, n)); 
	       RenderSolid(View3D(LitSolid, Peye, Pfix, UP, hither, yon, 
				  150.0, 150.0, 150.0, 150.0), 300, 300);
	  }
	  else
	       DrawSolid(View3D(solid*ctm, Peye, Pfix, UP, hither, yon, 
				150.0, 150.0, 150.0, 150.0));

	  ForceOutput();

	  if (scanf("%1s", &z) == EOF)
	       break;

	  switch (z) {
	  case 'I':
	       scanf("%1s %f", &z, &a);
	       switch (z) {
	       case 'a':
		    Ia = a;
		    break;
	       case 'p':
		    Ip = a;
		    break;
	       }
	       break;
	  case 'l':
	       scanf("%f %f %f", &a, &b, &c);
	       lsource(1, 1) = a; lsource(1, 2) = b; lsource(1, 3) = c;
	       break;
	  case 'm':
	  case 'M':
	       mode = !mode;
	       break;
	  case 'n':
	  case 'N':
	       scanf("%f", &n);
	       break;
	  case 'c':
	  case 'C':
	       scanf("%f %f %f", &border, &hither, &yon);
	       break;
	  case 'e':
	  case 'E':
	       scanf("%f %f %f", &a, &b, &c);
	       Peye(1, 1) = a; Peye(1, 2) = b; Peye(1, 3) = c;
	       break;
	  case 'f':
	  case 'F':
	       scanf("%f %f %f", &a, &b, &c);
	       Pfix(1, 1) = a; Pfix(1, 2) = b; Pfix(1, 3) = c;
	       break;
	  case 'u':
	  case 'U':
	       scanf("%f %f %f", &a, &b, &c);
	       UP(1, 1) = a; UP(1, 2) = b; UP(1, 3) = c;
	       break;
	  case 't':
	  case 'T':
	       scanf("%f %f %f", &a, &b, &c);
	       ctm.Translate(a, b, c);
	       break;
	  case 's':
	  case 'S':
	       scanf("%f %f %f", &a, &b, &c);
	       ctm.Scale(a, b, c);
	       break;
	  case 'r':
	  case 'R':
	       scanf("%1s %f", &z, &a);
	       switch (z) {
	       case 'X':
	       case 'x':
		    ctm.RotateX(DegToRad(a));
		    break;
	       case 'Y':
	       case 'y':
		    ctm.RotateY(DegToRad(a));
		    break;
	       case 'Z':
	       case 'z':
		    ctm.RotateZ(DegToRad(a));
		    break;
	       }
	       break;
	  case 'i':
	       ctm.Identify();
	       break;
	  }
     }

     /* Reset the EOF condition */
     clearerr(stdin);
     printf("\n");
}
     
main(int argc, char **argv)
{
     Panel p;

     p = NewPanel(WIDTH, HEIGHT);
     SetFunction(DRAW);
     UsePanel(p);

     problem2(argv[1]);

     DoneWithPanel(p);
     return 0;
}
