#include <stdio.h>
#include <stdlib.h>
#include "Matrix.H"
#include "MatrixMisc.H"
#include "LightSolid.H"
#include "CTM.H"
#include "View3D.h"
#include "render.h"

extern "C" {
#include <tgif/tgif.h>
}

const int WIDTH = 300;
const int HEIGHT = 300;
const int debug = 0;

int CubePoints[] = {
     /* Vertices */
     -50, -50, -50, 1,
     -50, 50, -50, 1,
     50, 50, -50, 1,
     50, -50, -50, 1,
     -50, -50, 50, 1,
     -50, 50, 50, 1,
     50, 50, 50, 1,
     50, -50, 50, 1
};
int CubeNormalPoints[] = {
     0, 0, -1, 1,
     -1, 0, 0, 1,
     0, 1, 0, 1,
     1, 0, 0, 1,
     0, 0, 1, 1,
     0, -1, 0, 1,
};
int CubeNormals[] = {
     1, 1, 1, 1,
     2, 2, 2, 2,
     3, 3, 3, 3,
     4, 4, 4, 4,
     5, 5, 5, 5,
     6, 6, 6, 6
};
double CubeKa[] = { 0.4, 0.4, 0.4, 0.4, 0.4, 0.4 };
double CubeKd[] = { 0.5, 0.5, 0.5, 0.5, 0.5, 0.5 };
double CubeKs[] = { 0.5, 0.5, 0.5, 0.5, 0.5, 0.5 };

int front[] = { 1, 4, 3, 2 };
int left[] = { 1, 2, 6, 5 };
int top[] = { 2, 3, 7, 6 };
int right[] = { 3, 4, 8, 7 };
int back[] = { 5, 6, 7, 8 };
int bottom[] = { 1, 5, 8, 4 };

int *CubeFaces[6] = {
     front, left, top, right, back, bottom,
};

Matrix pyrpts(5, 4,
	      -50, 0, -50, 1,
	      50, 0, -50, 1,
	      50, 0, 50, 1,
	      -50, 0, 50, 1,
	      0, 100, 0, 1);
Matrix pyrnrms(5, 4,
	       0.0, -1.0, 0.0, 1.0,	// bottom
	       0.894, 0.447, 0.0, 1.0,	// right
	       -0.894, 0.447, 0.0, 1.0,	// left
	       0.0, 0.447, -0.894, 1.0,	// front
	       0.0, 0.447, 0.894, 1.0); // back
	       

int p1[] = {1, 2, 3, 4};
int p2[] = {2, 1, 5};
int p3[] = {3, 2, 5};
int p4[] = {4, 3, 5};
int p5[] = {1, 4, 5};

int *pyrfaces[] = {p1, p2, p3, p4, p5};
int pyrnormals[] = {
     1, 1, 1, 1,
     4, 4, 4,
     2, 2, 2,
     5, 5, 5,
     3, 3, 3,
};
int pyrnfaces[] = {4, 3, 3, 3, 3};

double PyrKa[] = { 0.2, 0.2, 0.2, 0.2, 0.2, 0.2 };
double PyrKd[] = { 0.8, 0.8, 0.8, 0.8, 0.8, 0.8 };
double PyrKs[] = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };

Matrix CylinderPts(16, 4,
		   30.0, 100.0, 12.42, 1.0,
		   12.42, 100.0, 30.0, 1.0,
		   -12.42, 100.0, 30.0, 1.0,
		   -30.0, 100.0, 12.42, 1.0,
		   -30.0, 100.0, -12.42, 1.0,
		   -12.42, 100.0, -30.0, 1.0,
		   12.42, 100.0, -30.0, 1.0,
		   30.0, 100.0, -12.42, 1.0,

		   30.0, -100.0, 12.42, 1.0,
		   12.42, -100.0, 30.0, 1.0,
		   -12.42, -100.0, 30.0, 1.0,
		   -30.0, -100.0, 12.42, 1.0,
		   -30.0, -100.0, -12.42, 1.0,
		   -12.42, -100.0, -30.0, 1.0,
		   12.42, -100.0, -30.0, 1.0,
		   30.0, -100.0, -12.42, 1.0);
Matrix CylinderNorms(10, 4,
		     0.0, 1.0, 0.0, 1.0,
		     0.0, -1.0, 0.0, 1.0,
		     0.924, 0.0, 0.383, 1.0,
		     0.383, 0.0, 0.924, 1.0,
		     -0.383, 0.0, 0.924, 1.0,
		     -0.924, 0.0, 0.383, 1.0,
		     -0.924, 0.0, -0.383, 1.0,
		     -0.383, 0.0, -0.924, 1.0,
		     0.383, 0.0, -0.924, 1.0,
		     0.924, 0.0, -0.383, 1.0);
		     
int CylinderTop[] = { 1, 2, 3, 4, 5, 6, 7, 8 };
int CylinderBottom[] = { 9, 10, 11, 12, 13, 14, 15, 16 };
int Cyl1[] = { 1, 9, 10, 2 };
int Cyl2[] = { 2, 10, 11, 3 };
int Cyl3[] = { 3, 11, 12, 4 };
int Cyl4[] = { 4, 12, 13, 5 };
int Cyl5[] = { 5, 13, 14, 6 };
int Cyl6[] = { 6, 14, 15, 7 };
int Cyl7[] = { 7, 15, 16, 8 };
int Cyl8[] = { 8, 16, 9, 1 };
int *CylinderFaces[10] = {
     CylinderTop, CylinderBottom, Cyl1, Cyl2, Cyl3, Cyl4, Cyl5, Cyl6,
     Cyl7, Cyl8,
};
int CylinderNIndxs[] = {
     1, 1, 1, 1, 1, 1, 1, 1,
     2, 2, 2, 2, 2, 2, 2, 2,
     3, 3, 4, 4, 0, 0, 0, 0,
     4, 4, 5, 5, 0, 0, 0, 0,
     5, 5, 6, 6, 0, 0, 0, 0,
     6, 6, 7, 7, 0, 0, 0, 0,
     7, 7, 8, 8, 0, 0, 0, 0,
     8, 8, 9, 9, 0, 0, 0, 0,
     9, 9, 10, 10, 0, 0, 0, 0,
     10, 10, 3, 3, 0, 0, 0, 0,
};
double CylinderKa = 0.3;
double CylinderKd = 0.5;
double CylinderKs = 0.3;

inline double DegToRad(double d)
{
     return d * PI / 180.0;
}

#define DLine(a,b,c,d) Line(int(a), int(b), int(c), int(d))
void DrawSolid(const Solid& s)
{
     int i, j;

     for (i = 0; i < s.NumFaces(); i++) {
	  SolidFace *face = s.Face(i);
	  int *p = face->VIndxs();
	  for (j = 0; j < face->NumVerts()-1; j++) {
	       DLine(s.Vertices()(p[j], 1), s.Vertices()(p[j], 2), 
		     s.Vertices()(p[j+1], 1), s.Vertices()(p[j+1], 2));
	  }
	  DLine(s.Vertices()(p[j], 1), s.Vertices()(p[j], 2),
		s.Vertices()(p[0], 1), s.Vertices()(p[0], 2));
     }
}

void problem3()
{
     CTM ctm;
     double S, F, d_s, d_f, border = 20.0, hither = 10.0, yon = 500.0;
     double Ia = 50.0, Ip = 100.0, n = 4.0;
     double a, b, c;
     char z, mode = 0;
     int i;

     /* the viewing situation */
     Matrix Peye(1, 3, 0, 0, -350);
     Matrix UP(1, 3, 0, 1, 0);
     Matrix Pfix(1, 3, 0, 0, 0);

     /* the lighting situation */
     Matrix lsource(Peye);

     Matrix *points = new Matrix(8, 4, CubePoints);
     Matrix *normals = new Matrix(6, 4, CubeNormalPoints);
     
     Solid cube(*points, *normals);
     for (i = 0; i < 6; i++) {
	  SolidFace face(4, CubeFaces[i], 4, &CubeNormals[i*4],
			 CubeKa[i], CubeKd[i], CubeKs[i]);
	  cube.AddFace(face);
     }

#if 0
     Solid pyr(pyrpts*Translate(0.0, 0.0, -100.0), pyrnrms);
     for (i = 0; i < 5; i++) {
	  SolidFace face(pyrnfaces[i], pyrfaces[i], pyrnfaces[i],
			 &pyrnormals[i ? i*3 + 1 : 0],
			 PyrKa[i], PyrKd[i], PyrKs[i]);
	  pyr.AddFace(face);
     }
#endif

     Solid cylinder(CylinderPts, CylinderNorms);
     for (i = 0; i < 2; i++) {
	  SolidFace face(8, CylinderFaces[i], 8, &CylinderNIndxs[i*8],
			 CylinderKa, CylinderKd, CylinderKs);
	  cylinder.AddFace(face);
     }
     for (i = 2; i < 10; i++) {
	  SolidFace face(4, CylinderFaces[i], 4, &CylinderNIndxs[i*8],
			 CylinderKa, CylinderKd, CylinderKs);
	  cylinder.AddFace(face);
     }

     Solid solid(cube + cylinder);

     if (debug) {
	  printf("Solid:\n");
	  solid.Print();
     }	  
     
     while (1) {
	  Clear();

	  if (debug) {
	       printf("Transformed solid:\n");
	       View3D(solid*ctm, Peye, Pfix, UP, hither, yon, 
		      150.0, 150.0, 150.0, 150.0).Print();
	  }

	  if (mode) {
	       Solid LitSolid(GouraudLightSolid(solid*ctm, Ia, lsource,
						Peye, Ip, n)); 
	       RenderSolid(View3D(LitSolid, Peye, Pfix, UP, hither, yon, 
				  150.0, 150.0, 150.0, 150.0), 300, 300);
	  }
	  else
	       DrawSolid(View3D(solid*ctm, Peye, Pfix, UP, hither, yon, 
				150.0, 150.0, 150.0, 150.0));

	  ForceOutput();

	  if (scanf("%1s", &z) == EOF)
	       break;

	  switch (z) {
	  case 'I':
	       scanf("%1s %f", &z, &a);
	       switch (z) {
	       case 'a':
		    Ia = a;
		    break;
	       case 'p':
		    Ip = a;
		    break;
	       }
	       break;
	  case 'l':
	       scanf("%f %f %f", &a, &b, &c);
	       lsource(1, 1) = a; lsource(1, 2) = b; lsource(1, 3) = c;
	       break;
	  case 'm':
	  case 'M':
	       mode = !mode;
	       break;
	  case 'n':
	  case 'N':
	       scanf("%f", &n);
	       break;
	  case 'c':
	  case 'C':
	       scanf("%f %f %f", &border, &hither, &yon);
	       break;
	  case 'e':
	  case 'E':
	       scanf("%f %f %f", &a, &b, &c);
	       Peye(1, 1) = a; Peye(1, 2) = b; Peye(1, 3) = c;
	       break;
	  case 'f':
	  case 'F':
	       scanf("%f %f %f", &a, &b, &c);
	       Pfix(1, 1) = a; Pfix(1, 2) = b; Pfix(1, 3) = c;
	       break;
	  case 'u':
	  case 'U':
	       scanf("%f %f %f", &a, &b, &c);
	       UP(1, 1) = a; UP(1, 2) = b; UP(1, 3) = c;
	       break;
	  case 't':
	  case 'T':
	       scanf("%f %f %f", &a, &b, &c);
	       ctm.Translate(a, b, c);
	       break;
	  case 's':
	  case 'S':
	       scanf("%f %f %f", &a, &b, &c);
	       ctm.Scale(a, b, c);
	       break;
	  case 'r':
	  case 'R':
	       scanf("%1s %f", &z, &a);
	       switch (z) {
	       case 'X':
	       case 'x':
		    ctm.RotateX(DegToRad(a), 0.0, 0.0, 0.0);
		    break;
	       case 'Y':
	       case 'y':
		    ctm.RotateY(DegToRad(a), 0.0, 0.0, 0.0);
		    break;
	       case 'Z':
	       case 'z':
		    ctm.RotateZ(DegToRad(a), 0.0, 0.0, 0.0);
		    break;
	       }
	       break;
	  case 'i':
	       ctm.Identify();
	       break;
	  }
     }

     /* Reset the EOF condition */
     clearerr(stdin);
     printf("\n");
}
     
main(int argc, char **argv)
{
     Panel p;

     p = NewPanel(WIDTH, HEIGHT);
     SetFunction(DRAW);
     UsePanel(p);

     problem3();

     DoneWithPanel(p);
     return 0;
}
