#include <stdio.h>
#include <math.h>

extern "C" {
#include <tgif/tgif.h>
}

#ifdef DEBUG
#define inline
#endif

int width = 100;
int height = 500;

void PlotToScale(int px, int py, int s)
{
     int x, y, l, r;

     /* s odd --> plot s/2 pixels on either side; */
     /* s even --> plot (s/2)-1 to the left, s/2 to the right */
     l = r = s / 2;
     if (! (s % 2)) l -= 1;

     for (x = px - l; x <= px + r; x += 1)
	  for (y = py - l; y <= py + r; y += 1)
	       Plot(x, y);
}

void DrawCircle(double Z)
{
     PlotToScale(width / 2, int(Z*100.0), 3);
}

/*
 * Z increases towards the top of the screen.  A positive force acts
 * towards the top of the screen.  A positive acceleration and/or
 * velocity is towards the top of the screen.
 */

inline double spring_accel(double Z, double R, double k, double M)
{
     /* A spring exerts a force proportional to the amount it is */
     /* stretched from its natural length. */
     if (Z < R)
	  return (R - Z) * k / M;
     else
	  return 0.0;
}

inline double damper_accel(double Z, double R, double v, double B, double M)
{
     /* A damper exerts a negative force proportional to the velocity */
     /* at which it is being compressed */
     if (Z < R)
	  return (-v) * B / M;
     else
	  return 0.0;
}

inline double gravity_accel()
{
     return -9.8;
}

inline double total_acceleration(double Z, double v, double R, double k,
			   double B, double M)
{
     return gravity_accel() + spring_accel(Z, R, k, M) +
	  damper_accel(Z, R, v, B, M); 
}

void bounce(double zinit, double R, double k, double B, double M,
	    double dt,  int iter)
{
     double velocity = 0.0, Z = zinit, acceleration;

     while (--iter > 0) {
	  Clear();
	  DrawCircle(Z);
	  ForceOutput();

	  Z += velocity * dt;
	  acceleration = total_acceleration(Z, velocity, R, k, B, M);
	  velocity += acceleration * dt;
#ifdef DEBUG
	  printf("Accel = %6.3f, Velo = %6.3f, Z = %6.3f\n",
		 acceleration, velocity,  Z);
#endif
     }
}

void problem1()
{
     double zinit = 4.0, k = 120.0, B = 1, M = 0.5, R = 0.5, dt = 0.01;
     int iter = 1000, i;
     double d;
     char c;

     while (1) {
	  if (scanf("%1s", &c) == EOF)
	       break;
	  
	  switch (c) {
	  case 'i':
	       scanf("%d", &iter);
	       break;
	  case 'z':
	       scanf("%f", &zinit);
	       break;
	  case 'k':
	       scanf("%f", &k);
	       break;
	  case 'b':
	       scanf("%f", &B);
	       break;
	  case 'm':
	       scanf("%f", &M);
	       break;
	  case 'r':
	       scanf("%f", &R);
	       break;
	  case 't':
	       scanf("%f", &dt);
	       break;
	  case 'g':
	       bounce(zinit, R, k, B, M, dt, iter);
	       break;
	  }
     }
}

main()
{
     Panel p;

     p = NewPanel(width, height);
     UsePanel(p);

     problem1();

     DoneWithPanel(p);
     return 0;
}
